@@ -60,15 +60,19 @@
 #include "slurm/slurm_errno.h"
 
 #include "src/common/parse_time.h"
-#include "src/interfaces/site_factor.h"
+#include "src/common/slurm_time.h"
+#include "src/common/state_save.h"
+#include "src/common/xstring.h"
+
+#include "src/interfaces/gres.h"
 #include "src/interfaces/mcs.h"
 #include "src/interfaces/priority.h"
-#include "src/common/slurm_time.h"
-#include "src/common/xstring.h"
-#include "src/interfaces/gres.h"
+#include "src/interfaces/site_factor.h"
 
+#include "src/slurmctld/acct_policy.h"
 #include "src/slurmctld/licenses.h"
 #include "src/slurmctld/read_config.h"
+#include "src/slurmctld/state_save.h"
 
 #include "fair_tree.h"
 
@@ -82,8 +86,9 @@
 #if defined (__APPLE__)
 extern void *acct_db_conn  __attribute__((weak_import));
 extern uint32_t cluster_cpus __attribute__((weak_import));
-extern List job_list  __attribute__((weak_import));
+extern list_t *job_list  __attribute__((weak_import));
 extern time_t last_job_update __attribute__((weak_import));
+extern time_t last_part_update __attribute__((weak_import));
 extern slurm_conf_t slurm_conf __attribute__((weak_import));
 extern int slurmctld_tres_cnt __attribute__((weak_import));
 extern uint16_t accounting_enforce __attribute__((weak_import));
@@ -91,8 +96,9 @@ extern int active_node_record_count __attribute__((weak_import));
 #else
 void *acct_db_conn = NULL;
 uint32_t cluster_cpus = NO_VAL;
-List job_list = NULL;
+list_t *job_list = NULL;
 time_t last_job_update = (time_t) 0;
+time_t last_part_update = (time_t) 0;
 slurm_conf_t slurm_conf;
 int slurmctld_tres_cnt = 0;
 uint16_t accounting_enforce = 0;
@@ -148,6 +154,12 @@ static time_t g_last_ran = 0; /* when the last poll ran */
 static time_t g_last_reset = 0; /* when the last reset was done */
 static double decay_factor = 1; /* The decay factor when decaying time. */
 
+typedef struct {
+	job_record_t *job_ptr;
+	part_record_t *part_ptr;
+	list_t *ret_list;
+} create_prio_list_t;
+
 /* variables defined in priority_multifactor.h */
 
 static void _priority_p_set_assoc_usage_debug(slurmdb_assoc_rec_t *assoc);
@@ -174,7 +186,7 @@ static void _destroy_priority_factors_obj_light(void *object)
 static int _apply_decay(double real_decay)
 {
 	int i;
-	ListIterator itr = NULL;
+	list_itr_t *itr = NULL;
 	slurmdb_assoc_rec_t *assoc = NULL;
 	slurmdb_qos_rec_t *qos = NULL;
 	assoc_mgr_lock_t locks = { WRITE_LOCK, NO_LOCK, WRITE_LOCK, NO_LOCK,
@@ -235,7 +247,7 @@ static int _apply_decay(double real_decay)
  */
 static int _reset_usage(void)
 {
-	ListIterator itr = NULL;
+	list_itr_t *itr = NULL;
 	slurmdb_assoc_rec_t *assoc = NULL;
 	slurmdb_qos_rec_t *qos = NULL;
 	int i;
@@ -323,10 +335,8 @@ unpack_error:
 static int _write_last_decay_ran(time_t last_ran, time_t last_reset)
 {
 	/* Save high-water mark to avoid buffer growth with copies */
-	static int high_buffer_size = BUF_SIZE;
+	static uint32_t high_buffer_size = BUF_SIZE;
 	int error_code = SLURM_SUCCESS;
-	int state_fd;
-	char *old_file, *new_file, *state_file;
 	buf_t *buffer;
 
 	if (!xstrcmp(slurm_conf.state_save_location, "/dev/null")) {
@@ -339,56 +349,9 @@ static int _write_last_decay_ran(time_t last_ran, time_t last_reset)
 	pack_time(last_ran, buffer);
 	pack_time(last_reset, buffer);
 
-	/* read the file */
-	old_file = xstrdup(slurm_conf.state_save_location);
-	xstrcat(old_file, "/priority_last_decay_ran.old");
-	state_file = xstrdup(slurm_conf.state_save_location);
-	xstrcat(state_file, "/priority_last_decay_ran");
-	new_file = xstrdup(slurm_conf.state_save_location);
-	xstrcat(new_file, "/priority_last_decay_ran.new");
+	error_code = save_buf_to_state("priority_last_decay_ran", buffer,
+				       &high_buffer_size);
 
-	lock_state_files();
-	state_fd = creat(new_file, 0600);
-	if (state_fd < 0) {
-		error("Can't save decay state, create file %s error %m",
-		      new_file);
-		error_code = errno;
-	} else {
-		int pos = 0, nwrite = get_buf_offset(buffer), amount;
-		char *data = (char *)get_buf_data(buffer);
-		high_buffer_size = MAX(nwrite, high_buffer_size);
-		while (nwrite > 0) {
-			amount = write(state_fd, &data[pos], nwrite);
-			if ((amount < 0) && (errno != EINTR)) {
-				error("Error writing file %s, %m", new_file);
-				error_code = errno;
-				break;
-			}
-			nwrite -= amount;
-			pos    += amount;
-		}
-		fsync(state_fd);
-		close(state_fd);
-	}
-
-	if (error_code != SLURM_SUCCESS)
-		(void) unlink(new_file);
-	else {			/* file shuffle */
-		(void) unlink(old_file);
-		if (link(state_file, old_file))
-			debug3("unable to create link for %s -> %s: %m",
-			       state_file, old_file);
-		(void) unlink(state_file);
-		if (link(new_file, state_file))
-			debug3("unable to create link for %s -> %s: %m",
-			       new_file, state_file);
-		(void) unlink(new_file);
-	}
-	xfree(old_file);
-	xfree(state_file);
-	xfree(new_file);
-
-	unlock_state_files();
 	debug4("done writing time %ld", (long)last_ran);
 	FREE_NULL_BUFFER(buffer);
 
@@ -405,10 +368,10 @@ static int _write_last_decay_ran(time_t last_ran, time_t last_reset)
  *
  * NOTE: acct_mgr_assoc_lock must be locked before this is called.
  */
-static int _set_children_usage_efctv(List children_list)
+static int _set_children_usage_efctv(list_t *children_list)
 {
 	slurmdb_assoc_rec_t *assoc = NULL;
-	ListIterator itr = NULL;
+	list_itr_t *itr = NULL;
 
 	if (!children_list || !list_count(children_list))
 		return SLURM_SUCCESS;
@@ -527,6 +490,116 @@ static double _get_tres_prio_weighted(double *tres_factors)
 	return tmp_tres;
 }
 
+typedef struct {
+	int counter;
+	job_record_t *job_ptr;
+	char *multi_prio_str;
+	part_record_t *part_ptr;
+	double part_tres;
+	double priority_part_init;
+} mult_prio_args_t;
+
+static int _priority_each_qos(void *object, void *arg)
+{
+	slurmdb_qos_rec_t *qos_ptr = object;
+	mult_prio_args_t *args = arg;
+	job_record_t *job_ptr = args->job_ptr;
+	double priority_qos = job_ptr->prio_factors->priority_qos;
+	double priority_part = args->priority_part_init;
+	uint64_t tmp_64;
+
+	if (qos_ptr && qos_ptr->priority && weight_qos) {
+		priority_qos =
+			((flags & PRIORITY_FLAGS_NO_NORMAL_QOS) ?
+			 qos_ptr->priority : qos_ptr->usage->norm_priority) *
+			(double) weight_qos;
+	}
+
+	priority_part +=
+		(job_ptr->prio_factors->priority_age +
+		 job_ptr->prio_factors->priority_assoc +
+		 job_ptr->prio_factors->priority_fs +
+		 job_ptr->prio_factors->priority_js +
+		 priority_qos +
+		 args->part_tres +
+		 (double)(((int64_t)job_ptr->prio_factors->priority_site)
+			  - NICE_OFFSET) -
+		 (double)(((int64_t)job_ptr->prio_factors->nice)
+			     - NICE_OFFSET));
+
+	/* Priority 0 is reserved for held jobs */
+	if (priority_part < 1)
+		priority_part = 1;
+
+	tmp_64 = (uint64_t) priority_part;
+	if (tmp_64 > 0xffffffff) {
+		error("%pJ priority '%"PRIu64"' exceeds 32 bits. Reducing it to 4294967295 (2^32 - 1)",
+		      job_ptr, tmp_64);
+		tmp_64 = 0xffffffff;
+		priority_part = (double) tmp_64;
+	}
+	if (((flags & PRIORITY_FLAGS_INCR_ONLY) == 0) ||
+	    (job_ptr->prio_mult->priority_array[args->counter] <
+	     (uint32_t) priority_part)) {
+		job_ptr->prio_mult->priority_array[args->counter] =
+			(uint32_t) priority_part;
+	}
+	if (slurm_conf.debug_flags & DEBUG_FLAG_PRIO) {
+		if (args->multi_prio_str)
+			xstrcat(args->multi_prio_str, ", ");
+		if (qos_ptr)
+			xstrfmtcat(args->multi_prio_str,
+				   "%s/%s=%u", args->part_ptr->name,
+				   qos_ptr->name,
+				   job_ptr->prio_mult->
+				   priority_array[args->counter]);
+		else
+			xstrfmtcat(args->multi_prio_str,
+				   "%s=%u", args->part_ptr->name,
+				   job_ptr->prio_mult->
+				   priority_array[args->counter]);
+	}
+
+	args->counter++;
+	return SLURM_SUCCESS;
+}
+
+static int _priority_each_partition(void *object, void *arg)
+{
+	mult_prio_args_t *args = arg;
+	part_record_t *part_ptr = (part_record_t *)object;
+	job_record_t *job_ptr = args->job_ptr;
+
+	args->part_ptr = part_ptr;
+
+	if (weight_tres) {
+		double part_tres_factors[slurmctld_tres_cnt];
+		memset(part_tres_factors, 0,
+		       sizeof(double) * slurmctld_tres_cnt);
+		_get_tres_factors(job_ptr, part_ptr, part_tres_factors);
+		args->part_tres = _get_tres_prio_weighted(part_tres_factors);
+	}
+
+	args->priority_part_init =
+		((flags & PRIORITY_FLAGS_NO_NORMAL_PART) ?
+		 part_ptr->priority_job_factor :
+		 part_ptr->norm_priority) * (double)weight_part;
+
+	if (job_ptr->qos_list) {
+		assoc_mgr_lock_t locks = {
+			.qos = READ_LOCK,
+		};
+		assoc_mgr_lock(&locks);
+		(void) list_for_each(job_ptr->qos_list,
+				     _priority_each_qos, args);
+		assoc_mgr_unlock(&locks);
+	} else {
+		(void) _priority_each_qos(NULL, args);
+	}
+
+	return SLURM_SUCCESS;
+}
+
 /* Returns the priority after applying the weight factors */
 static uint32_t _get_priority_internal(time_t start_time,
 				       job_record_t *job_ptr)
@@ -535,7 +608,6 @@ static uint32_t _get_priority_internal(time_t start_time,
 	priority_factors_t pre_factors;
 	uint64_t tmp_64;
 	double tmp_tres = 0.0;
-	char *multi_part_str = NULL;
 
 	if (job_ptr->direct_set_prio && (job_ptr->priority > 0)) {
 		if (job_ptr->prio_factors) {
@@ -612,80 +684,65 @@ static uint32_t _get_priority_internal(time_t start_time,
 		priority = (double) tmp_64;
 	}
 
-	if (job_ptr->part_ptr_list) {
-		part_record_t *part_ptr;
-		double priority_part;
-		ListIterator part_iterator;
+	/*
+	 * Free after transitioning from multi-part/qos job to single part job.
+	 */
+	if (!job_ptr->part_ptr_list && !job_ptr->qos_list &&
+	    job_ptr->prio_mult) {
+		xfree(job_ptr->prio_mult->priority_array);
+		xfree(job_ptr->prio_mult->priority_array_names);
+		xfree(job_ptr->prio_mult);
+	}
+
+	if (job_ptr->part_ptr_list || job_ptr->qos_list) {
 		int i = 0;
+		mult_prio_args_t arg = {
+			.job_ptr = job_ptr
+		};
 
-		if (!job_ptr->priority_array) {
-			i = list_count(job_ptr->part_ptr_list) + 1;
-			job_ptr->priority_array = xcalloc(i, sizeof(uint32_t));
+		if (!job_ptr->prio_mult)
+			job_ptr->prio_mult =
+				xmalloc(sizeof(*job_ptr->prio_mult));
+
+		if (job_ptr->prio_mult &&
+		    (!job_ptr->prio_mult->priority_array ||
+		     (job_ptr->prio_mult->last_update < last_part_update))) {
+
+			xfree(job_ptr->prio_mult->priority_array);
+			if (job_ptr->part_ptr_list) {
+				i = list_count(job_ptr->part_ptr_list);
+				/* part_ptr_list is already sorted */
+				xfree(job_ptr->prio_mult->priority_array_names);
+				job_ptr->prio_mult->priority_array_names =
+					part_list_to_xstr(
+						job_ptr->part_ptr_list);
+			}
+
+			if (job_ptr->qos_list) {
+				int qos_count = list_count(job_ptr->qos_list);
+				if (i)
+					i *= qos_count;
+				else
+					i = qos_count;
+			}
+
+			job_ptr->prio_mult->priority_array =
+				xcalloc(i, sizeof(uint32_t));
+
+			job_ptr->prio_mult->last_update = time(NULL);
 		}
 
-		i = 0;
-		list_sort(job_ptr->part_ptr_list, priority_sort_part_tier);
-		part_iterator = list_iterator_create(job_ptr->part_ptr_list);
-		while ((part_ptr = list_next(part_iterator))) {
-			double part_tres = 0.0;
+		if (job_ptr->part_ptr_list)
+			(void) list_for_each(job_ptr->part_ptr_list,
+					     _priority_each_partition,
+					     &arg);
+		else
+			(void) _priority_each_partition(
+				job_ptr->part_ptr, &arg);
 
-			if (weight_tres) {
-				double part_tres_factors[slurmctld_tres_cnt];
-				memset(part_tres_factors, 0,
-				       sizeof(double) * slurmctld_tres_cnt);
-				_get_tres_factors(job_ptr, part_ptr,
-						  part_tres_factors);
-				part_tres = _get_tres_prio_weighted(
-							part_tres_factors);
-			}
-
-			priority_part =
-				((flags & PRIORITY_FLAGS_NO_NORMAL_PART) ?
-				 part_ptr->priority_job_factor :
-				 part_ptr->norm_priority) *
-				(double)weight_part;
-			priority_part +=
-				 (job_ptr->prio_factors->priority_age
-				 + job_ptr->prio_factors->priority_assoc
-				 + job_ptr->prio_factors->priority_fs
-				 + job_ptr->prio_factors->priority_js
-				 + job_ptr->prio_factors->priority_qos
-				 + part_tres
-				 + (double)
-				   (((int64_t)job_ptr->prio_factors->priority_site)
-				    - NICE_OFFSET)
-				 - (double)
-				   (((int64_t)job_ptr->prio_factors->nice)
-				    - NICE_OFFSET));
-
-			/* Priority 0 is reserved for held jobs */
-			if (priority_part < 1)
-				priority_part = 1;
-
-			tmp_64 = (uint64_t) priority_part;
-			if (tmp_64 > 0xffffffff) {
-				error("%pJ priority '%"PRIu64"' exceeds 32 bits. Reducing it to 4294967295 (2^32 - 1)",
-				      job_ptr, tmp_64);
-				tmp_64 = 0xffffffff;
-				priority_part = (double) tmp_64;
-			}
-			if (((flags & PRIORITY_FLAGS_INCR_ONLY) == 0) ||
-			    (job_ptr->priority_array[i] <
-			     (uint32_t) priority_part)) {
-				job_ptr->priority_array[i] =
-					(uint32_t) priority_part;
-			}
-			if (slurm_conf.debug_flags & DEBUG_FLAG_PRIO) {
-				xstrfmtcat(multi_part_str, multi_part_str ?
-					   ", %s=%u" : "%s=%u", part_ptr->name,
-					   job_ptr->priority_array[i]);
-			}
-			i++;
-		}
-		log_flag(PRIO, "%pJ multi-partition priorities: %s",
-			 job_ptr, multi_part_str);
-		xfree(multi_part_str);
-		list_iterator_destroy(part_iterator);
+		log_flag(PRIO, "%pJ multi-partition/qos priorities: %s",
+			 job_ptr, arg.multi_prio_str);
+		xfree(arg.multi_prio_str);
 	}
 
 	if (slurm_conf.debug_flags & DEBUG_FLAG_PRIO) {
@@ -813,14 +870,25 @@ static time_t _next_reset(uint16_t reset_period, time_t last_reset)
 
 static void _handle_qos_tres_run_secs(long double *tres_run_decay,
 				      uint64_t *tres_run_delta,
-				      uint32_t job_id,
+				      job_record_t *job_ptr,
 				      slurmdb_qos_rec_t *qos)
 {
+	slurmdb_used_limits_t *used_limits_a = NULL;
+	slurmdb_used_limits_t *used_limits_u = NULL;
+
 	int i;
 
 	if (!qos || !(accounting_enforce & ACCOUNTING_ENFORCE_LIMITS))
 		return;
 
+	if (!job_ptr->assoc_ptr)
+		return;
+
+	used_limits_a = acct_policy_get_acct_used_limits(
+		&qos->usage->acct_limit_list, job_ptr->assoc_ptr->acct);
+	used_limits_u = acct_policy_get_user_used_limits(
+		&qos->usage->user_limit_list, job_ptr->user_id);
+
 	for (i=0; i<slurmctld_tres_cnt; i++) {
 		if (i == TRES_ARRAY_ENERGY)
 			continue;
@@ -833,7 +901,7 @@ static void _handle_qos_tres_run_secs(long double *tres_run_decay,
 			      "QOS %s TRES %s grp_used_tres_run_secs "
 			      "underflow, tried to remove %"PRIu64" seconds "
 			      "when only %"PRIu64" remained.",
-			      job_id,
+			      job_ptr->job_id,
 			      qos->name,
 			      assoc_mgr_tres_name_array[i],
 			      tres_run_delta[i],
@@ -844,9 +912,48 @@ static void _handle_qos_tres_run_secs(long double *tres_run_decay,
 				tres_run_delta[i];
 
 		log_flag(PRIO, "%s: job %u: Removed %"PRIu64" unused seconds from QOS %s TRES %s grp_used_tres_run_secs = %"PRIu64,
-			 __func__, job_id, tres_run_delta[i], qos->name,
+			 __func__, job_ptr->job_id, tres_run_delta[i], qos->name,
 			 assoc_mgr_tres_name_array[i],
 			 qos->usage->grp_used_tres_run_secs[i]);
+
+		if (tres_run_delta[i] >
+		    used_limits_a->tres_run_secs[i]) {
+			error("_handle_qos_tres_run_secs: job %u: "
+			      "QOS %s TRES %s account used limit tres_run_secs "
+			      "underflow, tried to remove %"PRIu64" seconds "
+			      "when only %"PRIu64" remained.",
+			      job_ptr->job_id,
+			      qos->name,
+			      assoc_mgr_tres_name_array[i],
+			      tres_run_delta[i],
+			      used_limits_a->tres_run_secs[i]);
+			used_limits_a->tres_run_secs[i] = 0;
+		} else
+			used_limits_a->tres_run_secs[i] -=
+				tres_run_delta[i];
+
+		log_flag(PRIO, "%s: job %u: Removed %"PRIu64" unused seconds from QOS %s TRES %s user used limit tres_run_secs = %"PRIu64,
+			 __func__, job_ptr->job_id, tres_run_delta[i], qos->name,
+			 assoc_mgr_tres_name_array[i],
+			 used_limits_a->tres_run_secs[i]);
+
+		if (tres_run_delta[i] >
+		    used_limits_u->tres_run_secs[i]) {
+			error("_handle_qos_tres_run_secs: job %u: QOS %s TRES %s user used limit tres_run_secs underflow, tried to remove %"PRIu64" seconds when only %"PRIu64" remained.",
+			      job_ptr->job_id,
+			      qos->name,
+			      assoc_mgr_tres_name_array[i],
+			      tres_run_delta[i],
+			      used_limits_u->tres_run_secs[i]);
+			used_limits_u->tres_run_secs[i] = 0;
+		} else
+			used_limits_u->tres_run_secs[i] -=
+				tres_run_delta[i];
+
+		log_flag(PRIO, "%s: job %u: Removed %"PRIu64" unused seconds from QOS %s TRES %s account used limit tres_run_secs = %"PRIu64,
+			 __func__, job_ptr->job_id, tres_run_delta[i], qos->name,
+			 assoc_mgr_tres_name_array[i],
+			 used_limits_a->tres_run_secs[i]);
 	}
 }
 
@@ -896,12 +1003,12 @@ static void _handle_tres_run_secs(uint64_t *tres_run_delta,
 	slurmdb_assoc_rec_t *assoc = job_ptr->assoc_ptr;
 
 	_handle_qos_tres_run_secs(NULL, tres_run_delta,
-				  job_ptr->job_id, job_ptr->qos_ptr);
+				  job_ptr, job_ptr->qos_ptr);
 
 	/* Only update partition qos if not being used by job */
 	if (job_ptr->part_ptr &&
 	    (job_ptr->part_ptr->qos_ptr != job_ptr->qos_ptr))
-		_handle_qos_tres_run_secs(NULL, tres_run_delta, job_ptr->job_id,
+		_handle_qos_tres_run_secs(NULL, tres_run_delta, job_ptr,
 					  job_ptr->part_ptr->qos_ptr);
 
 	while (assoc) {
@@ -925,7 +1032,7 @@ static void _handle_tres_run_secs(uint64_t *tres_run_delta,
 static void _init_grp_used_tres_run_secs(time_t last_ran)
 {
 	job_record_t *job_ptr = NULL;
-	ListIterator itr;
+	list_itr_t *itr;
 	assoc_mgr_lock_t locks = { WRITE_LOCK, NO_LOCK, WRITE_LOCK, NO_LOCK,
 				   READ_LOCK, NO_LOCK, NO_LOCK };
 	slurmctld_lock_t job_read_lock =
@@ -1131,14 +1238,14 @@ static int _apply_new_usage(job_record_t *job_ptr, time_t start_period,
 
 			_handle_qos_tres_run_secs(tres_run_nodecay,
 						  tres_run_delta,
-						  job_ptr->job_id, qos);
+						  job_ptr, qos);
 		} else {
 			qos->usage->grp_used_wall += run_decay;
 			qos->usage->usage_raw += (long double)real_decay;
 
 			_handle_qos_tres_run_secs(tres_run_decay,
 						  tres_run_delta,
-						  job_ptr->job_id, qos);
+						  job_ptr, qos);
 		}
 	}
 
@@ -1162,14 +1269,14 @@ static int _apply_new_usage(job_record_t *job_ptr, time_t start_period,
 
 			_handle_qos_tres_run_secs(tres_run_nodecay,
 						  tres_run_delta,
-						  job_ptr->job_id, qos);
+						  job_ptr, qos);
 		} else {
 			qos->usage->grp_used_wall += run_decay;
 			qos->usage->usage_raw += (long double)real_decay;
 
 			_handle_qos_tres_run_secs(tres_run_decay,
 						  tres_run_delta,
-						  job_ptr->job_id, qos);
+						  job_ptr, qos);
 		}
 	}
 
@@ -1369,21 +1476,30 @@ static void *_decay_thread(void *no_data)
 	return NULL;
 }
 
-static priority_factors_object_t *_create_prio_factors_obj(
-	job_record_t *job_ptr, part_record_t *job_part_ptr)
+static int _create_prio_list_qos(void *x, void *arg)
 {
+	slurmdb_qos_rec_t *qos_ptr = x;
+	create_prio_list_t *prio_list = arg;
+	part_record_t *part_ptr = prio_list->part_ptr;
+	job_record_t *job_ptr = prio_list->job_ptr;
+
 	priority_factors_object_t *obj =
 		xmalloc(sizeof(priority_factors_object_t));
 
+	if (!prio_list->ret_list)
+		prio_list->ret_list =
+			list_create(_destroy_priority_factors_obj_light);
+	list_append(prio_list->ret_list, obj);
+
 	/*
 	 * Don't xstrdup dup anything here, the list is freed
 	 * with _destroy_priority_factors_obj_light.
 	 */
 	obj->account = job_ptr->account;
 	obj->job_id = job_ptr->job_id;
-	obj->partition = job_part_ptr ?
-		job_part_ptr->name : job_ptr->part_ptr->name;
-	obj->qos = job_ptr->qos_ptr ? job_ptr->qos_ptr->name : NULL;
+	obj->partition = part_ptr ? part_ptr->name : job_ptr->part_ptr->name;
+	obj->qos = qos_ptr ? qos_ptr->name :
+		job_ptr->qos_ptr ? job_ptr->qos_ptr->name : NULL;
 	obj->user_id = job_ptr->user_id;
 
 	if (job_ptr->direct_set_prio) {
@@ -1394,127 +1510,48 @@ static priority_factors_object_t *_create_prio_factors_obj(
 					    job_ptr->prio_factors);
 
 		/* This portion is only needed for multi-partition jobs */
-		if (job_part_ptr) {
+		if (part_ptr) {
 			obj->prio_factors->priority_part =
 				((flags & PRIORITY_FLAGS_NO_NORMAL_PART) ?
-				 job_part_ptr->priority_job_factor :
-				 job_part_ptr->norm_priority) *
+				 part_ptr->priority_job_factor :
+				 part_ptr->norm_priority) *
 				(double)weight_part;
 			if (obj->prio_factors->priority_tres) {
 				_get_tres_factors(
-					job_ptr, job_part_ptr,
+					job_ptr, part_ptr,
 					obj->prio_factors->priority_tres);
 				_get_tres_prio_weighted(
 					obj->prio_factors->priority_tres);
 			}
 		}
+
+		if (qos_ptr && qos_ptr->priority && weight_qos) {
+			obj->prio_factors->priority_qos =
+				((flags & PRIORITY_FLAGS_NO_NORMAL_QOS) ?
+				 qos_ptr->priority :
+				 qos_ptr->usage->norm_priority) *
+				(double) weight_qos;
+		}
+
 	}
 
-	return obj;
+	return SLURM_SUCCESS;
 }
 
-/* this function can be removed 2 versions after 23.02 */
-/* If the specified job record satisfies the filter specifications in req_msg
- * and part_ptr_list (partition name filters), then add its priority specs
- * to ret_list */
-static void _filter_job(job_record_t *job_ptr,
-			priority_factors_request_msg_t *req_msg,
-			List part_ptr_list, List ret_list)
+static int _create_prio_list_part(void *x, void *arg)
 {
-	part_record_t *job_part_ptr = NULL, *filter_part_ptr = NULL;
-	List req_job_list, req_user_list;
-	int filter = 0;
-	ListIterator iterator, job_iter, filter_iter;
-	uint32_t *job_id;
-	uint32_t *user_id;
+	create_prio_list_t *prio_list = arg;
+	job_record_t *job_ptr = prio_list->job_ptr;
+	prio_list->part_ptr = x;
 
-	/* Filter by job ID */
-	req_job_list = req_msg->job_id_list;
-	if (req_job_list) {
-		filter = 1;
-		iterator = list_iterator_create(req_job_list);
-		while ((job_id = list_next(iterator))) {
-			if (*job_id == job_ptr->job_id) {
-				filter = 0;
-				break;
-			}
-		}
-		list_iterator_destroy(iterator);
-		if (filter == 1)
-			return;
-	}
+	if (job_ptr->qos_list)
+		(void) list_for_each(job_ptr->qos_list,
+				     _create_prio_list_qos,
+				     prio_list);
+	else
+		(void) _create_prio_list_qos(NULL, prio_list);
 
-	/* Filter by user/UID */
-	req_user_list = req_msg->uid_list;
-	if (req_user_list) {
-		filter = 1;
-		iterator = list_iterator_create(req_user_list);
-		while ((user_id = list_next(iterator))) {
-			if (*user_id == job_ptr->user_id) {
-				filter = 0;
-				break;
-			}
-		}
-		list_iterator_destroy(iterator);
-		if (filter == 1)
-			return;
-	}
-
-	/*
-	 * Job is not in any partition, so there is nothing to return.
-	 * This can happen if the Partition was deleted, CALCULATE_RUNNING
-	 * is enabled, and this job is still waiting out MinJobAge before
-	 * being removed from the system.
-	 */
-	if (!job_ptr->part_ptr && !job_ptr->part_ptr_list)
-		return;
-
-	/* Filter by partition, job in one partition */
-	if (!job_ptr->part_ptr_list) {
-		job_part_ptr =  job_ptr->part_ptr;
-		filter = 0;
-		if (part_ptr_list) {
-			filter = 1;
-			filter_iter = list_iterator_create(part_ptr_list);
-			while ((filter_part_ptr = list_next(filter_iter))) {
-				if (filter_part_ptr == job_part_ptr) {
-					filter = 0;
-					break;
-				}
-			}
-			list_iterator_destroy(filter_iter);
-		}
-
-		if (filter == 0) {
-			list_append(ret_list,
-				    _create_prio_factors_obj(job_ptr, NULL));
-		}
-		return;
-	}
-
-	/* Filter by partition, job in multiple partitions */
-	job_iter = list_iterator_create(job_ptr->part_ptr_list);
-	while ((job_part_ptr = list_next(job_iter))) {
-		filter = 0;
-		if (part_ptr_list) {
-			filter = 1;
-			filter_iter = list_iterator_create(part_ptr_list);
-			while ((filter_part_ptr = list_next(filter_iter))) {
-				if (filter_part_ptr == job_part_ptr) {
-					filter = 0;
-					break;
-				}
-			}
-			list_iterator_destroy(filter_iter);
-		}
-
-		if (filter == 0) {
-			list_append(ret_list,
-				    _create_prio_factors_obj(job_ptr,
-							     job_part_ptr));
-		}
-	}
-	list_iterator_destroy(job_iter);
+	return SLURM_SUCCESS;
 }
 
 static void _internal_setup(void)
@@ -1549,9 +1586,9 @@ static void _internal_setup(void)
 /* Reursively call assoc_mgr_normalize_assoc_shares from assoc_mgr.c on
  * children of an assoc
  */
-static void _set_norm_shares(List children_list)
+static void _set_norm_shares(list_t *children_list)
 {
-	ListIterator itr = NULL;
+	list_itr_t *itr = NULL;
 	slurmdb_assoc_rec_t *assoc = NULL;
 
 	if (!children_list || list_is_empty(children_list))
@@ -1617,7 +1654,7 @@ static void _depth_oblivious_set_usage_efctv(slurmdb_assoc_rec_t *assoc)
 {
 	long double ratio_p, ratio_l, k, f, ratio_s;
 	slurmdb_assoc_rec_t *parent_assoc = NULL;
-	ListIterator sib_itr = NULL;
+	list_itr_t *sib_itr = NULL;
 	slurmdb_assoc_rec_t *sibling = NULL;
 	char *child;
 	char *child_str;
@@ -1799,8 +1836,7 @@ int fini ( void )
 	slurm_mutex_unlock(&decay_lock);
 
 	/* Now join outside the lock */
-	if (decay_handler_thread)
-		pthread_join(decay_handler_thread, NULL);
+	slurm_thread_join(decay_handler_thread);
 
 	site_factor_g_fini();
 
@@ -1860,8 +1896,6 @@ extern void priority_p_reconfig(bool assoc_clear)
 	if (assoc_clear)
 		_init_grp_used_tres_run_secs(g_last_ran);
 
-	site_factor_g_reconfig();
-
 	debug2("%s reconfigured", plugin_name);
 
 	return;
@@ -1919,37 +1953,20 @@ extern double priority_p_calc_fs_factor(long double usage_efctv,
 	return priority_fs;
 }
 
-/* req_msg can be removed 2 versions after 23.02 */
-extern List priority_p_get_priority_factors_list(
-	priority_factors_request_msg_t *req_msg, uid_t uid)
+extern list_t *priority_p_get_priority_factors_list(uid_t uid)
 {
-	List ret_list = NULL, part_filter_list = NULL;
-	ListIterator itr, job_iter;
+	list_itr_t *itr;
 	job_record_t *job_ptr = NULL;
-	part_record_t *part_ptr, *job_part_ptr = NULL;
 	time_t start_time = time(NULL);
-	char *part_str, *tok, *last = NULL;
+	create_prio_list_t prio_list = { 0 };
 
 	xassert(verify_lock(JOB_LOCK, READ_LOCK));
 	xassert(verify_lock(NODE_LOCK, READ_LOCK));
 	xassert(verify_lock(PART_LOCK, READ_LOCK));
 
-	if (req_msg && req_msg->partitions) {
-		part_filter_list = list_create(NULL);
-		part_str = xstrdup(req_msg->partitions);
-		tok = strtok_r(part_str, ",", &last);
-		while (tok) {
-			if ((part_ptr = find_part_record(tok)))
-				list_append(part_filter_list, part_ptr);
-			tok = strtok_r(NULL, ",", &last);
-		}
-		xfree(part_str);
-	}
-
 	if (job_list && list_count(job_list)) {
 		time_t use_time;
 
-		ret_list = list_create(_destroy_priority_factors_obj_light);
 		itr = list_iterator_create(job_list);
 		while ((job_ptr = list_next(itr))) {
 			if (!(flags & PRIORITY_FLAGS_CALCULATE_RUNNING) &&
@@ -1989,13 +2006,6 @@ extern List priority_p_get_priority_factors_list(
 						     false) != 0))))
 				continue;
 
-			/* this can be removed 2 versions after 23.02 */
-			if(req_msg) {
-				_filter_job(job_ptr, req_msg, part_filter_list,
-					    ret_list);
-				continue;
-			}
-
 			/*
 			 * Job is not in any partition, so there is nothing to
 			 * return. This can happen if the Partition was deleted,
@@ -2006,30 +2016,23 @@ extern List priority_p_get_priority_factors_list(
 			if (!job_ptr->part_ptr && !job_ptr->part_ptr_list)
 				continue;
 
+			prio_list.job_ptr = job_ptr;
+
 			/* Job in one partition */
 			if (!job_ptr->part_ptr_list) {
-				list_append(ret_list,
-					    _create_prio_factors_obj(
-						    job_ptr, NULL));
-				continue;
+				_create_prio_list_part(NULL, &prio_list);
+			} else {
+				list_for_each(job_ptr->part_ptr_list,
+					      _create_prio_list_part,
+					      &prio_list);
 			}
-
-			/* Job in multiple partitions */
-			job_iter = list_iterator_create(job_ptr->part_ptr_list);
-			while ((job_part_ptr = list_next(job_iter))) {
-				list_append(ret_list,
-					    _create_prio_factors_obj(
-						    job_ptr, job_part_ptr));
-			}
-			list_iterator_destroy(job_iter);
 		}
 		list_iterator_destroy(itr);
-		if (!list_count(ret_list))
-			FREE_NULL_LIST(ret_list);
+		if (!list_count(prio_list.ret_list))
+			FREE_NULL_LIST(prio_list.ret_list);
 	}
-	FREE_NULL_LIST(part_filter_list);
 
-	return ret_list;
+	return prio_list.ret_list;
 }
 
 /* at least slurmctld_lock_t job_write_lock = { NO_LOCK, WRITE_LOCK,
