@@ -89,7 +89,6 @@
 #include "src/common/slurm_protocol_api.h"
 #include "src/common/slurm_protocol_socket.h"
 #include "src/common/uid.h"
-#include "src/common/xsignal.h"
 #include "src/common/xassert.h"
 #include "src/common/xmalloc.h"
 #include "src/common/xstring.h"
@@ -144,7 +143,7 @@ typedef struct {
 					 * will not do nodelist if set */
 	hostlist_t *nodelist;		/* list of nodes to send to */
 	char *nodename;			/* node to send to */
-	List ret_list;
+	list_t *ret_list;
 } thd_t;
 
 typedef struct {
@@ -207,7 +206,6 @@ static void _queue_update_node(char *node_name);
 static void _queue_update_srun(slurm_step_id_t *step_id);
 static int  _setup_requeue(agent_arg_t *agent_arg_ptr, thd_t *thread_ptr,
 			   int *count, int *spot);
-static void _sig_handler(int dummy);
 static void *_thread_per_group_rpc(void *args);
 static int   _valid_agent_arg(agent_arg_t *agent_arg_ptr);
 static void *_wdog(void *args);
@@ -221,10 +219,10 @@ static char **_build_mail_env(job_record_t *job_ptr, uint32_t mail_type);
 static pthread_mutex_t defer_mutex = PTHREAD_MUTEX_INITIALIZER;
 static pthread_mutex_t mail_mutex  = PTHREAD_MUTEX_INITIALIZER;
 static pthread_mutex_t retry_mutex = PTHREAD_MUTEX_INITIALIZER;
-static List defer_list = NULL;		/* agent_arg_t list for requests
+static list_t *defer_list = NULL;	/* agent_arg_t list for requests
 					 * requiring job write lock */
-static List mail_list = NULL;		/* pending e-mail requests */
-static List retry_list = NULL;		/* agent_arg_t list for retry */
+static list_t *mail_list = NULL;	/* pending e-mail requests */
+static list_t *retry_list = NULL;	/* agent_arg_t list for retry */
 
 static list_t *update_node_list = NULL;	/* node list for update */
 static pthread_mutex_t update_nodes_mutex = PTHREAD_MUTEX_INITIALIZER;
@@ -719,7 +717,7 @@ static void _notify_slurmctld_jobs(agent_info_t *agent_ptr)
 			*agent_ptr->msg_args_pptr;
 		step_id.job_id = msg->job_id;
 	} else if (agent_ptr->msg_type == RESPONSE_HET_JOB_ALLOCATION) {
-		List het_alloc_list = *agent_ptr->msg_args_pptr;
+		list_t *het_alloc_list = *agent_ptr->msg_args_pptr;
 		resource_allocation_response_msg_t *msg;
 		if (!het_alloc_list || (list_count(het_alloc_list) == 0))
 			return;
@@ -915,6 +913,33 @@ static int _wif_status(void)
 	return rc;
 }
 
+/*
+ * slurm_send_msg_maybe
+ * opens a connection, sends a message across while ignoring any errors,
+ * then closes the connection
+ *
+ * Open a connection to the "address" specified in the slurm msg `req'
+ * Then, immediately close the connection w/out waiting for a reply.
+ * Ignore any errors. This should only be used when you do not care if
+ * the message is ever received.
+ *
+ * IN request_msg	- slurm_msg request
+ */
+static void _send_msg_maybe(slurm_msg_t *req)
+{
+	int fd = -1;
+
+	if ((fd = slurm_open_msg_conn(&req->address)) < 0) {
+		log_flag(NET, "%s: slurm_open_msg_conn(%pA): %m",
+			 __func__, &req->address);
+		return;
+	}
+
+	(void) slurm_send_node_msg(fd, req);
+
+	(void) close(fd);
+}
+
 /*
  * _thread_per_group_rpc - thread to issue an RPC for a group of nodes
  *                         sending message out to one and forwarding it to
@@ -938,10 +963,9 @@ static void *_thread_per_group_rpc(void *args)
 	state_t thread_state = DSH_NO_RESP;
 	slurm_msg_type_t msg_type = task_ptr->msg_type;
 	bool is_kill_msg, srun_agent, sack_agent;
-	List ret_list = NULL;
+	list_t *ret_list = NULL;
 	list_itr_t *itr;
 	ret_data_info_t *ret_data_info = NULL;
-	int sig_array[2] = {SIGUSR1, 0};
 	/* Locks: Write job, write node */
 	slurmctld_lock_t job_write_lock = {
 		NO_LOCK, WRITE_LOCK, WRITE_LOCK, NO_LOCK, READ_LOCK };
@@ -954,8 +978,6 @@ static void *_thread_per_group_rpc(void *args)
 	uint32_t job_id;
 
 	xassert(args != NULL);
-	xsignal(SIGUSR1, _sig_handler);
-	xsignal_unblock(sig_array);
 	is_kill_msg = (	(msg_type == REQUEST_KILL_TIMELIMIT)	||
 			(msg_type == REQUEST_KILL_PREEMPTED)	||
 			(msg_type == REQUEST_TERMINATE_JOB) );
@@ -1049,7 +1071,7 @@ static void *_thread_per_group_rpc(void *args)
 			 * flings the message out and disregards any
 			 * communication problems that may arise.
 			 */
-			slurm_send_msg_maybe(&msg);
+			_send_msg_maybe(&msg);
 			thread_state = DSH_DONE;
 		} else if (slurm_send_only_node_msg(&msg) == SLURM_SUCCESS) {
 			thread_state = DSH_DONE;
@@ -1141,7 +1163,7 @@ static void *_thread_per_group_rpc(void *args)
 			/* Communication issue to srun that launched the job
 			 * Cancel rather than leave a stray-but-empty job
 			 * behind on the allocated nodes. */
-			List het_alloc_list = task_ptr->msg_args_ptr;
+			list_t *het_alloc_list = task_ptr->msg_args_ptr;
 			resource_allocation_response_msg_t *msg_ptr;
 			if (!het_alloc_list ||
 			    (list_count(het_alloc_list) == 0))
@@ -1289,14 +1311,6 @@ cleanup:
 	return NULL;
 }
 
-/*
- * Signal handler.  We are really interested in interrupting hung communictions
- * and causing them to return EINTR. Multiple interrupts might be required.
- */
-static void _sig_handler(int dummy)
-{
-}
-
 static int _setup_requeue(agent_arg_t *agent_arg_ptr, thd_t *thread_ptr,
 			  int *count, int *spot)
 {
@@ -1470,8 +1484,12 @@ static void *_agent_init(void *arg)
 					     &ts);
 		}
 		if (slurmctld_config.shutdown_time) {
-			slurm_mutex_unlock(&pending_mutex);
-			break;
+			if (!retry_list_size() ||
+			    (slurmctld_config.shutdown_time +
+			     AGENT_SHUTDOWN_WAIT < time(NULL))) {
+				slurm_mutex_unlock(&pending_mutex);
+				break;
+			}
 		}
 		mail_too = pending_mail;
 		min_wait = pending_wait_time;
@@ -1516,11 +1534,13 @@ static void *_agent_nodes_update(void *arg)
 			break;
 		}
 
-		if (!list_count(update_node_list))
-			continue;
-		lock_slurmctld(node_write_lock);
-		list_delete_all(update_node_list, _foreach_node_did_resp, NULL);
-		unlock_slurmctld(node_write_lock);
+		if (list_count(update_node_list)) {
+			lock_slurmctld(node_write_lock);
+			list_delete_all(update_node_list,
+					_foreach_node_did_resp, NULL);
+			unlock_slurmctld(node_write_lock);
+		}
+		ping_nodes_update();
 	}
 
 	return NULL;
@@ -1746,7 +1766,7 @@ static void _agent_defer(void)
 	lock_slurmctld(job_write_lock);
 	slurm_mutex_lock(&defer_mutex);
 	if (defer_list) {
-		List tmp_list = NULL;
+		list_t *tmp_list = NULL;
 		/* first try to find a new (never tried) record */
 		while ((queued_req_ptr = list_pop(defer_list))) {
 			agent_arg_ptr = queued_req_ptr->agent_arg_ptr;
@@ -1957,21 +1977,18 @@ extern void agent_purge(void)
 {
 	int i;
 
-	if (retry_list) {
-		slurm_mutex_lock(&retry_mutex);
-		FREE_NULL_LIST(retry_list);
-		slurm_mutex_unlock(&retry_mutex);
-	}
-	if (defer_list) {
-		slurm_mutex_lock(&defer_mutex);
-		FREE_NULL_LIST(defer_list);
-		slurm_mutex_unlock(&defer_mutex);
-	}
-	if (mail_list) {
-		slurm_mutex_lock(&mail_mutex);
-		FREE_NULL_LIST(mail_list);
-		slurm_mutex_unlock(&mail_mutex);
-	}
+	slurm_mutex_lock(&retry_mutex);
+	FREE_NULL_LIST(retry_list);
+	slurm_mutex_unlock(&retry_mutex);
+
+	slurm_mutex_lock(&defer_mutex);
+	FREE_NULL_LIST(defer_list);
+	slurm_mutex_unlock(&defer_mutex);
+
+	slurm_mutex_lock(&mail_mutex);
+	FREE_NULL_LIST(mail_list);
+	slurm_mutex_unlock(&mail_mutex);
+
 	slurm_mutex_lock(&update_nodes_mutex);
 	FREE_NULL_LIST(update_node_list);
 	slurm_mutex_unlock(&update_nodes_mutex);
