@@ -5,7 +5,7 @@
  *****************************************************************************
  *  Copyright (C) 2002-2007 The Regents of the University of California.
  *  Copyright (C) 2008-2010 Lawrence Livermore National Security.
- *  Portions Copyright (C) 2010-2016 SchedMD <https://www.schedmd.com>.
+ *  Copyright (C) SchedMD LLC.
  *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
  *  Written by Morris Jette <jette@llnl.gov> et. al.
  *  CODE-OCEC-09-009. All rights reserved.
@@ -56,12 +56,15 @@
 #include "src/common/hostlist.h"
 #include "src/common/list.h"
 #include "src/common/pack.h"
+#include "src/common/part_record.h"
 #include "src/common/slurm_protocol_pack.h"
 #include "src/common/slurm_resource_info.h"
+#include "src/common/state_save.h"
 #include "src/common/uid.h"
 #include "src/common/xstring.h"
 
 #include "src/interfaces/burst_buffer.h"
+#include "src/interfaces/priority.h"
 #include "src/interfaces/select.h"
 
 #include "src/slurmctld/gang.h"
@@ -89,7 +92,7 @@ typedef struct {
 } _foreach_pack_part_info_t;
 
 /* Global variables */
-List part_list = NULL;			/* partition list */
+list_t *part_list = NULL;		/* partition list */
 char *default_part_name = NULL;		/* name of default partition */
 part_record_t *default_part_loc = NULL;	/* default partition location */
 time_t last_part_update = (time_t) 0;	/* time of last update to partition records */
@@ -336,52 +339,16 @@ static void _unlink_free_nodes(bitstr_t *old_bitmap, part_record_t *part_ptr)
 }
 
 /*
- * Sync with _init_conf_part().
- *
- * _init_conf_part() initializes default values from slurm.conf parameters.
- * After parsing slurm.conf, _build_single_partitionline_info() copies
- * slurm_conf_partition_t to part_record_t. Default values between
- * slurm_conf_partition_t and part_record_t should stay in sync in case a
- * part_record_t is created outside of slurm.conf parsing.
- */
-static void _init_part_record(part_record_t *part_ptr)
-{
-	part_ptr->magic = PART_MAGIC;
-	if (slurm_conf.conf_flags & CTL_CONF_DRJ)
-		part_ptr->flags |= PART_FLAG_NO_ROOT;
-	part_ptr->max_nodes_orig = INFINITE;
-	part_ptr->min_nodes = 1;
-	part_ptr->min_nodes_orig = 1;
-
-	/* sync with slurm_conf_partition_t */
-	part_ptr->default_time = NO_VAL;
-	part_ptr->max_cpus_per_node = INFINITE;
-	part_ptr->max_cpus_per_socket = INFINITE;
-	part_ptr->max_nodes = INFINITE;
-	part_ptr->max_share = 1;
-	part_ptr->max_time = INFINITE;
-	part_ptr->over_time_limit = NO_VAL16;
-	part_ptr->preempt_mode = NO_VAL16;
-	part_ptr->priority_job_factor = 1;
-	part_ptr->priority_tier = 1;
-	part_ptr->resume_timeout = NO_VAL16;
-	part_ptr->state_up = PARTITION_UP;
-	part_ptr->suspend_time = NO_VAL;
-	part_ptr->suspend_timeout = NO_VAL16;
-}
-
-/*
- * create_part_record - create a partition record
+ * create_ctld_part_record - create a partition record
  * RET a pointer to the record or NULL if error
  * global: part_list - global partition list
  */
-part_record_t *create_part_record(const char *name)
+part_record_t *create_ctld_part_record(const char *name)
 {
-	part_record_t *part_ptr = xmalloc(sizeof(*part_ptr));
+	part_record_t *part_ptr = part_record_create();
 
 	last_part_update = time(NULL);
 
-	_init_part_record(part_ptr);
 	part_ptr->name = xstrdup(name);
 
 	list_append(part_list, part_ptr);
@@ -393,9 +360,7 @@ part_record_t *create_part_record(const char *name)
 int dump_all_part_state(void)
 {
 	/* Save high-water mark to avoid buffer growth with copies */
-	static int high_buffer_size = BUF_SIZE;
-	int error_code = 0, log_fd;
-	char *old_file, *new_file, *reg_file;
+	static uint32_t high_buffer_size = BUF_SIZE;
 	/* Locks: Read partition */
 	slurmctld_lock_t part_read_lock =
 	    { READ_LOCK, NO_LOCK, NO_LOCK, READ_LOCK, NO_LOCK };
@@ -411,60 +376,9 @@ int dump_all_part_state(void)
 	/* write partition records to buffer */
 	lock_slurmctld(part_read_lock);
 	list_for_each_ro(part_list, _dump_part_state, buffer);
-
-	old_file = xstrdup(slurm_conf.state_save_location);
-	xstrcat(old_file, "/part_state.old");
-	reg_file = xstrdup(slurm_conf.state_save_location);
-	xstrcat(reg_file, "/part_state");
-	new_file = xstrdup(slurm_conf.state_save_location);
-	xstrcat(new_file, "/part_state.new");
 	unlock_slurmctld(part_read_lock);
 
-	/* write the buffer to file */
-	lock_state_files();
-	log_fd = creat(new_file, 0600);
-	if (log_fd < 0) {
-		error("Can't save state, error creating file %s, %m",
-		      new_file);
-		error_code = errno;
-	} else {
-		int pos = 0, nwrite = get_buf_offset(buffer), amount, rc;
-		char *data = (char *)get_buf_data(buffer);
-		high_buffer_size = MAX(nwrite, high_buffer_size);
-		while (nwrite > 0) {
-			amount = write(log_fd, &data[pos], nwrite);
-			if ((amount < 0) && (errno != EINTR)) {
-				error("Error writing file %s, %m", new_file);
-				error_code = errno;
-				break;
-			}
-			nwrite -= amount;
-			pos    += amount;
-		}
-
-		rc = fsync_and_close(log_fd, "partition");
-		if (rc && !error_code)
-			error_code = rc;
-	}
-	if (error_code)
-		(void) unlink(new_file);
-	else {			/* file shuffle */
-		(void) unlink(old_file);
-		if (link(reg_file, old_file)) {
-			debug4("unable to create link for %s -> %s: %m",
-			       reg_file, old_file);
-		}
-		(void) unlink(reg_file);
-		if (link(new_file, reg_file)) {
-			debug4("unable to create link for %s -> %s: %m",
-			       new_file, reg_file);
-		}
-		(void) unlink(new_file);
-	}
-	xfree(old_file);
-	xfree(reg_file);
-	xfree(new_file);
-	unlock_state_files();
+	save_buf_to_state("part_state", buffer, &high_buffer_size);
 
 	FREE_NULL_BUFFER(buffer);
 	END_TIMER2(__func__);
@@ -492,36 +406,7 @@ static int _dump_part_state(void *x, void *arg)
 	else
 		part_ptr->flags &= (~PART_FLAG_DEFAULT);
 
-	pack32(part_ptr->cpu_bind,	 buffer);
-	packstr(part_ptr->name,          buffer);
-	pack32(part_ptr->grace_time,	 buffer);
-	pack32(part_ptr->max_time,       buffer);
-	pack32(part_ptr->default_time,   buffer);
-	pack32(part_ptr->max_cpus_per_node, buffer);
-	pack32(part_ptr->max_cpus_per_socket, buffer);
-	pack32(part_ptr->max_nodes_orig, buffer);
-	pack32(part_ptr->min_nodes_orig, buffer);
-
-	pack16(part_ptr->flags,          buffer);
-	pack16(part_ptr->max_share,      buffer);
-	pack16(part_ptr->over_time_limit,buffer);
-	pack16(part_ptr->preempt_mode,   buffer);
-	pack16(part_ptr->priority_job_factor, buffer);
-	pack16(part_ptr->priority_tier,  buffer);
-
-	pack16(part_ptr->state_up,       buffer);
-	pack16(part_ptr->cr_type,        buffer);
-
-	packstr(part_ptr->allow_accounts, buffer);
-	packstr(part_ptr->allow_groups,  buffer);
-	packstr(part_ptr->allow_qos,     buffer);
-	packstr(part_ptr->qos_char,      buffer);
-	packstr(part_ptr->allow_alloc_nodes, buffer);
-	packstr(part_ptr->alternate,     buffer);
-	packstr(part_ptr->deny_accounts, buffer);
-	packstr(part_ptr->deny_qos,      buffer);
-	/* Save orig_nodes as nodes will be built from orig_nodes */
-	packstr(part_ptr->orig_nodes, buffer);
+	part_record_pack(part_ptr, buffer, SLURM_PROTOCOL_VERSION);
 
 	return 0;
 }
@@ -558,24 +443,13 @@ static buf_t *_open_part_state_file(char **state_file)
  */
 extern int load_all_part_state(uint16_t reconfig_flags)
 {
-	char *part_name = NULL, *nodes = NULL;
-	char *allow_accounts = NULL, *allow_groups = NULL, *allow_qos = NULL;
-	char *deny_accounts = NULL, *deny_qos = NULL, *qos_char = NULL;
 	char *state_file = NULL;
-	uint32_t max_time, default_time, max_nodes, min_nodes;
-	uint32_t max_cpus_per_node = INFINITE, cpu_bind = 0, grace_time = 0;
-	uint32_t max_cpus_per_socket = INFINITE;
 	time_t time;
-	uint16_t flags, priority_job_factor, priority_tier;
-	uint16_t max_share, over_time_limit = NO_VAL16, preempt_mode;
-	uint16_t state_up, cr_type;
 	part_record_t *part_ptr;
 	int error_code = 0, part_cnt = 0;
 	buf_t *buffer;
 	char *ver_str = NULL;
-	char* allow_alloc_nodes = NULL;
 	uint16_t protocol_version = NO_VAL16;
-	char* alternate = NULL;
 
 	xassert(verify_lock(CONF_LOCK, READ_LOCK));
 
@@ -617,192 +491,111 @@ extern int load_all_part_state(uint16_t reconfig_flags)
 	safe_unpack_time(&time, buffer);
 
 	while (remaining_buf(buffer) > 0) {
-		if (protocol_version >= SLURM_23_02_PROTOCOL_VERSION) {
-			safe_unpack32(&cpu_bind, buffer);
-			safe_unpackstr(&part_name, buffer);
-			safe_unpack32(&grace_time, buffer);
-			safe_unpack32(&max_time, buffer);
-			safe_unpack32(&default_time, buffer);
-			safe_unpack32(&max_cpus_per_node, buffer);
-			safe_unpack32(&max_cpus_per_socket, buffer);
-			safe_unpack32(&max_nodes, buffer);
-			safe_unpack32(&min_nodes, buffer);
+		part_record_t *part_rec_state = NULL;
 
-			safe_unpack16(&flags,        buffer);
-			safe_unpack16(&max_share,    buffer);
-			safe_unpack16(&over_time_limit, buffer);
-			safe_unpack16(&preempt_mode, buffer);
-
-			safe_unpack16(&priority_job_factor, buffer);
-			safe_unpack16(&priority_tier, buffer);
-			if (priority_job_factor > part_max_priority)
-				part_max_priority = priority_job_factor;
-
-			safe_unpack16(&state_up, buffer);
-			safe_unpack16(&cr_type, buffer);
-
-			safe_unpackstr(&allow_accounts, buffer);
-			safe_unpackstr(&allow_groups, buffer);
-			safe_unpackstr(&allow_qos, buffer);
-			safe_unpackstr(&qos_char, buffer);
-			safe_unpackstr(&allow_alloc_nodes, buffer);
-			safe_unpackstr(&alternate, buffer);
-			safe_unpackstr(&deny_accounts, buffer);
-			safe_unpackstr(&deny_qos, buffer);
-			safe_unpackstr(&nodes, buffer);
-			if ((flags & PART_FLAG_DEFAULT_CLR)   ||
-			    (flags & PART_FLAG_EXC_USER_CLR)  ||
-			    (flags & PART_FLAG_HIDDEN_CLR)    ||
-			    (flags & PART_FLAG_NO_ROOT_CLR)   ||
-			    (flags & PART_FLAG_PDOI_CLR)      ||
-			    (flags & PART_FLAG_ROOT_ONLY_CLR) ||
-			    (flags & PART_FLAG_REQ_RESV_CLR)  ||
-			    (flags & PART_FLAG_LLN_CLR)) {
-				error("Invalid data for partition %s: flags=%u",
-				      part_name, flags);
-				error_code = EINVAL;
-			}
-		} else if (protocol_version >= SLURM_MIN_PROTOCOL_VERSION) {
-			safe_unpack32(&cpu_bind, buffer);
-			safe_unpackstr(&part_name, buffer);
-			safe_unpack32(&grace_time, buffer);
-			safe_unpack32(&max_time, buffer);
-			safe_unpack32(&default_time, buffer);
-			safe_unpack32(&max_cpus_per_node, buffer);
-			safe_unpack32(&max_nodes, buffer);
-			safe_unpack32(&min_nodes, buffer);
-
-			safe_unpack16(&flags,        buffer);
-			safe_unpack16(&max_share,    buffer);
-			safe_unpack16(&over_time_limit, buffer);
-			safe_unpack16(&preempt_mode, buffer);
-
-			safe_unpack16(&priority_job_factor, buffer);
-			safe_unpack16(&priority_tier, buffer);
-			if (priority_job_factor > part_max_priority)
-				part_max_priority = priority_job_factor;
-
-			safe_unpack16(&state_up, buffer);
-			safe_unpack16(&cr_type, buffer);
-
-			safe_unpackstr(&allow_accounts, buffer);
-			safe_unpackstr(&allow_groups, buffer);
-			safe_unpackstr(&allow_qos, buffer);
-			safe_unpackstr(&qos_char, buffer);
-			safe_unpackstr(&allow_alloc_nodes, buffer);
-			safe_unpackstr(&alternate, buffer);
-			safe_unpackstr(&deny_accounts, buffer);
-			safe_unpackstr(&deny_qos, buffer);
-			safe_unpackstr(&nodes, buffer);
-			if ((flags & PART_FLAG_DEFAULT_CLR)   ||
-			    (flags & PART_FLAG_EXC_USER_CLR)  ||
-			    (flags & PART_FLAG_HIDDEN_CLR)    ||
-			    (flags & PART_FLAG_NO_ROOT_CLR)   ||
-			    (flags & PART_FLAG_ROOT_ONLY_CLR) ||
-			    (flags & PART_FLAG_REQ_RESV_CLR)  ||
-			    (flags & PART_FLAG_LLN_CLR)) {
-				error("Invalid data for partition %s: flags=%u",
-				      part_name, flags);
-				error_code = EINVAL;
-			}
-		} else {
-			error("%s: protocol_version %hu not supported",
-			      __func__, protocol_version);
+		if ((error_code = part_record_unpack(&part_rec_state, buffer,
+						     protocol_version)))
 			goto unpack_error;
+
+		if ((part_rec_state->flags & PART_FLAG_DEFAULT_CLR)   ||
+		    (part_rec_state->flags & PART_FLAG_EXC_USER_CLR)  ||
+		    (part_rec_state->flags & PART_FLAG_EXC_TOPO_CLR)  ||
+		    (part_rec_state->flags & PART_FLAG_HIDDEN_CLR)    ||
+		    (part_rec_state->flags & PART_FLAG_NO_ROOT_CLR)   ||
+		    (part_rec_state->flags & PART_FLAG_PDOI_CLR)      ||
+		    (part_rec_state->flags & PART_FLAG_ROOT_ONLY_CLR) ||
+		    (part_rec_state->flags & PART_FLAG_REQ_RESV_CLR)  ||
+		    (part_rec_state->flags & PART_FLAG_LLN_CLR)) {
+			error("Invalid data for partition %s: flags=%u",
+			      part_rec_state->name, part_rec_state->flags);
+			error_code = EINVAL;
 		}
 		/* validity test as possible */
-		if (state_up > PARTITION_UP) {
+		if (part_rec_state->state_up > PARTITION_UP) {
 			error("Invalid data for partition %s: state_up=%u",
-			      part_name, state_up);
+			      part_rec_state->name, part_rec_state->state_up);
 			error_code = EINVAL;
 		}
 		if (error_code) {
 			error("No more partition data will be processed from "
 			      "the checkpoint file");
-			xfree(allow_accounts);
-			xfree(allow_groups);
-			xfree(allow_qos);
-			xfree(qos_char);
-			xfree(allow_alloc_nodes);
-			xfree(alternate);
-			xfree(deny_accounts);
-			xfree(deny_qos);
-			xfree(part_name);
-			xfree(nodes);
+			part_record_delete(part_rec_state);
 			error_code = EINVAL;
 			break;
 		}
 
 		/* find record and perform update */
 		part_ptr = list_find_first(part_list, &list_find_part,
-					   part_name);
+					   part_rec_state->name);
 		if (!part_ptr && (reconfig_flags & RECONFIG_KEEP_PART_INFO)) {
 			info("%s: partition %s missing from configuration file, creating",
-			     __func__, part_name);
-			part_ptr = create_part_record(part_name);
+			     __func__, part_rec_state->name);
+			part_ptr = create_ctld_part_record(part_rec_state->name);
 		} else if (!part_ptr) {
 			info("%s: partition %s removed from configuration file, skipping",
-			     __func__, part_name);
+			     __func__, part_rec_state->name);
 		}
 
 		/* Handle RECONFIG_KEEP_PART_STAT */
 		if (part_ptr) {
 			part_cnt++;
-			part_ptr->state_up = state_up;
+			part_ptr->state_up = part_rec_state->state_up;
 		}
 
 		if (!(reconfig_flags & RECONFIG_KEEP_PART_INFO)) {
-			xfree(allow_accounts);
-			xfree(allow_groups);
-			xfree(allow_qos);
-			xfree(qos_char);
-			xfree(allow_alloc_nodes);
-			xfree(alternate);
-			xfree(deny_accounts);
-			xfree(deny_qos);
-			xfree(part_name);
-			xfree(nodes);
+			part_record_delete(part_rec_state);
 			continue;
 		}
 
-		part_ptr->cpu_bind       = cpu_bind;
-		part_ptr->flags          = flags;
+		part_ptr->cpu_bind = part_rec_state->cpu_bind;
+		part_ptr->flags = part_rec_state->flags;
 		if (part_ptr->flags & PART_FLAG_DEFAULT) {
 			xfree(default_part_name);
-			default_part_name = xstrdup(part_name);
+			default_part_name = xstrdup(part_rec_state->name);
 			default_part_loc = part_ptr;
 		}
-		part_ptr->max_time       = max_time;
-		part_ptr->default_time   = default_time;
-		part_ptr->max_cpus_per_node = max_cpus_per_node;
-		part_ptr->max_cpus_per_socket = max_cpus_per_socket;
-		part_ptr->max_nodes      = max_nodes;
-		part_ptr->max_nodes_orig = max_nodes;
-		part_ptr->min_nodes      = min_nodes;
-		part_ptr->min_nodes_orig = min_nodes;
-		part_ptr->max_share      = max_share;
-		part_ptr->grace_time     = grace_time;
-		part_ptr->over_time_limit = over_time_limit;
-		if (preempt_mode != NO_VAL16)
-			part_ptr->preempt_mode   = preempt_mode;
-		part_ptr->priority_job_factor = priority_job_factor;
-		part_ptr->priority_tier  = priority_tier;
-		part_ptr->cr_type	 = cr_type;
+		part_ptr->max_time = part_rec_state->max_time;
+		part_ptr->default_time = part_rec_state->default_time;
+		part_ptr->max_cpus_per_node = part_rec_state->max_cpus_per_node;
+		part_ptr->max_cpus_per_socket =
+			part_rec_state->max_cpus_per_socket;
+		part_ptr->max_nodes = part_rec_state->max_nodes;
+		part_ptr->max_nodes_orig = part_rec_state->max_nodes;
+		part_ptr->min_nodes = part_rec_state->min_nodes;
+		part_ptr->min_nodes_orig = part_rec_state->min_nodes;
+		part_ptr->max_share = part_rec_state->max_share;
+		part_ptr->grace_time = part_rec_state->grace_time;
+		part_ptr->over_time_limit = part_rec_state->over_time_limit;
+		if (part_rec_state->preempt_mode != NO_VAL16)
+			part_ptr->preempt_mode = part_rec_state->preempt_mode;
+		part_ptr->priority_job_factor =
+			part_rec_state->priority_job_factor;
+		part_ptr->priority_tier = part_rec_state->priority_tier;
+		part_ptr->cr_type = part_rec_state->cr_type;
+
 		xfree(part_ptr->allow_accounts);
-		part_ptr->allow_accounts = allow_accounts;
-		xfree(part_ptr->allow_groups);
+		part_ptr->allow_accounts = part_rec_state->allow_accounts;
+		part_rec_state->allow_accounts = NULL;
+
 		FREE_NULL_LIST(part_ptr->allow_accts_list);
 		part_ptr->allow_accts_list =
 			accounts_list_build(part_ptr->allow_accounts, false);
-		part_ptr->allow_groups   = allow_groups;
-		xfree(part_ptr->allow_qos);
-		part_ptr->allow_qos      = allow_qos;
-		qos_list_build(part_ptr->allow_qos,&part_ptr->allow_qos_bitstr);
 
-		if (qos_char) {
+		xfree(part_ptr->allow_groups);
+		part_ptr->allow_groups   = part_rec_state->allow_groups;
+		part_rec_state->allow_groups = NULL;
+
+		xfree(part_ptr->allow_qos);
+		part_ptr->allow_qos = part_rec_state->allow_qos;
+		part_rec_state->allow_qos = NULL;
+		qos_list_build(part_ptr->allow_qos, false,
+			       &part_ptr->allow_qos_bitstr);
+
+		if (part_rec_state->qos_char) {
 			slurmdb_qos_rec_t qos_rec;
 			xfree(part_ptr->qos_char);
-			part_ptr->qos_char = qos_char;
+			part_ptr->qos_char = part_rec_state->qos_char;
+			part_rec_state->qos_char = NULL;
 
 			memset(&qos_rec, 0, sizeof(slurmdb_qos_rec_t));
 			qos_rec.name = part_ptr->qos_char;
@@ -818,17 +611,25 @@ extern int load_all_part_state(uint16_t reconfig_flags)
 		}
 
 		xfree(part_ptr->allow_alloc_nodes);
-		part_ptr->allow_alloc_nodes   = allow_alloc_nodes;
+		part_ptr->allow_alloc_nodes = part_rec_state->allow_alloc_nodes;
+		part_rec_state->allow_alloc_nodes = NULL;
+
 		xfree(part_ptr->alternate);
-		part_ptr->alternate      = alternate;
+		part_ptr->alternate = part_rec_state->alternate;
+		part_rec_state->alternate = NULL;
+
 		xfree(part_ptr->deny_accounts);
-		part_ptr->deny_accounts  = deny_accounts;
+		part_ptr->deny_accounts = part_rec_state->deny_accounts;
+		part_rec_state->deny_accounts = NULL;
 		FREE_NULL_LIST(part_ptr->deny_accts_list);
 		part_ptr->deny_accts_list =
 			accounts_list_build(part_ptr->deny_accounts, false);
+
 		xfree(part_ptr->deny_qos);
-		part_ptr->deny_qos       = deny_qos;
-		qos_list_build(part_ptr->deny_qos, &part_ptr->deny_qos_bitstr);
+		part_ptr->deny_qos = part_rec_state->deny_qos;
+		part_rec_state->deny_qos = NULL;
+		qos_list_build(part_ptr->deny_qos, false,
+			       &part_ptr->deny_qos_bitstr);
 
 		/*
 		 * Store saved nodelist in orig_nodes. nodes will be regenerated
@@ -836,9 +637,10 @@ extern int load_all_part_state(uint16_t reconfig_flags)
 		 */
 		xfree(part_ptr->nodes);
 		xfree(part_ptr->orig_nodes);
-		part_ptr->orig_nodes = nodes;
+		part_ptr->orig_nodes = part_rec_state->nodes;
+		part_rec_state->nodes = NULL;
 
-		xfree(part_name);
+		part_record_delete(part_rec_state);
 	}
 
 	info("Recovered state of %d partitions", part_cnt);
@@ -873,11 +675,11 @@ part_record_t *find_part_record(char *name)
  * IN part_list_src - a job's part_list
  * RET copy of part_list_src, must be freed by caller
  */
-extern List part_list_copy(List part_list_src)
+extern list_t *part_list_copy(list_t *part_list_src)
 {
 	part_record_t *part_ptr;
-	ListIterator iter;
-	List part_list_dest = NULL;
+	list_itr_t *iter;
+	list_t *part_list_dest = NULL;
 
 	if (!part_list_src)
 		return part_list_dest;
@@ -895,34 +697,36 @@ extern List part_list_copy(List part_list_src)
 /*
  * get_part_list - find record for named partition(s)
  * IN name - partition name(s) in a comma separated list
+ * OUT part_ptr_list - sorted list of pointers to the partitions or NULL
+ * OUT prim_part_ptr - pointer to the primary partition
  * OUT err_part - The first invalid partition name.
- * RET List of pointers to the partitions or NULL if not found
  * NOTE: Caller must free the returned list
  * NOTE: Caller must free err_part
  */
-extern List get_part_list(char *name, char **err_part)
+extern void get_part_list(char *name, list_t **part_ptr_list,
+			  part_record_t **prim_part_ptr, char **err_part)
 {
 	part_record_t *part_ptr;
-	List job_part_list = NULL;
 	char *token, *last = NULL, *tmp_name;
 
+	*part_ptr_list = NULL;
+	*prim_part_ptr = NULL;
+
 	if (name == NULL)
-		return job_part_list;
+		return;
 
 	tmp_name = xstrdup(name);
 	token = strtok_r(tmp_name, ",", &last);
 	while (token) {
 		part_ptr = list_find_first(part_list, &list_find_part, token);
 		if (part_ptr) {
-			if (job_part_list == NULL) {
-				job_part_list = list_create(NULL);
-			}
-			if (!list_find_first(job_part_list, &_match_part_ptr,
-					     part_ptr)) {
-				list_append(job_part_list, part_ptr);
-			}
+			if (!(*part_ptr_list))
+				*part_ptr_list = list_create(NULL);
+			if (!list_find_first(*part_ptr_list, &_match_part_ptr,
+					     part_ptr))
+				list_append(*part_ptr_list, part_ptr);
 		} else {
-			FREE_NULL_LIST(job_part_list);
+			FREE_NULL_LIST(*part_ptr_list);
 			if (err_part) {
 				xfree(*err_part);
 				*err_part = xstrdup(token);
@@ -931,8 +735,23 @@ extern List get_part_list(char *name, char **err_part)
 		}
 		token = strtok_r(NULL, ",", &last);
 	}
+
+	if (*part_ptr_list) {
+		/*
+		 * Return the first part_ptr in the list before sorting. On
+		 * state load, the first partition in the list is the running
+		 * partition -- for multi-partition jobs. Other times it doesn't
+		 * matter what the returned part_ptr is because it will be
+		 * modified when scheduling the different job_queue_rec_t's.
+		 *
+		 * The part_ptr_list always needs to be sorted by priority_tier.
+		 */
+		*prim_part_ptr = list_peek(*part_ptr_list);
+		list_sort(*part_ptr_list, priority_sort_part_tier);
+		if (list_count(*part_ptr_list) == 1)
+			FREE_NULL_LIST(*part_ptr_list);
+	}
 	xfree(tmp_name);
-	return job_part_list;
 }
 
 /*
@@ -953,26 +772,6 @@ void init_part_conf(void)
 	default_part_loc = NULL;
 }
 
-/*
- * Free memory for cached backfill data in partition record
- */
-static void _bf_data_free(bf_part_data_t **datap)
-{
-	bf_part_data_t *data;
-	if (!datap || !*datap)
-		return;
-
-	data = *datap;
-
-	slurmdb_destroy_bf_usage(data->job_usage);
-        slurmdb_destroy_bf_usage(data->resv_usage);
-	xhash_free(data->user_usage);
-	xfree(data);
-
-	*datap = NULL;
-	return;
-}
-
 /*
  * _list_delete_part - delete an entry from the global partition list,
  *	see common/list.h for documentation
@@ -1003,35 +802,9 @@ static void _list_delete_part(void *part_entry)
 		}
 	}
 
-	xfree(part_ptr->allow_accounts);
-	FREE_NULL_LIST(part_ptr->allow_accts_list);
-	xfree(part_ptr->allow_alloc_nodes);
-	xfree(part_ptr->allow_groups);
-	xfree(part_ptr->allow_uids);
-	xfree(part_ptr->allow_qos);
-	FREE_NULL_BITMAP(part_ptr->allow_qos_bitstr);
-	xfree(part_ptr->alternate);
-	xfree(part_ptr->billing_weights_str);
-	xfree(part_ptr->billing_weights);
-	xfree(part_ptr->deny_accounts);
-	FREE_NULL_LIST(part_ptr->deny_accts_list);
-	xfree(part_ptr->deny_qos);
-	FREE_NULL_BITMAP(part_ptr->deny_qos_bitstr);
-	FREE_NULL_LIST(part_ptr->job_defaults_list);
-	xfree(part_ptr->name);
-	xfree(part_ptr->orig_nodes);
-	xfree(part_ptr->nodes);
-	xfree(part_ptr->nodesets);
-	FREE_NULL_BITMAP(part_ptr->node_bitmap);
-	xfree(part_ptr->qos_char);
-	xfree(part_ptr->tres_cnt);
-	xfree(part_ptr->tres_fmt_str);
-	_bf_data_free(&part_ptr->bf_data);
-
-	xfree(part_entry);
+	part_record_delete(part_ptr);
 }
 
-
 /*
  * list_find_part - find an entry in the partition list, see common/list.h
  *	for documentation
@@ -1174,17 +947,14 @@ static int _pack_part(void *object, void *arg)
 /*
  * pack_all_part - dump all partition information for all partitions in
  *	machine independent form (for network transmission)
- * OUT buffer_ptr - the pointer is set to the allocated buffer.
- * OUT buffer_size - set to size of the buffer in bytes
  * IN show_flags - partition filtering options
  * IN uid - uid of user making request (for partition filtering)
  * global: part_list - global list of partition records
- * NOTE: the buffer at *buffer_ptr must be xfreed by the caller
+ * OUT buffer
  * NOTE: change slurm_load_part() in api/part_info.c if data format changes
  */
-extern void pack_all_part(char **buffer_ptr, int *buffer_size,
-			  uint16_t show_flags, uid_t uid,
-			  uint16_t protocol_version)
+extern buf_t *pack_all_part(uint16_t show_flags, uid_t uid,
+			    uint16_t protocol_version)
 {
 	int tmp_offset;
 	time_t now = time(NULL);
@@ -1199,9 +969,6 @@ extern void pack_all_part(char **buffer_ptr, int *buffer_size,
 		.visible_parts = build_visible_parts(uid, privileged),
 	};
 
-	buffer_ptr[0] = NULL;
-	*buffer_size = 0;
-
 	/* write header: version and time */
 	pack32(0, pack_info.buffer);
 	pack_time(now, pack_info.buffer);
@@ -1214,9 +981,8 @@ extern void pack_all_part(char **buffer_ptr, int *buffer_size,
 	pack32(pack_info.parts_packed, pack_info.buffer);
 	set_buf_offset(pack_info.buffer, tmp_offset);
 
-	*buffer_size = get_buf_offset(pack_info.buffer);
-	buffer_ptr[0] = xfer_buf_data(pack_info.buffer);
 	xfree(pack_info.visible_parts);
+	return pack_info.buffer;
 }
 
 
@@ -1232,7 +998,7 @@ extern void pack_all_part(char **buffer_ptr, int *buffer_size,
  */
 void pack_part(part_record_t *part_ptr, buf_t *buffer, uint16_t protocol_version)
 {
-	if (protocol_version >= SLURM_23_02_PROTOCOL_VERSION) {
+	if (protocol_version >= SLURM_24_05_PROTOCOL_VERSION) {
 		if (default_part_loc == part_ptr)
 			part_ptr->flags |= PART_FLAG_DEFAULT;
 		else
@@ -1252,7 +1018,7 @@ void pack_part(part_record_t *part_ptr, buf_t *buffer, uint16_t protocol_version
 		pack32(part_ptr->max_cpus_per_socket, buffer);
 		pack64(part_ptr->max_mem_per_cpu, buffer);
 
-		pack16(part_ptr->flags, buffer);
+		pack32(part_ptr->flags, buffer);
 		pack16(part_ptr->max_share, buffer);
 		pack16(part_ptr->over_time_limit, buffer);
 		pack16(part_ptr->preempt_mode, buffer);
@@ -1281,6 +1047,7 @@ void pack_part(part_record_t *part_ptr, buf_t *buffer, uint16_t protocol_version
 				      job_defaults_pack, buffer,
 				      protocol_version);
 	} else if (protocol_version >= SLURM_MIN_PROTOCOL_VERSION) {
+		uint16_t tmp_uint16;
 		if (default_part_loc == part_ptr)
 			part_ptr->flags |= PART_FLAG_DEFAULT;
 		else
@@ -1297,10 +1064,12 @@ void pack_part(part_record_t *part_ptr, buf_t *buffer, uint16_t protocol_version
 		pack32(part_ptr->total_cpus, buffer);
 		pack64(part_ptr->def_mem_per_cpu, buffer);
 		pack32(part_ptr->max_cpus_per_node, buffer);
+		pack32(part_ptr->max_cpus_per_socket, buffer);
 		pack64(part_ptr->max_mem_per_cpu, buffer);
 
-		pack16(part_ptr->flags,      buffer);
-		pack16(part_ptr->max_share,  buffer);
+		tmp_uint16 = part_ptr->flags;
+		pack16(tmp_uint16, buffer);
+		pack16(part_ptr->max_share, buffer);
 		pack16(part_ptr->over_time_limit, buffer);
 		pack16(part_ptr->preempt_mode, buffer);
 		pack16(part_ptr->priority_job_factor, buffer);
@@ -1396,7 +1165,7 @@ extern int update_part(update_part_msg_t * part_desc, bool create_flag)
 		}
 		info("%s: partition %s being created", __func__,
 		     part_desc->name);
-		part_ptr = create_part_record(part_desc->name);
+		part_ptr = create_ctld_part_record(part_desc->name);
 	} else {
 		if (!part_ptr) {
 			verbose("%s: Update for partition not found (%s)",
@@ -1532,6 +1301,16 @@ extern int update_part(update_part_msg_t * part_desc, bool create_flag)
 		part_ptr->flags &= (~PART_FLAG_EXCLUSIVE_USER);
 	}
 
+	if (part_desc->flags & PART_FLAG_EXCLUSIVE_TOPO) {
+		info("%s: setting exclusive_topo for partition %s", __func__,
+		     part_desc->name);
+		part_ptr->flags |= PART_FLAG_EXCLUSIVE_TOPO;
+	} else if (part_desc->flags & PART_FLAG_EXC_TOPO_CLR) {
+		info("%s: clearing exclusive_topo for partition %s", __func__,
+		     part_desc->name);
+		part_ptr->flags &= (~PART_FLAG_EXCLUSIVE_TOPO);
+	}
+
 	if (part_desc->flags & PART_FLAG_DEFAULT) {
 		if (default_part_name == NULL) {
 			info("%s: setting default partition to %s", __func__,
@@ -1593,22 +1372,45 @@ extern int update_part(update_part_msg_t * part_desc, bool create_flag)
 	}
 
 	if (part_desc->preempt_mode != NO_VAL16) {
-		uint16_t new_mode;
-		new_mode = part_desc->preempt_mode & (~PREEMPT_MODE_GANG);
-		if (new_mode <= PREEMPT_MODE_CANCEL) {
-			info("%s: setting preempt_mode to %s for partition %s",
-			     __func__, preempt_mode_string(new_mode),
-			     part_desc->name);
-			part_ptr->preempt_mode = new_mode;
+		if (!(part_desc->preempt_mode & PREEMPT_MODE_GANG)) {
+			uint16_t new_mode;
+
+			new_mode =
+				part_desc->preempt_mode & (~PREEMPT_MODE_GANG);
+
+			if (new_mode <= PREEMPT_MODE_CANCEL) {
+				/*
+				 * This is a valid mode, but if GANG was enabled
+				 * at cluster level, always leave it set.
+				 */
+				if ((part_ptr->preempt_mode != NO_VAL16) &&
+				    (part_ptr->preempt_mode &
+				     PREEMPT_MODE_GANG))
+					new_mode = new_mode | PREEMPT_MODE_GANG;
+				info("%s: setting preempt_mode to %s for partition %s",
+				     __func__,
+				     preempt_mode_string(new_mode),
+				     part_desc->name);
+				part_ptr->preempt_mode = new_mode;
+			} else {
+				info("%s: invalid preempt_mode %u", __func__, new_mode);
+			}
 		} else {
-			info("%s: invalid preempt_mode %u", __func__, new_mode);
+			info("%s: PreemptMode=GANG is a cluster-wide option and cannot be set at partition level",
+			      __func__);
 		}
 	}
 
 	if (part_desc->priority_tier != NO_VAL16) {
+		bool changed =
+			part_ptr->priority_tier != part_desc->priority_tier;
 		info("%s: setting PriorityTier to %u for partition %s",
 		     __func__, part_desc->priority_tier, part_desc->name);
 		part_ptr->priority_tier = part_desc->priority_tier;
+
+		/* Need to resort all job partition lists */
+		if (changed)
+			sort_all_jobs_partition_lists();
 	}
 
 	if (part_desc->priority_job_factor != NO_VAL16) {
@@ -1629,7 +1431,7 @@ extern int update_part(update_part_msg_t * part_desc, bool create_flag)
 		 * partitions. If not then just set this partition.
 		 */
 		if (redo_prio) {
-			ListIterator itr = list_iterator_create(part_list);
+			list_itr_t *itr = list_iterator_create(part_list);
 			part_record_t *part2 = NULL;
 
 			if (redo_prio == 2) {
@@ -1707,7 +1509,8 @@ extern int update_part(update_part_msg_t * part_desc, bool create_flag)
 			info("%s: setting AllowQOS to %s for partition %s",
 			     __func__, part_ptr->allow_qos, part_desc->name);
 		}
-		qos_list_build(part_ptr->allow_qos,&part_ptr->allow_qos_bitstr);
+		qos_list_build(part_ptr->allow_qos, false,
+			       &part_ptr->allow_qos_bitstr);
 	}
 
 	if (part_desc->qos_char && part_desc->qos_char[0] == '\0') {
@@ -1745,10 +1548,10 @@ extern int update_part(update_part_msg_t * part_desc, bool create_flag)
 		memset(&qos_rec, 0, sizeof(slurmdb_qos_rec_t));
 		qos_rec.name = part_desc->qos_char;
 		assoc_mgr_lock(&locks);
-		if (assoc_mgr_fill_in_qos(
+		if ((assoc_mgr_fill_in_qos(
 			    acct_db_conn, &qos_rec, accounting_enforce,
 			    (slurmdb_qos_rec_t **)&qos, true)
-		    != SLURM_SUCCESS) {
+		     != SLURM_SUCCESS) || !qos) {
 			error("%s: invalid qos (%s) given",
 			      __func__, qos_rec.name);
 			error_code = ESLURM_INVALID_QOS;
@@ -1873,7 +1676,8 @@ extern int update_part(update_part_msg_t * part_desc, bool create_flag)
 		part_desc->deny_qos = NULL;
 		info("%s: setting DenyQOS to %s for partition %s", __func__,
 		     part_ptr->deny_qos, part_desc->name);
-		qos_list_build(part_ptr->deny_qos, &part_ptr->deny_qos_bitstr);
+		qos_list_build(part_ptr->deny_qos, false,
+			       &part_ptr->deny_qos_bitstr);
 	}
 	if (part_desc->allow_qos && part_desc->deny_qos) {
 		error("%s: Both AllowQOS and DenyQOS are defined, DenyQOS will be ignored",
@@ -1896,7 +1700,7 @@ extern int update_part(update_part_msg_t * part_desc, bool create_flag)
 	}
 
 	if (part_desc->job_defaults_str) {
-		List new_job_def_list = NULL;
+		list_t *new_job_def_list = NULL;
 		if (part_desc->job_defaults_str[0] == '\0') {
 			FREE_NULL_LIST(part_ptr->job_defaults_list);
 		} else if (job_defaults_list(part_desc->job_defaults_str,
@@ -2193,7 +1997,7 @@ static int _find_acct_in_list(void *x, void *arg)
  *	if required (updated group file or force set)
  * IN force - if set then always reload the allow_uid list
  */
-void load_part_uid_allow_list(int force)
+void load_part_uid_allow_list(bool force)
 {
 	static time_t last_update_time;
 	int updated = 0;
@@ -2202,7 +2006,7 @@ void load_part_uid_allow_list(int force)
 
 	START_TIMER;
 	temp_time = get_group_tlm();
-	if ((force == 0) && (temp_time == last_update_time))
+	if (!force && (temp_time == last_update_time))
 		return;
 	debug("Updating partition uid access list");
 	last_update_time = temp_time;
@@ -2253,40 +2057,6 @@ extern int delete_partition(delete_part_msg_t *part_desc_ptr)
 	return SLURM_SUCCESS;
 }
 
-/*
- * Determine of the specified job can execute right now or is currently
- * blocked by a miscellaneous limit. This does not re-validate job state,
- * but relies upon schedule() in src/slurmctld/job_scheduler.c to do so.
- */
-extern bool misc_policy_job_runnable_state(job_record_t *job_ptr)
-{
-	if ((job_ptr->state_reason == FAIL_ACCOUNT) ||
-	    (job_ptr->state_reason == FAIL_QOS) ||
-	    (job_ptr->state_reason == WAIT_NODE_NOT_AVAIL)) {
-		return false;
-	}
-
-	return true;
-}
-
-/*
- * Determine of the specified job can execute right now or is currently
- * blocked by a partition state or limit. These job states should match the
- * reason values returned by job_limits_check().
- */
-extern bool part_policy_job_runnable_state(job_record_t *job_ptr)
-{
-	if ((job_ptr->state_reason == WAIT_PART_DOWN) ||
-	    (job_ptr->state_reason == WAIT_PART_INACTIVE) ||
-	    (job_ptr->state_reason == WAIT_PART_NODE_LIMIT) ||
-	    (job_ptr->state_reason == WAIT_PART_TIME_LIMIT) ||
-	    (job_ptr->state_reason == WAIT_QOS_THRES)) {
-		return false;
-	}
-
-	return true;
-}
-
 /*
  * Validate a job's account against the partition's AllowAccounts or
  *	DenyAccounts parameters.
@@ -2443,7 +2213,6 @@ extern void part_list_update_assoc_lists(void)
 	list_for_each(part_list, part_update_assoc_lists, NULL);
 	assoc_mgr_unlock(&locks);
 	unlock_slurmctld(part_write_lock);
-	return;
 }
 
 extern int part_update_assoc_lists(void *x, void *arg)
@@ -2461,3 +2230,30 @@ extern int part_update_assoc_lists(void *x, void *arg)
 
 	return 0;
 }
+
+typedef struct {
+	char *names;
+	char *pos;
+} _foreach_part_names_t;
+
+static int _foreach_part_name_to_xstr(void *x, void *arg)
+{
+	part_record_t *part_ptr = x;
+	_foreach_part_names_t *part_names = arg;
+
+	xstrfmtcatat(part_names->names, &part_names->pos, "%s%s",
+		     part_names->names ? "," : "", part_ptr->name);
+
+	return SLURM_SUCCESS;
+}
+
+extern char *part_list_to_xstr(list_t *list)
+{
+	_foreach_part_names_t part_names = {0};
+
+	xassert(list);
+
+	list_for_each(list, _foreach_part_name_to_xstr, &part_names);
+
+	return part_names.names;
+}
