@@ -60,6 +60,7 @@
 #define PING_TIMEOUT 100
 
 static pthread_mutex_t lock_mutex = PTHREAD_MUTEX_INITIALIZER;
+static bool ping_updated = false;
 static int ping_count = 0;
 static time_t ping_start = 0;
 
@@ -83,8 +84,17 @@ bool is_ping_done (void)
 			      PING_TIMEOUT);
 			ping_msg_sent = true;
 		}
-	} else
+	} else {
 		ping_msg_sent = false;
+		/*
+		 * We can only consider the last node ping cycle to be fully
+		 * completed if ping_updated is true, meaning the
+		 * _agent_nodes_update thread finished updating the nodes
+		 * response. Otherwise, we could hit a race and incorrectly set
+		 * responding/healthy nodes to DOWN.
+		 */
+		is_done = ping_updated;
+	}
 	slurm_mutex_unlock(&lock_mutex);
 
 	return is_done;
@@ -104,6 +114,18 @@ void ping_begin (void)
 	slurm_mutex_unlock(&lock_mutex);
 }
 
+/*
+ * ping_nodes_update - A ping cycle can end but the update can still be pending
+ * for the _agent_nodes_update thread. This call will confirm node info was
+ * updated.
+ */
+void ping_nodes_update(void)
+{
+	slurm_mutex_lock(&lock_mutex);
+	ping_updated = true;
+	slurm_mutex_unlock(&lock_mutex);
+}
+
 /*
  * ping_end - record that a ping cycle has ended. This can be called more
  *	than once (for REQUEST_PING and simultaneous REQUEST_NODE_REGISTRATION
@@ -121,7 +143,7 @@ void ping_end (void)
 
 	if (ping_count == 0) /* no more running ping cycles */
 		ping_start = 0;
-
+	ping_updated = false;
 	slurm_mutex_unlock(&lock_mutex);
 }
 
@@ -382,17 +404,42 @@ void ping_nodes (void)
 /* Spawn health check function for every node that is not DOWN */
 extern void run_health_check(void)
 {
+	char *host_str = NULL;
+	agent_arg_t *check_agent_args = NULL;
 #ifdef HAVE_FRONT_END
 	front_end_record_t *front_end_ptr;
 	int i;
 #else
 	node_record_t *node_ptr;
-	int node_test_cnt = 0, node_limit, node_states, run_cyclic;
+	int node_test_cnt = 0;
+	int node_limit = 0;
+	int node_states = slurm_conf.health_check_node_state &
+		(~HEALTH_CHECK_CYCLE);
+	int run_cyclic = slurm_conf.health_check_node_state &
+		HEALTH_CHECK_CYCLE;
 	static int base_node_loc = 0;
 	static time_t cycle_start_time = (time_t) 0;
+
+	if (run_cyclic) {
+		time_t now = time(NULL);
+		if (cycle_start_time == (time_t) 0)
+			cycle_start_time = now;
+		else if (base_node_loc > 0)
+			;	/* mid-cycle */
+		else if (difftime(now, cycle_start_time) <
+		         slurm_conf.health_check_interval)
+			return;	/* Wait to start next cycle */
+		cycle_start_time = now;
+		/*
+		 * Determine how many nodes we want to test on each call of
+		 * run_health_check() to spread out the work.
+		 */
+		node_limit = (active_node_record_count * 2) /
+		             slurm_conf.health_check_interval;
+		node_limit = MAX(node_limit, 10);
+	}
+
 #endif
-	char *host_str = NULL;
-	agent_arg_t *check_agent_args = NULL;
 
 	check_agent_args = xmalloc (sizeof (agent_arg_t));
 	check_agent_args->msg_type = REQUEST_HEALTH_CHECK;
@@ -422,30 +469,6 @@ extern void run_health_check(void)
 		check_agent_args->node_count++;
 	}
 #else
-	node_limit = 0;
-	run_cyclic = slurm_conf.health_check_node_state &
-	             HEALTH_CHECK_CYCLE;
-	node_states = slurm_conf.health_check_node_state &
-	              (~HEALTH_CHECK_CYCLE);
-	if (run_cyclic) {
-		time_t now = time(NULL);
-		if (cycle_start_time == (time_t) 0)
-			cycle_start_time = now;
-		else if (base_node_loc > 0)
-			;	/* mid-cycle */
-		else if (difftime(now, cycle_start_time) <
-		         slurm_conf.health_check_interval) {
-			hostlist_destroy(check_agent_args->hostlist);
-			xfree(check_agent_args);
-			return;	/* Wait to start next cycle */
-		}
-		cycle_start_time = now;
-		/* Determine how many nodes we want to test on each call of
-		 * run_health_check() to spread out the work. */
-		node_limit = (active_node_record_count * 2) /
-		             slurm_conf.health_check_interval;
-		node_limit = MAX(node_limit, 10);
-	}
 
 	for (; (node_ptr = next_node(&base_node_loc)); base_node_loc++) {
 		if (run_cyclic &&
