@@ -134,7 +134,7 @@
 \*****************************************************************************/
 
 typedef struct slurmctld_config {
-	List acct_update_list;
+	list_t *acct_update_list;
 	pthread_cond_t acct_update_cond;
 	pthread_mutex_t acct_update_lock;
 	pthread_cond_t backup_finish_cond;
@@ -153,9 +153,7 @@ typedef struct slurmctld_config {
 	pthread_t thread_id_acct_update;
 	pthread_t thread_id_main;
 	pthread_t thread_id_save;
-	pthread_t thread_id_sig;
 	pthread_t thread_id_purge_files;
-	pthread_t thread_id_rpc;
 } slurmctld_config_t;
 
 typedef enum {
@@ -256,10 +254,8 @@ extern bool cloud_dns;
 extern uint32_t   cluster_cpus;
 extern bool disable_remote_singleton;
 extern int listen_nports;
-extern struct pollfd *listen_fds;
 extern int max_depend_depth;
 extern uint32_t max_powered_nodes;
-extern bool node_features_updated;
 extern pthread_cond_t purge_thread_cond;
 extern pthread_mutex_t purge_thread_lock;
 extern pthread_mutex_t check_bf_running_lock;
@@ -281,15 +277,7 @@ extern bool ping_nodes_now;		/* if set, ping nodes immediately */
 extern bool want_nodes_reboot;		/* if set, check for idle nodes */
 extern bool ignore_state_errors;
 
-typedef struct node_features {
-	uint32_t magic;		/* magic cookie to test data integrity */
-	char *name;		/* name of a feature */
-	bitstr_t *node_bitmap;	/* bitmap of nodes with this feature */
-} node_feature_t;
-
-extern List active_feature_list;/* list of currently active node features */
-extern List avail_feature_list;	/* list of available node features */
-extern List conf_includes_list; /* list of conf_includes_map_t */
+extern list_t *conf_includes_list;  /* list of conf_includes_map_t */
 
 #define PACK_FANOUT_ADDRS(_X) \
 	(IS_NODE_DYNAMIC_FUTURE(_X) || \
@@ -298,7 +286,7 @@ extern List conf_includes_list; /* list of conf_includes_map_t */
 
 /*****************************************************************************\
  *  NODE states and bitmaps
- *
+ *  asap_node_bitmap 	    Set if the node is marked to be rebooted asap
  *  avail_node_bitmap       Set if node's state is not DOWN, DRAINING/DRAINED,
  *                          FAILING or NO_RESPOND (i.e. available to run a job)
  *  booting_node_bitmap     Set if node in process of booting
@@ -314,6 +302,7 @@ extern List conf_includes_list; /* list of conf_includes_map_t */
  *                          configured for the job's partition)
  *  up_node_bitmap          Set if the node's state is not DOWN
 \*****************************************************************************/
+extern bitstr_t *asap_node_bitmap; /* reboot asap nodes */
 extern bitstr_t *avail_node_bitmap;	/* bitmap of available nodes,
 					 * state not DOWN, DRAIN or FAILING */
 extern bitstr_t *bf_ignore_node_bitmap;	/* bitmap of nodes made available during
@@ -332,7 +321,7 @@ extern bitstr_t *rs_node_bitmap;	/* next_state=resume nodes */
 /*****************************************************************************\
  *  PARTITION parameters and data structures
 \*****************************************************************************/
-extern List part_list;			/* list of part_record entries */
+extern list_t *part_list;		/* list of part_record entries */
 extern time_t last_part_update;		/* time of last part_list update */
 extern part_record_t default_part;	/* default configuration values */
 extern char *default_part_name;		/* name of default partition */
@@ -373,12 +362,12 @@ typedef struct slurmctld_resv {
 				 * running on it */
 	char *features;		/* required node features		*/
 	uint64_t flags;		/* see RESERVE_FLAG_* in slurm.h	*/
-	List gres_list_alloc;	/* Allocated generic resource allocation
+	list_t *gres_list_alloc;/* Allocated generic resource allocation
 				 * detail */
 	char *groups;		/* names of linux groups permitted to use */
 	uint32_t job_pend_cnt;	/* number of pending jobs		*/
 	uint32_t job_run_cnt;	/* number of running jobs		*/
-	List license_list;	/* structure with license info		*/
+	list_t *license_list;	/* structure with license info		*/
 	char *licenses;		/* required system licenses (including those
 				 * from TRES requests */
 	uint32_t max_start_delay;/* Maximum delay in which jobs outside of the
@@ -417,13 +406,13 @@ typedef struct {
 	bitstr_t **exc_cores;
 } resv_exc_t;
 
-extern List resv_list;		/* list of slurmctld_resv_t entries */
+extern list_t *resv_list;	/* list of slurmctld_resv_t entries */
 extern time_t last_resv_update;	/* time of last resv_list update */
 
 /*****************************************************************************\
  *  Job lists
 \*****************************************************************************/
-extern List job_list;			/* list of job_record entries */
+extern list_t *job_list;		/* list of job_record entries */
 extern list_t *purge_jobs_list;		/* list of job_record_t to free */
 
 /*****************************************************************************\
@@ -437,15 +426,9 @@ extern list_t *purge_jobs_list;		/* list of job_record_t to free */
  * useful when updating other types of consumable resources as well
 */
 enum select_plugindata_info {
-	SELECT_CR_PLUGIN,    /* data-> uint32 See SELECT_TYPE_* below */
-	SELECT_BITMAP,       /* Unused since version 2.0 */
-	SELECT_ALLOC_CPUS,   /* data-> uint16 alloc cpus (CR support) */
-	SELECT_ALLOC_LPS,    /* data-> uint32 alloc lps  (CR support) */
-	SELECT_AVAIL_MEMORY, /* data-> uint64 avail mem  (CR support) */
-	SELECT_STATIC_PART,  /* data-> uint16, 1 if static partitioning
-			      * BlueGene support */
-	SELECT_CONFIG_INFO,  /* data-> List get .conf info from select
-			      * plugin */
+	SELECT_CR_PLUGIN = 0,    /* data-> uint32 See SELECT_TYPE_* below */
+	SELECT_CONFIG_INFO = 6,  /* data-> list_t * get .conf info from select
+				  * plugin */
 };
 #define SELECT_TYPE_CONS_RES	1
 #define SELECT_TYPE_CONS_TRES	2
@@ -510,6 +493,9 @@ extern void set_job_tres_req_str(job_record_t *job_ptr, bool assoc_mgr_locked);
  * This function can be called multiple times. */
 extern void backup_slurmctld_restart(void);
 
+/* Handle SIGHUP while in backup mode */
+extern void backup_on_sighup(void);
+
 /* Complete a batch job requeue logic after all steps complete so that
  * subsequent jobs appear in a separate accounting record. */
 extern void batch_requeue_fini(job_record_t *job_ptr);
@@ -573,7 +559,7 @@ extern void delete_step_record(job_record_t *job_ptr, step_record_t *step_ptr);
  * IN depend_list_src - a job's depend_lst
  * RET copy of depend_list_src, must bee freed by caller
  */
-extern List depended_list_copy(List depend_list_src);
+extern list_t *depended_list_copy(list_t *depend_list_src);
 
 /*
  * drain_nodes - drain one or more nodes,
@@ -659,7 +645,7 @@ extern void node_mgr_reset_node_stats(node_record_t *node_ptr);
  * IN feature_list_src - a job's depend_lst
  * RET copy of depend_list_src, must be freed by caller
  */
-extern List feature_list_copy(List feature_list_src);
+extern list_t *feature_list_copy(list_t *feature_list_src);
 
 typedef enum {
 	FOR_EACH_JOB_BY_ID_EACH_INVALID = 0,
@@ -687,6 +673,8 @@ typedef foreach_job_by_id_control_t (*JobROForEachFunc)(
 					void *arg);
 /*
  * Function prototype for operating on a job id that is not found
+ * This is called just once for an array expression with a bitmap of array
+ * tasks that were not found.
  * Returns control requested for processing
  */
 typedef foreach_job_by_id_control_t
@@ -802,12 +790,14 @@ extern uint32_t get_next_job_id(bool test_only);
 /*
  * get_part_list - find record for named partition(s)
  * IN name - partition name(s) in a comma separated list
+ * OUT part_ptr_list - sorted list of pointers to the partitions or NULL
+ * OUT prim_part_ptr - pointer to the primary partition
  * OUT err_part - The first invalid partition name.
- * RET List of pointers to the partitions or NULL if not found
  * NOTE: Caller must free the returned list
  * NOTE: Caller must free err_part
  */
-extern List get_part_list(char *name, char **err_part);
+extern void get_part_list(char *name, list_t **part_ptr_list,
+			  part_record_t **prim_part_ptr, char **err_part);
 
 /*
  * init_depend_policy()
@@ -936,14 +926,15 @@ extern int job_allocate(job_desc_msg_t *job_desc, int immediate,
 extern void job_array_pre_sched(job_record_t *job_ptr);
 
 /* If this is a job array meta-job, clean up after scheduling attempt */
-extern job_record_t *job_array_post_sched(job_record_t *job_ptr);
+extern job_record_t *job_array_post_sched(job_record_t *job_ptr, bool list_add);
 
 /* Create an exact copy of an existing job record for a job array.
  * IN job_ptr - META job record for a job array, which is to become an
  *		individial task of the job array.
  *		Set the job's array_task_id to the task to be split out.
+ * IN list_add - add to the job_list or not.
  * RET - The new job record, which is the new META job record. */
-extern job_record_t *job_array_split(job_record_t *job_ptr);
+extern job_record_t *job_array_split(job_record_t *job_ptr, bool list_add);
 
 /* Record the start of one job array task */
 extern void job_array_start(job_record_t *job_ptr);
@@ -986,7 +977,7 @@ extern void job_completion_logger(job_record_t *job_ptr, bool requeue);
 extern uint64_t job_get_tres_mem(struct job_resources *job_res,
 				 uint64_t pn_min_memory, uint32_t cpu_cnt,
 				 uint32_t node_cnt, part_record_t *part_ptr,
-				 List gres_list, bool user_set_mem,
+				 list_t *gres_list, bool user_set_mem,
 				 uint16_t min_sockets_per_node,
 				 uint32_t num_tasks);
 
@@ -1105,20 +1096,16 @@ extern int job_mgr_signal_jobs(kill_jobs_msg_t *kill_msg, uid_t auth_uid,
  *
  * IN sus_ptr - suspend/resume request message
  * IN uid - user id of the user issuing the RPC
- * IN conn_fd - file descriptor on which to send reply,
- *              -1 if none
  * indf_susp IN - set if job is being suspended indefinitely by user or admin
  *                and we should clear it's priority, otherwise suspended
  *		  temporarily for gang scheduling
  * IN protocol_version - slurm protocol version of client
  * RET 0 on success, otherwise ESLURM error code
  */
-extern int job_suspend(suspend_msg_t *sus_ptr, uid_t uid,
-		       int conn_fd, bool indf_susp,
-		       uint16_t protocol_version);
-extern int job_suspend2(suspend_msg_t *sus_ptr, uid_t uid,
-			int conn_fd, bool indf_susp,
-			uint16_t protocol_version);
+extern int job_suspend(slurm_msg_t *msg, suspend_msg_t *sus_ptr, uid_t uid,
+		       bool indf_susp, uint16_t protocol_version);
+extern int job_suspend2(slurm_msg_t *msg, suspend_msg_t *sus_ptr, uid_t uid,
+			bool indf_susp, uint16_t protocol_version);
 
 /*
  * job_complete - note the normal termination the specified job
@@ -1182,14 +1169,13 @@ extern int job_requeue2(uid_t uid, requeue_msg_t *req_ptr, slurm_msg_t *msg,
  * job_set_top - Move the specified job to the top of the queue (at least
  *	for that user ID, partition, account, and QOS).
  *
+ * IN msg - original request msg
  * IN top_ptr - user request
  * IN uid - user id of the user issuing the RPC
- * IN conn_fd - file descriptor on which to send reply,
- *              -1 if none
  * IN protocol_version - slurm protocol version of client
  * RET 0 on success, otherwise ESLURM error code
  */
-extern int job_set_top(top_job_msg_t *top_ptr, uid_t uid, int conn_fd,
+extern int job_set_top(slurm_msg_t *msg, top_job_msg_t *top_ptr, uid_t uid,
 		       uint16_t protocol_version);
 
 /*
@@ -1313,11 +1299,6 @@ extern void load_part_uid_allow_list(bool force);
  */
 extern int load_all_part_state(uint16_t reconfig_flags);
 
-/*
- * Log contents of avail_feature_list and active_feature_list
- */
-extern void log_feature_lists(void);
-
 /* make_node_alloc - flag specified node as allocated to a job
  * IN node_ptr - pointer to node being allocated
  * IN job_ptr  - pointer to job that is starting
@@ -1534,7 +1515,7 @@ extern void part_fini (void);
  * IN part_list_src - a job's part_list
  * RET copy of part_list_src, must be freed by caller
  */
-extern List part_list_copy(List part_list_src);
+extern list_t *part_list_copy(list_t *part_list_src);
 
 /*
  * Validate a job's account against the partition's AllowAccounts or
@@ -1627,7 +1608,7 @@ void purge_old_job(void);
 extern void free_old_jobs(void);
 
 /* Convert a comma delimited list of QOS names into a bitmap */
-extern void qos_list_build(char *qos, bitstr_t **qos_bits);
+extern void qos_list_build(char *qos, bool locked, bitstr_t **qos_bits);
 
 /* Request that the job scheduler execute soon (typically within seconds) */
 extern void queue_job_scheduler(void);
@@ -1644,17 +1625,6 @@ extern void rehash_jobs(void);
  */
 extern void setup_job_state_hash(int new_hash_table_size);
 
-#ifndef NDEBUG
-/*
- * Walk entire job state cache and verify every job matchs job_ptr states
- * WARNING: slow and expensive to run
- * WARNING: caller must hold atleast job read lock
- */
-extern void verify_job_state_cache_synced(void);
-#else
-#define verify_job_state_cache_synced() {}
-#endif
-
 /* update first assigned job id as needed on reconfigure */
 extern void reset_first_job_id(void);
 
@@ -1681,6 +1651,11 @@ extern void restore_node_features(int recover);
  *	mode, assuming control when the primary controller stops responding */
 extern void run_backup(void);
 
+/* conmgr RPC connection callbacks */
+extern void *on_backup_connection(conmgr_fd_t *con, void *arg);
+extern void on_backup_finish(conmgr_fd_t *con, void *arg);
+extern int on_backup_msg(conmgr_fd_t *con, slurm_msg_t *msg, void *arg);
+
 /*
  * ping_controllers - ping other controllers in HA configuration.
  * IN active_controller - true if active controller, false if backup
@@ -1978,6 +1953,18 @@ extern int set_partition_billing_weights(char *billing_weights_str,
  */
 extern int update_part (update_part_msg_t * part_desc, bool create_flag);
 
+/*
+ * Sort all jobs' part_ptr_list to be in descending order according to
+ * partition priority tier. This Should be called anytime a partition's priority
+ * tier is modified.
+ */
+extern void sort_all_jobs_partition_lists();
+
+/*
+ * Common code to handle a job when a cred can't be created.
+ */
+extern void job_mgr_handle_cred_failure(job_record_t *job_ptr);
+
 /*
  * validate_alloc_node - validate that the allocating node
  * is allowed to use this partition
@@ -2248,7 +2235,7 @@ extern char **job_common_env_vars(job_record_t *job_ptr, bool is_complete);
 
 /*
  * update_node_active_features - Update active features associated with nodes
- * IN node_names - List of nodes to update
+ * IN node_names - list of nodes to update
  * IN active_features - New active features value
  * IN mode - FEATURE_MODE_IND : Print each node change indivually
  *           FEATURE_MODE_COMB: Try to combine like changes (SEE NOTE BELOW)
@@ -2263,7 +2250,7 @@ extern int update_node_active_features(char *node_names, char *active_features,
 /*
  * update_node_avail_features - Update available features associated with
  *	nodes, build new config list records as needed
- * IN node_names - List of nodes to update
+ * IN node_names - list of nodes to update
  * IN avail_features - New available features value
  * IN mode - FEATURE_MODE_IND : Print each node change indivually
  *           FEATURE_MODE_COMB: Try to combine like changes (SEE NOTE BELOW)
@@ -2360,4 +2347,10 @@ extern void controller_reconfig_scheduling(void);
  */
 extern char *part_list_to_xstr(list_t *list);
 
+/* Allow listener sockets to accept() new incoming requests */
+extern void listeners_unquiesce(void);
+
+/* Stop listener sockets from accept()ing new incoming requests */
+extern void listeners_quiesce(void);
+
 #endif /* !_HAVE_SLURMCTLD_H */
