@@ -1,8 +1,7 @@
 /*****************************************************************************\
  *  slurmscriptd.c - Slurm script functions.
  *****************************************************************************
- *  Copyright (C) 2021 SchedMD LLC.
- *  Written by Marshall Garey <marshall@schedmd.com>
+ *  Copyright (C) SchedMD LLC.
  *
  *  This file is part of Slurm, a resource management program.
  *  For details, see <https://slurm.schedmd.com/>.
@@ -42,6 +41,7 @@
 #  include <sys/prctl.h>
 #endif
 
+#include <fcntl.h>
 #include <poll.h>
 #include <signal.h>
 #include <sys/wait.h>
@@ -54,6 +54,7 @@
 #include "src/common/fd.h"
 #include "src/common/fetch_config.h"
 #include "src/common/log.h"
+#include "src/common/msg_type.h"
 #include "src/common/run_command.h"
 #include "src/common/setproctitle.h"
 #include "src/common/slurm_protocol_pack.h"
@@ -61,8 +62,10 @@
 #include "src/common/xmalloc.h"
 #include "src/common/xstring.h"
 
+#include "src/conmgr/conmgr.h"
+
 #include "src/interfaces/burst_buffer.h"
-#include "src/interfaces/select.h"
+#include "src/interfaces/hash.h"
 
 #include "src/slurmctld/locks.h"
 #include "src/slurmctld/slurmctld.h"
@@ -84,8 +87,8 @@
  */
 
 static bool _msg_readable(eio_obj_t *obj);
-static int _msg_accept(eio_obj_t *obj, List objs);
-static int _handle_close(eio_obj_t *obj, List objs);
+static int _msg_accept(eio_obj_t *obj, list_t *objs);
+static int _handle_close(eio_obj_t *obj, list_t *objs);
 
 struct io_operations msg_ops = {
 	.readable = _msg_readable,
@@ -206,7 +209,7 @@ static void _wait_for_script_resp(script_response_t *script_resp,
 				  int *status, char **resp_msg,
 				  bool *track_script_signalled)
 {
-	slurm_mutex_lock(&script_resp->mutex);
+	/* script_resp->mutex should already be locked */
 	slurm_cond_wait(&script_resp->cond, &script_resp->mutex);
 	/* The script is done now, and we should have the response */
 	*status = script_resp->rc;
@@ -214,7 +217,6 @@ static void _wait_for_script_resp(script_response_t *script_resp,
 		*resp_msg = xstrdup(script_resp->resp_msg);
 	if (track_script_signalled)
 		*track_script_signalled = script_resp->track_script_signalled;
-	slurm_mutex_unlock(&script_resp->mutex);
 }
 
 static void _wait_for_powersave_scripts()
@@ -286,7 +288,7 @@ static void _wait_for_powersave_scripts()
 
 }
 
-static int _handle_close(eio_obj_t *obj, List objs)
+static int _handle_close(eio_obj_t *obj, list_t *objs)
 {
 	debug3("Called %s", __func__);
 
@@ -323,11 +325,12 @@ static bool _msg_readable(eio_obj_t *obj)
 	return true;
 }
 
-static int _write_msg(int fd, int req, buf_t *buffer)
+static int _write_msg(int fd, int req, buf_t *buffer, bool lock)
 {
 	int len = 0;
 
-	slurm_mutex_lock(&write_mutex);
+	if (lock)
+		slurm_mutex_lock(&write_mutex);
 	safe_write(fd, &req, sizeof(int));
 	if (buffer) {
 		len = get_buf_offset(buffer);
@@ -335,7 +338,8 @@ static int _write_msg(int fd, int req, buf_t *buffer)
 		safe_write(fd, get_buf_data(buffer), len);
 	} else /* Write 0 length so the receiver knows not to read anymore */
 		safe_write(fd, &len, sizeof(int));
-	slurm_mutex_unlock(&write_mutex);
+	if (lock)
+		slurm_mutex_unlock(&write_mutex);
 
 	return SLURM_SUCCESS;
 
@@ -343,7 +347,8 @@ rwfail:
 	if (running_in_slurmctld())
 		error("%s: read/write op failed, restart slurmctld now: %m",
 		      __func__);
-	slurm_mutex_unlock(&write_mutex);
+	if (lock)
+		slurm_mutex_unlock(&write_mutex);
 	return SLURM_ERROR;
 }
 
@@ -384,10 +389,16 @@ static int _send_to_slurmscriptd(uint32_t msg_type, void *msg_data, bool wait,
 	}
 	if (msg_type == SLURMSCRIPTD_REQUEST_RUN_SCRIPT)
 		_incr_script_cnt();
-	rc = _write_msg(slurmctld_writefd, msg.msg_type, buffer);
+
+	if (wait)
+		slurm_mutex_lock(&script_resp->mutex);
+	rc = _write_msg(slurmctld_writefd, msg.msg_type, buffer, true);
 
 	if ((rc == SLURM_SUCCESS) && wait) {
 		_wait_for_script_resp(script_resp, &rc, resp_msg, signalled);
+	}
+	if (wait) {
+		slurm_mutex_unlock(&script_resp->mutex);
 		_script_resp_map_remove(script_resp->key);
 	}
 
@@ -443,7 +454,7 @@ static int _respond_to_slurmctld(char *key, uint32_t job_id, char *resp_msg,
 		rc = SLURM_ERROR;
 		goto cleanup;
 	}
-	_write_msg(slurmscriptd_writefd, msg.msg_type, buffer);
+	_write_msg(slurmscriptd_writefd, msg.msg_type, buffer, true);
 
 cleanup:
 	FREE_NULL_BUFFER(buffer);
@@ -466,12 +477,120 @@ static void _incr_script_cnt(void)
 	slurm_mutex_unlock(&script_count_mutex);
 }
 
+static void _change_proc_name(int argc, char **argv, char *proc_name)
+{
+	char *log_prefix;
+	/*
+	 * Since running_in_slurmctld() is called before we fork()'d,
+	 * the result is cached in static variables, so calling it now
+	 * would return true even though we're now slurmscriptd.
+	 * Reset those cached variables so running_in_slurmctld()
+	 * returns false if called from slurmscriptd.
+	 * But first change slurm_prog_name since that is
+	 * read by run_in_daemon().
+	 */
+	xfree(slurm_prog_name);
+	slurm_prog_name = xstrdup(proc_name);
+	running_in_slurmctld_reset();
+
+	/*
+	 * Change the process name to slurmscriptd.
+	 * Since slurmscriptd logs to the slurmctld log file, add a
+	 * prefix to make it clear which daemon a log comes from.
+	 */
+	init_setproctitle(argc, argv);
+	setproctitle("%s", proc_name);
+#if HAVE_SYS_PRCTL_H
+	if (prctl(PR_SET_NAME, proc_name, NULL, NULL, NULL) < 0) {
+		error("%s: cannot set my name to %s %m",
+		      __func__, proc_name);
+	}
+#endif
+	/* log_set_prefix takes control of an xmalloc()'d string */
+	log_prefix = xstrdup_printf("%s: ", proc_name);
+	log_set_prefix(&log_prefix);
+}
+
+static void _send_bb_script_msg(int write_fd, void *cb_arg)
+{
+	run_script_msg_t *script_msg = cb_arg;
+	buf_t *buffer = init_buf(0);
+	bb_script_info_msg_t bb_msg = {
+		.cluster_name = slurm_conf.cluster_name,
+		.extra_buf = script_msg->extra_buf,
+		.extra_buf_size = script_msg->extra_buf_size,
+		.function = script_msg->script_name,
+		.job_id = script_msg->job_id,
+		.slurmctld_debug = slurm_conf.slurmctld_debug,
+		.slurmctld_logfile = slurm_conf.slurmctld_logfile,
+		.log_fmt = slurm_conf.log_fmt,
+		.plugindir = slurm_conf.plugindir,
+		.slurm_user_name = slurm_conf.slurm_user_name,
+		.slurm_user_id = slurm_conf.slurm_user_id,
+	};
+	slurmscriptd_msg_t msg = {
+		.msg_data = &bb_msg,
+		.msg_type = SLURMSCRIPTD_REQUEST_BB_SCRIPT_INFO,
+	};
+
+	/*
+	 * Send bb_script_info_msg_t. The write_mutex controls writing on
+	 * slurmctld_writefd or slurmscriptd_writefd. We are writing to a pipe
+	 * to a running script, so we do not need to lock write_mutex.
+	 */
+	slurmscriptd_pack_msg(&msg, buffer);
+	if (_write_msg(write_fd, SLURMSCRIPTD_REQUEST_BB_SCRIPT_INFO, buffer,
+		       false) != SLURM_SUCCESS) {
+		error("%s: Failed writing data to script: burst_buffer.lua:%s, JobId=%u",
+		      __func__, script_msg->script_name, script_msg->job_id);
+		goto fini;
+	}
+
+fini:
+	FREE_NULL_BUFFER(buffer);
+}
+
+static int _recv_bb_script_msg(bb_script_info_msg_t **msg_pptr)
+{
+	int rc = SLURM_SUCCESS, req = 0, len = 0;
+	int fd = STDIN_FILENO;
+	char *data = NULL;
+	buf_t *buffer = NULL;
+	slurmscriptd_msg_t scriptd_msg = { 0 };
+
+	safe_read(fd, &req, sizeof(req));
+	scriptd_msg.msg_type = req;
+	if (scriptd_msg.msg_type != SLURMSCRIPTD_REQUEST_BB_SCRIPT_INFO) {
+		fatal("%s: Invalid msg_type=%u",
+		      __func__, scriptd_msg.msg_type);
+	}
+
+	safe_read(fd, &len, sizeof(len));
+	if (!len)
+		fatal("%s: Invalid message length == 0", __func__);
+
+	data = xmalloc(len);
+	safe_read(fd, data, len);
+
+	/* Unpack bb_script_info_msg_t */
+	buffer = create_buf(data, len);
+	rc = slurmscriptd_unpack_msg(&scriptd_msg, buffer);
+	*msg_pptr = scriptd_msg.msg_data;
+	FREE_NULL_BUFFER(buffer);
+
+	return rc;
+
+rwfail:
+	error("%s Failed", __func__);
+	return SLURM_ERROR;
+}
+
 /*
  * Run a script with a given timeout (in seconds).
  * Return the status or SLURM_ERROR if fork() fails.
  */
-static int _run_script(run_command_args_t *run_command_args, uint32_t job_id,
-		       int timeout, char *tmp_file_env_name, char *tmp_file_str,
+static int _run_script(run_command_args_t *run_command_args,
+		       run_script_msg_t *script_msg,
 		       char **resp_msg, bool *signalled)
 {
 	int status = SLURM_ERROR;
@@ -479,6 +598,10 @@ static int _run_script(run_command_args_t *run_command_args, uint32_t job_id,
 	char *resp = NULL;
 	bool killed = false;
 	int tmp_fd = 0;
+	uint32_t job_id = script_msg->job_id;
+	int timeout = script_msg->timeout;
+	char *tmp_file_env_name = script_msg->tmp_file_env_name;
+	char *tmp_file_str = script_msg->tmp_file_str;
 
 	if ((timeout <= 0) || (timeout == NO_VAL16))
 		ms_timeout = -1; /* wait indefinitely in run_command() */
@@ -587,142 +710,6 @@ static int _handle_flush_job(slurmscriptd_msg_t *recv_msg)
 	return SLURM_SUCCESS;
 }
 
-static int _handle_reconfig(slurmscriptd_msg_t *recv_msg)
-{
-	slurmctld_lock_t config_write_lock =
-		{ .conf = WRITE_LOCK };
-	reconfig_msg_t *reconfig_msg = recv_msg->msg_data;
-
-	log_flag(SCRIPT, "Handling %s", rpc_num2string(recv_msg->msg_type));
-
-	lock_slurmctld(config_write_lock);
-	slurm_conf.debug_flags = reconfig_msg->debug_flags;
-	xfree(slurm_conf.slurmctld_logfile);
-	slurm_conf.slurmctld_logfile = xstrdup(reconfig_msg->logfile);
-	slurm_conf.log_fmt = reconfig_msg->log_fmt;
-	slurm_conf.slurmctld_debug = reconfig_msg->slurmctld_debug;
-	slurm_conf.slurmctld_syslog_debug = reconfig_msg->syslog_debug;
-	update_logging();
-	unlock_slurmctld(config_write_lock);
-
-	return SLURM_SUCCESS;
-}
-
-static void _run_bb_script_child(int fd, char *script_func, uint32_t job_id,
-				 uint32_t argc, char **argv,
-				 job_info_msg_t *job_info)
-{
-	int exit_code;
-	char *resp = NULL;
-
-	setpgid(0, 0);
-
-	exit_code = bb_g_run_script(script_func, job_id, argc, argv, job_info,
-				    &resp);
-	if (resp)
-		safe_write(fd, resp, strlen(resp));
-
-rwfail:
-	_exit(exit_code);
-}
-
-
-/*
- * Run the burst buffer script in a fork()'d process so that if the script
- * runs for longer than the timeout, or if the script is cancelled, we can
- * SIGTERM/SIGKILL the process. This is based on the code in run_command(),
- * but instead of calling exec() in the child, we call a burst buffer plugin
- * API to run the script.
- *
- * Set the response of the script in resp_msg.
- * Return the exit code of the script.
- */
-static int _run_bb_script(run_script_msg_t *script_msg,
-			  char **resp_msg,
-			  bool *track_script_signalled)
-{
-	int pfd[2] = {-1, -1};
-	int status = 0;
-	uint32_t job_id = script_msg->job_id, timeout = script_msg->timeout;
-	uint32_t argc = script_msg->argc;
-	char **argv = script_msg->argv;
-	char *script_func = script_msg->script_name;
-	pid_t cpid;
-	job_info_msg_t *job_info = NULL;
-
-	xassert(resp_msg);
-	xassert(track_script_signalled);
-
-	*track_script_signalled = false;
-
-	if (script_msg->extra_buf_size) {
-		buf_t *extra_buf;
-		slurm_msg_t *extra_msg = xmalloc(sizeof *extra_msg);
-
-		slurm_msg_t_init(extra_msg);
-		extra_msg->protocol_version = SLURM_PROTOCOL_VERSION;
-		extra_msg->msg_type = RESPONSE_JOB_INFO;
-		extra_buf = create_buf(script_msg->extra_buf,
-				       script_msg->extra_buf_size);
-		unpack_msg(extra_msg, extra_buf);
-		job_info = extra_msg->data;
-		extra_msg->data = NULL;
-
-		/*
-		 * create_buf() does not duplicate the data, just points to it.
-		 * So just NULL it out here. It will get free'd later.
-		 */
-		extra_buf->head = NULL;
-		FREE_NULL_BUFFER(extra_buf);
-		slurm_free_msg(extra_msg);
-	}
-
-	if (pipe(pfd) != 0) {
-		*resp_msg = xstrdup_printf("pipe(): %m");
-		error("%s: Error running %s for JobId=%u: %s",
-		      __func__, script_func, job_id, *resp_msg);
-		return 127;
-	}
-
-
-	cpid = fork();
-	if (cpid < 0) { /* fork() failed */
-		*resp_msg = xstrdup_printf("fork(): %m");
-		error("%s: Error running %s for JobId=%u: %s",
-		      __func__, script_func, job_id, *resp_msg);
-		close(pfd[0]);
-		close(pfd[1]);
-		return 127;
-	} else if (cpid == 0) { /* child - run the script */
-		close(pfd[0]); /* Close the read fd, we're only writing */
-		_run_bb_script_child(pfd[1], script_func, job_id, argc, argv,
-				     job_info);
-	} else { /* parent */
-		close(pfd[1]); /* Close the write fd, we're only reading */
-		track_script_rec_add(job_id, cpid, pthread_self());
-		*resp_msg = run_command_poll_child(cpid,
-						   timeout * 1000,
-						   false,
-						   pfd[0],
-						   script_msg->script_path,
-						   script_msg->script_name,
-						   pthread_self(),
-						   &status,
-						   NULL);
-		close(pfd[0]);
-
-		/* If we were killed by track_script, let the caller know. */
-		*track_script_signalled =
-			track_script_killed(pthread_self(), status, true);
-
-		track_script_remove(pthread_self());
-	}
-
-	slurm_free_job_info_msg(job_info);
-
-	return status;
-}
-
 static int _handle_shutdown(slurmscriptd_msg_t *recv_msg)
 {
 	log_flag(SCRIPT, "Handling %s", rpc_num2string(recv_msg->msg_type));
@@ -730,6 +717,7 @@ static int _handle_shutdown(slurmscriptd_msg_t *recv_msg)
 	_wait_for_powersave_scripts();
 	track_script_flush();
 
+	conmgr_request_shutdown();
 	eio_signal_shutdown(msg_handle);
 
 	return SLURM_ERROR; /* Don't handle any more requests. */
@@ -737,6 +725,7 @@ static int _handle_shutdown(slurmscriptd_msg_t *recv_msg)
 
 static int _handle_run_script(slurmscriptd_msg_t *recv_msg)
 {
+	extern char **environ;
 	run_script_msg_t *script_msg = recv_msg->msg_data;
 	int rc, status = 0;
 	char *resp_msg = NULL;
@@ -764,9 +753,26 @@ static int _handle_run_script(slurmscriptd_msg_t *recv_msg)
 
 	switch (script_msg->script_type) {
 	case SLURMSCRIPTD_BB_LUA:
-		status = _run_bb_script(script_msg,
-					&resp_msg,
-					&signalled);
+		/*
+		 * Set SLURM_SCRIPT_CONTEXT in env for slurmctld, but we also
+		 * need to preserve the parent's environment. There was not any
+		 * env passed to us in script_msg.
+		 */
+		xassert(!run_command_args.env);
+		run_command_args.env = env_array_copy((const char **) environ);
+		env_array_append(&run_command_args.env, "SLURM_SCRIPT_CONTEXT",
+				 "burst_buffer.lua");
+
+		/* burst_buffer.lua is not exec'd, it is run directly from C */
+		run_command_args.ignore_path_exec_check = true;
+
+		/* Send needed script info and configs to slurmctld */
+		run_command_args.write_to_child = true;
+		run_command_args.cb = _send_bb_script_msg;
+		run_command_args.cb_arg = script_msg;
+
+		status = _run_script(&run_command_args, script_msg,
+				     &resp_msg, &signalled);
 		break;
 	case SLURMSCRIPTD_EPILOG: /* fall-through */
 	case SLURMSCRIPTD_MAIL:
@@ -779,10 +785,7 @@ static int _handle_run_script(slurmscriptd_msg_t *recv_msg)
 		 * script_msg->timeout may also not be set (NO_VAL16).
 		 * Let _run_script handle the conversion.
 		 */
-		status = _run_script(&run_command_args, script_msg->job_id,
-				     script_msg->timeout,
-				     script_msg->tmp_file_env_name,
-				     script_msg->tmp_file_str,
+		status = _run_script(&run_command_args, script_msg,
 				     &resp_msg, &signalled);
 		break;
 	case SLURMSCRIPTD_POWER:
@@ -797,10 +800,7 @@ static int _handle_run_script(slurmscriptd_msg_t *recv_msg)
 		 */
 		run_command_args.tid = 0;
 		run_command_args.orphan_on_shutdown = true;
-		status = _run_script(&run_command_args, script_msg->job_id,
-				     script_msg->timeout,
-				     script_msg->tmp_file_env_name,
-				     script_msg->tmp_file_str,
+		status = _run_script(&run_command_args, script_msg,
 				     &resp_msg, &signalled);
 
 		break;
@@ -982,9 +982,6 @@ static int _handle_request(int req, buf_t *buffer)
 		case SLURMSCRIPTD_REQUEST_FLUSH_JOB:
 			rc = _handle_flush_job(&recv_msg);
 			break;
-		case SLURMSCRIPTD_REQUEST_RECONFIG:
-			rc = _handle_reconfig(&recv_msg);
-			break;
 		case SLURMSCRIPTD_REQUEST_RUN_SCRIPT:
 			rc = _handle_run_script(&recv_msg);
 			break;
@@ -1023,7 +1020,7 @@ static void *_handle_accept(void *args)
 	return NULL;
 }
 
-static int _msg_accept(eio_obj_t *obj, List objs)
+static int _msg_accept(eio_obj_t *obj, list_t *objs)
 {
 	int rc = SLURM_SUCCESS, req, buf_len = 0;
 	char *incoming_buffer = NULL;
@@ -1082,14 +1079,158 @@ static void _setup_eio(int fd)
 	eio_new_initial_obj(msg_handle, eio_obj);
 }
 
-static void _slurmscriptd_mainloop(void)
+static void _on_sigint(conmgr_callback_args_t conmgr_args, void *arg)
 {
-	run_command_init();
+	log_flag(SCRIPT, "Caught SIGINT. Ignoring.");
+}
+
+static void _on_sigterm(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	log_flag(SCRIPT, "Caught SIGTERM. Ignoring.");
+}
+
+static void _on_sigchld(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	log_flag(SCRIPT, "Caught SIGCHLD. Ignoring");
+}
+
+static void _on_sigquit(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	log_flag(SCRIPT, "Caught SIGQUIT. Ignoring.");
+}
+
+static void _on_sigtstp(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	log_flag(SCRIPT, "Caught SIGTSTP. Ignoring.");
+}
+
+static void _on_sighup(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	log_flag(SCRIPT, "Caught SIGHUP. Ignoring.");
+}
+
+static void _on_sigusr1(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	log_flag(SCRIPT, "Caught SIGUSR1. Ignoring.");
+}
+
+static void _on_sigusr2(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	log_flag(SCRIPT, "Caught SIGUSR2. Ignoring.");
+}
+
+static void _on_sigpipe(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	/* debug5 to avoid polluting the SCRIPT debug flag */
+	debug5("Caught SIGPIPE. Ignoring.");
+}
+
+static void _on_sigxcpu(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	log_flag(SCRIPT, "Caught SIGXCPU. Ignoring.");
+}
+
+static void _on_sigabrt(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	log_flag(SCRIPT, "Caught SIGABRT. Ignoring.");
+}
+
+static void _on_sigalrm(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	log_flag(SCRIPT, "Caught SIGALRM. Ignoring.");
+}
+
+static void _init_slurmscriptd_conmgr(void)
+{
+	conmgr_callbacks_t callbacks = { NULL, NULL };
+
+	if (slurm_conf.slurmctld_params)
+		conmgr_set_params(slurm_conf.slurmctld_params);
+
+	conmgr_init(0, 0, callbacks);
+
+	/*
+	 * Ignore signals. slurmscriptd should only handle requests directly
+	 * from slurmctld.
+	 */
+	conmgr_add_work_signal(SIGINT, _on_sigint, NULL);
+	conmgr_add_work_signal(SIGTERM, _on_sigterm, NULL);
+	conmgr_add_work_signal(SIGCHLD, _on_sigchld, NULL);
+	conmgr_add_work_signal(SIGQUIT, _on_sigquit, NULL);
+	conmgr_add_work_signal(SIGTSTP, _on_sigtstp, NULL);
+	conmgr_add_work_signal(SIGHUP, _on_sighup, NULL);
+	conmgr_add_work_signal(SIGUSR1, _on_sigusr1, NULL);
+	conmgr_add_work_signal(SIGUSR2, _on_sigusr2, NULL);
+	conmgr_add_work_signal(SIGPIPE, _on_sigpipe, NULL);
+	conmgr_add_work_signal(SIGXCPU, _on_sigxcpu, NULL);
+	conmgr_add_work_signal(SIGABRT, _on_sigabrt, NULL);
+	conmgr_add_work_signal(SIGALRM, _on_sigalrm, NULL);
+
+	conmgr_run(false);
+}
+
+__attribute__((noreturn))
+extern void slurmscriptd_run_slurmscriptd(int argc, char **argv,
+					  char *binary_path)
+{
+	ssize_t i;
+	int rc = SLURM_ERROR, ack;
+
+	slurmscriptd_writefd = SLURMSCRIPT_WRITE_FD;
+	slurmscriptd_readfd = SLURMSCRIPT_READ_FD;
+
+	_change_proc_name(argc, argv, "slurmscriptd");
+
+	/* Test communications with slurmctld. */
+	ack = SLURM_SUCCESS;
+	i = write(slurmscriptd_writefd, &ack, sizeof(int));
+	if (i != sizeof(int)) {
+		error("%s: slurmscriptd: failed to send return code to slurmctld: %m",
+		      __func__);
+		_exit(1);
+	}
+	i = read(slurmscriptd_readfd, &rc, sizeof(int));
+	if (i < 0) {
+		error("%s: slurmscriptd: Can not read ack from slurmctld: %m",
+		      __func__);
+		_exit(1);
+	} else if (i != sizeof(int)) {
+		error("%s: slurmscriptd: slurmctld failed to send ack: %m",
+		      __func__);
+		_exit(1);
+	}
+
+	debug("slurmscriptd: Got ack from slurmctld");
+
+	i = write(slurmscriptd_writefd, &ack, sizeof(int));
+	if (i != sizeof(int))
+		fatal("%s: Failed to send initialization code to slurmctld",
+		      __func__);
+
+	_init_slurmscriptd_conmgr();
+
+	debug("Initialization successful");
+
+	slurm_mutex_init(&powersave_script_count_mutex);
+	slurm_mutex_init(&write_mutex);
+	if ((run_command_init(0, NULL, binary_path) != SLURM_SUCCESS) &&
+	    binary_path && binary_path[0])
+		fatal("%s: Unable to reliably execute %s",
+		      __func__, binary_path);
+
 	_setup_eio(slurmscriptd_readfd);
 
 	debug("%s: started", __func__);
 	eio_handle_mainloop(msg_handle);
 	debug("%s: finished", __func__);
+
+#ifdef MEMORY_LEAK_DEBUG
+	track_script_fini();
+	slurm_mutex_destroy(&powersave_script_count_mutex);
+#endif
+
+	/* We never want to return from here, only exit. */
+	_exit(0);
 }
 
 static void *_slurmctld_listener_thread(void *x)
@@ -1197,6 +1338,61 @@ static run_script_msg_t *_init_run_script_msg(char **env,
 	return run_script_msg;
 }
 
+static job_info_msg_t *_unpack_bb_job_info(bb_script_info_msg_t *bb_msg)
+{
+	buf_t *extra_buf;
+	slurm_msg_t *extra_msg;
+	job_info_msg_t *job_info;
+
+	if (!bb_msg->extra_buf_size)
+		return NULL;
+
+	extra_msg = xmalloc(sizeof *extra_msg);
+	slurm_msg_t_init(extra_msg);
+	extra_msg->protocol_version = SLURM_PROTOCOL_VERSION;
+	extra_msg->msg_type = RESPONSE_JOB_INFO;
+	extra_buf = create_buf(bb_msg->extra_buf, bb_msg->extra_buf_size);
+	unpack_msg(extra_msg, extra_buf);
+	job_info = extra_msg->data;
+	extra_msg->data = NULL;
+
+	/* create_buf() does not duplicate the data, it just points to it. */
+	extra_buf->head = NULL;
+	FREE_NULL_BUFFER(extra_buf);
+	slurm_free_msg(extra_msg);
+
+	return job_info;
+}
+
+static void _init_bb_script_config(char **function, uint32_t *job_id,
+				   job_info_msg_t **job_info)
+{
+	bb_script_info_msg_t *bb_msg = NULL;
+
+	if (_recv_bb_script_msg(&bb_msg) != SLURM_SUCCESS)
+		fatal("Failed to receive burst buffer script msg");
+	if (!bb_msg->function || !bb_msg->function[0])
+		fatal_abort("%s: Invalid NULL function", __func__);
+
+	*function = bb_msg->function;
+	*job_id = bb_msg->job_id;
+	*job_info = _unpack_bb_job_info(bb_msg);
+
+	slurm_conf.cluster_name = bb_msg->cluster_name;
+	slurm_conf.slurmctld_debug = bb_msg->slurmctld_debug;
+	slurm_conf.slurmctld_logfile = bb_msg->slurmctld_logfile;
+	slurm_conf.log_fmt = bb_msg->log_fmt;
+	slurm_conf.plugindir = bb_msg->plugindir;
+	slurm_conf.slurm_user_name = bb_msg->slurm_user_name;
+	slurm_conf.slurm_user_id = bb_msg->slurm_user_id;
+
+	/*
+	 * We copied the pointers in bb_msg, so only free bb_msg, not its
+	 * contents.
+	 */
+	xfree(bb_msg);
+}
+
 extern void slurmscriptd_flush(void)
 {
 	_send_to_slurmscriptd(SLURMSCRIPTD_REQUEST_FLUSH, NULL, true, NULL,
@@ -1205,6 +1401,79 @@ extern void slurmscriptd_flush(void)
 	_wait_for_all_scripts();
 }
 
+extern void slurmscriptd_handle_bb_lua_mode(int argc, char **argv)
+{
+	int exit_code = 127;
+	char *function = NULL;
+	uint32_t job_id = 0;
+	char *proc_name = "burst_buffer.lua";
+	char *resp = NULL;
+	char **script_argv = NULL;
+	int script_argc = 0;
+	/* The lock is required for update_logging() */
+	slurmctld_lock_t config_write_lock = {
+		.conf = WRITE_LOCK,
+	};
+	/*
+	 * Only log errors until we read the config and update to the configured
+	 * level.
+	 */
+	log_options_t log_opts = LOG_OPTS_STDERR_ONLY;
+	job_info_msg_t *job_info = NULL;
+
+	setpgid(0, 0);
+	closeall(3); /* Do this before initializing logging */
+	/* coverity[leaked_handle] */
+
+	/* Logging will go to stdout/stderr until we call update_logging(). */
+	log_init(proc_name, log_opts, LOG_DAEMON, NULL);
+
+	/*
+	 * Change our process name and make it so running_in_slurmctld() and
+	 * run_in_daemon() return false.
+	 */
+	if (argc < RUN_COMMAND_LAUNCHER_ARGC) {
+		fatal("%s: Unexpected argc=%d, it should be >= %d",
+		      __func__, argc, RUN_COMMAND_LAUNCHER_ARGC);
+	}
+
+	script_argc = argc - RUN_COMMAND_LAUNCHER_ARGC;
+	/* _change_proc_name overwrites argv. Copy the args that we need. */
+	script_argv = slurm_char_array_copy(script_argc,
+					    &argv[RUN_COMMAND_LAUNCHER_ARGC]);
+	_change_proc_name(argc, argv, proc_name);
+
+	/* Minimal config setup: */
+	init_slurm_conf(&slurm_conf);
+	_init_bb_script_config(&function, &job_id, &job_info);
+
+	/*
+	 * Initialize plugins.
+	 */
+	slurm_conf.bb_type = "burst_buffer/lua";
+	if (bb_g_init() != SLURM_SUCCESS)
+		fatal("failed to initialize burst_buffer plugin");
+
+	/*
+	 * update_logging() makes logs go to the slurmctld log file.
+	 * Call update_logging() after initializing plugins to avoid seeing some
+	 * extra debug logs about loading plugins.
+	 */
+	lock_slurmctld(config_write_lock);
+	update_logging();
+	unlock_slurmctld(config_write_lock);
+
+	/* Run the script */
+	exit_code = bb_g_run_script(function, job_id, script_argc, script_argv,
+				    job_info, &resp);
+	if (resp)
+		safe_write(STDOUT_FILENO, resp, strlen(resp));
+
+	/* Ignore memory leaks because we are calling exit() */
+rwfail:
+	exit(exit_code);
+}
+
 extern void slurmscriptd_flush_job(uint32_t job_id)
 {
 	flush_job_msg_t *msg = xmalloc(sizeof(*msg));
@@ -1217,36 +1486,6 @@ extern void slurmscriptd_flush_job(uint32_t job_id)
 	slurm_thread_create_detached(_async_send_to_slurmscriptd, send_args);
 }
 
-extern void slurmscriptd_reconfig(void)
-{
-	reconfig_msg_t msg;
-	slurmctld_lock_t config_read_lock =
-		{ .conf = READ_LOCK };
-
-	memset(&msg, 0, sizeof(msg));
-
-	/*
-	 * slurmscriptd only needs a minimal configuration, so only send what
-	 * needs to be updated rather than sending the entire slurm_conf
-	 * or having slurmscriptd read/parse the slurm.conf file.
-	 */
-	lock_slurmctld(config_read_lock);
-	msg.debug_flags = slurm_conf.debug_flags;
-	msg.logfile = slurm_conf.slurmctld_logfile;
-	msg.log_fmt = slurm_conf.log_fmt;
-	msg.slurmctld_debug = slurm_conf.slurmctld_debug;
-	msg.syslog_debug = slurm_conf.slurmctld_syslog_debug;
-	/*
-	 * If we ever allow switching plugins on reconfig, then we will need to
-	 * pass slurm_conf.bb_type to slurmscriptd, since a child/fork() of
-	 * slurmscriptd calls bb_g_run_script().
-	 */
-	unlock_slurmctld(config_read_lock);
-
-	_send_to_slurmscriptd(SLURMSCRIPTD_REQUEST_RECONFIG, &msg, false,
-			      NULL, NULL);
-}
-
 extern int slurmscriptd_run_mail(char *script_path, uint32_t argc, char **argv,
 				 char **env, uint32_t timeout, char **resp)
 {
@@ -1312,10 +1551,11 @@ extern void slurmscriptd_run_power(char *script_path, char *hosts,
 
 extern int slurmscriptd_run_bb_lua(uint32_t job_id, char *function,
 				   uint32_t argc, char **argv, uint32_t timeout,
-				   char *job_buf, int job_buf_size,
-				   char **resp, bool *track_script_signalled)
+				   buf_t *job_buf, char **resp,
+				   bool *track_script_signalled)
 {
-	int status, rc;
+	int status, rc = SLURM_ERROR;
+	uint32_t extra_buf_size = job_buf ? job_buf->processed : 0;
 	run_script_msg_t run_script_msg;
 
 	memset(&run_script_msg, 0, sizeof(run_script_msg));
@@ -1323,12 +1563,11 @@ extern int slurmscriptd_run_bb_lua(uint32_t job_id, char *function,
 	/* Init run_script_msg */
 	run_script_msg.argc = argc;
 	run_script_msg.argv = argv;
-	run_script_msg.env = NULL;
-	run_script_msg.extra_buf = job_buf;
-	run_script_msg.extra_buf_size = job_buf_size;
+	run_script_msg.extra_buf = job_buf ? job_buf->head : NULL;
+	run_script_msg.extra_buf_size = extra_buf_size;
 	run_script_msg.job_id = job_id;
 	run_script_msg.script_name = function; /* Shallow copy, do not free */
-	run_script_msg.script_path = NULL;
+	run_script_msg.script_path = "burst_buffer.lua";
 	run_script_msg.script_type = SLURMSCRIPTD_BB_LUA;
 	run_script_msg.timeout = timeout;
 
@@ -1337,7 +1576,6 @@ extern int slurmscriptd_run_bb_lua(uint32_t job_id, char *function,
 				       &run_script_msg, true, resp,
 				       track_script_signalled);
 
-	/* Cleanup */
 	if (WIFEXITED(status))
 		rc = WEXITSTATUS(status);
 	else
@@ -1444,7 +1682,7 @@ extern void slurmscriptd_update_log_level(int debug_level, bool log_rotate)
 			      false, NULL, NULL);
 }
 
-extern int slurmscriptd_init(int argc, char **argv)
+extern int slurmscriptd_init(char **argv, char *binary_path)
 {
 	int to_slurmscriptd[2] = {-1, -1};
 	int to_slurmctld[2] = {-1, -1};
@@ -1470,14 +1708,21 @@ extern int slurmscriptd_init(int argc, char **argv)
 		 * slurmctld writes data to slurmscriptd with the
 		 * to_slurmscriptd pipe and slurmscriptd writes data to
 		 * slurmctld with the to_slurmctld pipe.
+		 * If there is a failure with startup, SIGKILL the slurmscriptd
+		 * and then exit. The slurmscriptd pid will be adopted and then
+		 * reaped by init, so we don't need to call waitpid().
 		 */
 		if (close(to_slurmscriptd[0]) < 0) {
-			_kill_slurmscriptd();
+			rc = errno;
+			killpg(slurmscriptd_pid, SIGKILL);
+			errno = rc;
 			fatal("%s: slurmctld: Unable to close read to_slurmscriptd in parent: %m",
 			      __func__);
 		}
 		if (close(to_slurmctld[1]) < 0) {
-			_kill_slurmscriptd();
+			rc = errno;
+			killpg(slurmscriptd_pid, SIGKILL);
+			errno = rc;
 			fatal("%s: slurmctld: Unable to close write to_slurmctld in parent: %m",
 			      __func__);
 		}
@@ -1485,25 +1730,31 @@ extern int slurmscriptd_init(int argc, char **argv)
 		/* Test communications with slurmscriptd. */
 		i = read(slurmctld_readfd, &rc, sizeof(int));
 		if (i < 0) {
-			_kill_slurmscriptd();
-			fatal("%s: slurmctld: Can not read return code from slurmscriptd: %m",
-			      __func__);
+			rc = errno;
+			killpg(slurmscriptd_pid, SIGKILL);
+			errno = rc;
+			fatal_abort("%s: slurmctld: Can not read return code from slurmscriptd: %m",
+				    __func__);
 		} else if (i != sizeof(int)) {
-			_kill_slurmscriptd();
-			fatal("%s: slurmctld: slurmscriptd failed to send return code: %m",
-			      __func__);
+			rc = errno;
+			killpg(slurmscriptd_pid, SIGKILL);
+			errno = rc;
+			fatal_abort("%s: slurmctld: slurmscriptd failed to send return code: %m",
+				    __func__);
 		}
 		if (rc != SLURM_SUCCESS) {
-			_kill_slurmscriptd();
-			fatal("%s: slurmctld: slurmscriptd did not initialize",
-			      __func__);
+			killpg(slurmscriptd_pid, SIGKILL);
+			fatal_abort("%s: slurmctld: slurmscriptd did not initialize",
+				    __func__);
 		}
 		ack = SLURM_SUCCESS;
 		i = write(slurmctld_writefd, &ack, sizeof(int));
 		if (i != sizeof(int)) {
-			_kill_slurmscriptd();
-			fatal("%s: slurmctld: failed to send ack to slurmscriptd: %m",
-			      __func__);
+			rc = errno;
+			killpg(slurmscriptd_pid, SIGKILL);
+			errno = rc;
+			fatal_abort("%s: slurmctld: failed to send ack to slurmscriptd: %m",
+				    __func__);
 		}
 
 		/* Get slurmscriptd initialization status */
@@ -1525,104 +1776,17 @@ extern int slurmscriptd_init(int argc, char **argv)
 				    _slurmctld_listener_thread, NULL);
 		debug("slurmctld: slurmscriptd fork()'d and initialized.");
 	} else { /* child (slurmscriptd_pid == 0) */
-		ssize_t i;
-		int rc = SLURM_ERROR, ack;
-		char *proc_name = "slurmscriptd";
-		char *log_prefix;
-		char *failed_plugin = NULL;
-
 		/*
-		 * Since running_in_slurmctld() is called before we fork()'d,
-		 * the result is cached in static variables, so calling it now
-		 * would return true even though we're now slurmscriptd.
-		 * Reset those cached variables so running_in_slurmctld()
-		 * returns false if called from slurmscriptd.
-		 * But first change slurm_prog_name since that is
-		 * read by run_in_daemon().
+		 * Dup needed file descriptors and re-exec self.
+		 * We do not need to closeall() here because it will happen on
+		 * the re-exec.
 		 */
-		xfree(slurm_prog_name);
-		slurm_prog_name = xstrdup(proc_name);
-		running_in_slurmctld_reset();
-
-		/*
-		 * Change the process name to slurmscriptd.
-		 * Since slurmscriptd logs to the slurmctld log file, add a
-		 * prefix to make it clear which daemon a log comes from.
-		 */
-		init_setproctitle(argc, argv);
-		setproctitle("%s", proc_name);
-#if HAVE_SYS_PRCTL_H
-		if (prctl(PR_SET_NAME, proc_name, NULL, NULL, NULL) < 0) {
-			error("%s: cannot set my name to %s %m",
-			      __func__, proc_name);
-		}
-#endif
-		/* log_set_prefix takes control of an xmalloc()'d string */
-		log_prefix = xstrdup_printf("%s: ", proc_name);
-		log_set_prefix(&log_prefix);
-
-		/* Close extra fd's. */
-		if (close(to_slurmscriptd[1]) < 0) {
-			error("%s: slurmscriptd: Unable to close write to_slurmscriptd in child: %m",
-			      __func__);
-			_exit(1);
-		}
-		if (close(to_slurmctld[0]) < 0) {
-			error("%s: slurmscriptd: Unable to close read to_slurmctld in child: %m",
-			      __func__);
-			_exit(1);
-		}
-		/* Test communiations with slurmctld. */
-		ack = SLURM_SUCCESS;
-		i = write(slurmscriptd_writefd, &ack, sizeof(int));
-		if (i != sizeof(int)) {
-			error("%s: slurmscriptd: failed to send return code to slurmctld: %m",
-			      __func__);
-			_exit(1);
-		}
-		i = read(slurmscriptd_readfd, &rc, sizeof(int));
-		if (i < 0) {
-			error("%s: slurmscriptd: Can not read ack from slurmctld: %m",
-			      __func__);
-			_exit(1);
-		} else if (i != sizeof(int)) {
-			error("%s: slurmscriptd: slurmctld failed to send ack: %m",
-			      __func__);
-			_exit(1);
-		}
-
-		debug("slurmscriptd: Got ack from slurmctld");
-
-		/*
-		 * Initialize required plugins to avoid lazy linking.
-		 * If plugins fail to initialize, send an error to slurmctld.
-		 */
-		if (bb_g_init() != SLURM_SUCCESS) {
-			failed_plugin = "burst_buffer";
-			ack = SLURM_ERROR;
-		}
-
-		i = write(slurmscriptd_writefd, &ack, sizeof(int));
-		if (i != sizeof(int))
-			fatal("%s: Failed to send initialization code to slurmctld",
-			      __func__);
-		if (ack != SLURM_SUCCESS)
-			fatal("%s: Failed to initialize %s plugin",
-			      __func__, failed_plugin);
-
-		debug("Initialization successful");
-
-		slurm_mutex_init(&powersave_script_count_mutex);
-		slurm_mutex_init(&write_mutex);
-		_slurmscriptd_mainloop();
-
-#ifdef MEMORY_LEAK_DEBUG
-		track_script_fini();
-		slurm_mutex_destroy(&powersave_script_count_mutex);
-#endif
-
-		/* We never want to return from here, only exit. */
-		_exit(0);
+		dup2(slurmscriptd_readfd, SLURMSCRIPT_READ_FD);
+		dup2(slurmscriptd_writefd, SLURMSCRIPT_WRITE_FD);
+		setenv(SLURMSCRIPTD_MODE_ENV, "1", 1);
+		execv(binary_path, argv);
+		fatal("%s: execv() failed: %m", __func__);
+		/* Never returns */
 	}
 
 	return SLURM_SUCCESS;
@@ -1635,7 +1799,7 @@ extern int slurmscriptd_fini(void)
 
 	/* Now shutdown communications. */
 	eio_signal_shutdown(msg_handle);
-	pthread_join(slurmctld_listener_tid, NULL);
+	slurm_thread_join(slurmctld_listener_tid);
 	slurm_mutex_destroy(&script_resp_map_mutex);
 	xhash_clear(script_resp_map);
 	slurm_mutex_destroy(&write_mutex);
