@@ -3,7 +3,7 @@
  *****************************************************************************
  *  Copyright (C) 2002-2007 The Regents of the University of California.
  *  Copyright (C) 2008-2009 Lawrence Livermore National Security.
- *  Copyright (C) 2010-2016 SchedMD LLC.
+ *  Copyright (C) SchedMD LLC.
  *  Copyright (C) 2013      Intel, Inc.
  *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
  *  Written by Mark Grondona <mgrondona@llnl.gov>.
@@ -52,6 +52,7 @@
 #include <poll.h>
 #include <pthread.h>
 #include <pwd.h>
+#include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -91,25 +92,27 @@
 #include "src/common/util-net.h"
 #include "src/common/x11_util.h"
 #include "src/common/xmalloc.h"
-#include "src/common/xsignal.h"
 #include "src/common/xstring.h"
 
 #include "src/interfaces/acct_gather_profile.h"
-#include "src/interfaces/core_spec.h"
 #include "src/interfaces/cred.h"
 #include "src/interfaces/gpu.h"
 #include "src/interfaces/gres.h"
 #include "src/interfaces/job_container.h"
 #include "src/interfaces/jobacct_gather.h"
 #include "src/interfaces/mpi.h"
+#include "src/interfaces/prep.h"
 #include "src/interfaces/proctrack.h"
 #include "src/interfaces/switch.h"
 #include "src/interfaces/task.h"
 
+#include "src/conmgr/conmgr.h"
+
 #include "src/slurmd/common/fname.h"
 #include "src/slurmd/common/privileges.h"
 #include "src/slurmd/common/set_oomadj.h"
 #include "src/slurmd/common/slurmd_cgroup.h"
+#include "src/slurmd/common/slurmd_common.h"
 #include "src/slurmd/common/xcpuinfo.h"
 
 #include "src/slurmd/slurmd/slurmd.h"
@@ -124,7 +127,6 @@
 #include "src/slurmd/slurmstepd/x11_forwarding.h"
 
 #define RETRY_DELAY 15		/* retry every 15 seconds */
-#define MAX_RETRY   240		/* retry 240 times (one hour max) */
 
 step_complete_t step_complete = {
 	PTHREAD_COND_INITIALIZER,
@@ -146,8 +148,12 @@ typedef struct kill_thread {
 	int       secs;
 } kill_thread_t;
 
-static pthread_t x11_signal_handler_thread = 0;
-static int sig_array[] = {SIGTERM, 0};
+#if defined(__linux__)
+typedef struct {
+	stepd_step_rec_t *step;
+	int id;
+} spank_task_args_t;
+#endif
 
 /*
  * Prototypes
@@ -174,8 +180,6 @@ static void _random_sleep(stepd_step_rec_t *step);
 static int  _run_script_as_user(const char *name, const char *path,
 				stepd_step_rec_t *step,
 				int max_wait, char **env);
-static void _unblock_signals(void);
-static void *_x11_signal_handler(void *arg);
 
 /*
  * Batch step management prototypes:
@@ -211,6 +215,7 @@ mgr_launch_tasks_setup(launch_tasks_request_msg_t *msg, slurm_addr_t *cli,
 	step->accel_bind_type = msg->accel_bind_type;
 	step->tres_bind = xstrdup(msg->tres_bind);
 	step->tres_freq = xstrdup(msg->tres_freq);
+	step->stepmgr = xstrdup(msg->stepmgr);
 
 	return step;
 }
@@ -604,7 +609,7 @@ _send_exit_msg(stepd_step_rec_t *step, uint32_t *tid, int n, int status)
 {
 	slurm_msg_t     resp;
 	task_exit_msg_t msg;
-	ListIterator    i       = NULL;
+	list_itr_t *i = NULL;
 	srun_info_t    *srun    = NULL;
 
 	debug3("%s: sending task exit msg for %d tasks (oom:%s exit_status:%s",
@@ -798,6 +803,22 @@ _one_step_complete_msg(stepd_step_rec_t *step, int first, int last)
 		       step_complete.rank, first, last);
 	}
 
+	if (step->stepmgr) {
+		slurm_msg_t resp_msg;
+
+		slurm_msg_t_init(&resp_msg);
+
+		slurm_conf_get_addr(step->stepmgr, &req.address,
+				    req.flags);
+		slurm_msg_set_r_uid(&req, slurm_conf.slurmd_user_id);
+		msg.send_to_stepmgr = true;
+		debug3("sending complete to step_ctld host:%s",
+		       step->stepmgr);
+		if (slurm_send_recv_node_msg(&req, &resp_msg, 0))
+			return;
+		goto finished;
+	}
+
 	/* Retry step complete RPC send to slurmctld indefinitely.
 	 * Prevent orphan job step if slurmctld is down */
 	i = 1;
@@ -916,6 +937,180 @@ extern void set_job_state(stepd_step_rec_t *step, slurmstepd_state_t new_state)
 	slurm_mutex_unlock(&step->state_mutex);
 }
 
+/*
+ * Run SPANK functions within the job container.
+ * WARNING: This is running as a separate process, but sharing the parent's
+ * memory space. Be careful not to leak memory, or free resources that the
+ * parent needs to continue processing. Only use _exit() here, otherwise
+ * any plugins that registered their own fini() hooks will wreck the parent.
+ */
+#if defined(__linux__)
+static int _spank_user_child(void *arg)
+{
+	stepd_step_rec_t *step = arg;
+	struct priv_state sprivs;
+	int rc = 0;
+
+	if (container_g_join(&step->step_id, step->uid, false)) {
+		error("container_g_join(%u): %m", step->step_id.job_id);
+		_exit(-1);
+	}
+
+	if (drop_privileges(step, true, &sprivs, true) < 0) {
+		error("drop_privileges: %m");
+		_exit(-1);
+	}
+
+	if (spank_user(step) < 0)
+		rc = 1;
+
+	/*
+	 * This is taken from the end of reclaim_privileges(). The child does
+	 * not need to reclaim here since we simply exit, so instead clean up
+	 * the structure and the lock so that the parent can continue to
+	 * operate.
+	 */
+	xfree(sprivs.gid_list);
+	auth_setuid_unlock();
+
+	_exit(rc);
+}
+
+static int _spank_task_post_fork_child(void *arg)
+{
+	spank_task_args_t *args = arg;
+	stepd_step_rec_t *step = args->step;
+
+	if (container_g_join(&step->step_id, step->uid, false)) {
+		error("container_g_join(%u): %m", step->step_id.job_id);
+		_exit(-1);
+	}
+
+	if (spank_task_post_fork(step, args->id) < 0)
+		_exit(1);
+
+	_exit(0);
+}
+
+static int _spank_task_exit_child(void *arg)
+{
+	spank_task_args_t *args = arg;
+	stepd_step_rec_t *step = args->step;
+
+	if (container_g_join(&step->step_id, step->uid, false)) {
+		error("container_g_join(%u): %m", step->step_id.job_id);
+		_exit(-1);
+	}
+
+	if (spank_task_exit(step, args->id) < 0)
+		_exit(1);
+
+	_exit(0);
+}
+#endif
+
+static int _run_spank_func(step_fn_t spank_func, stepd_step_rec_t *step, int id,
+			   struct priv_state *sprivs)
+{
+	int rc = SLURM_SUCCESS;
+
+#if defined(__linux__)
+	if (slurm_conf.conf_flags & CONF_FLAG_CONTAIN_SPANK) {
+		pid_t pid = -1;
+		int flags = CLONE_VM | SIGCHLD;
+		char *stack = NULL;
+		int status = 0;
+		spank_task_args_t *args = NULL;
+
+		/*
+		 * To enter the container, the process cannot share CLONE_FS
+		 * with another process. However, the spank plugins require
+		 * access to global memory that cannot be easily be shipped to
+		 * another process.
+		 *
+		 * clone() + CLONE_VM allows the new process to have distinct
+		 * filesystem attribites, but share memory space. By allowing
+		 * the current process to continue executing, shared locks can
+		 * be released allowing the new process to operate normally.
+		 */
+		if ((spank_func == SPANK_STEP_TASK_EXIT) &&
+		    spank_has_task_exit()) {
+			args = xmalloc(sizeof(*args));
+			args->step = step;
+			args->id = id;
+			stack = xmalloc(STACK_SIZE);
+			pid = clone(_spank_task_exit_child,
+				    stack + STACK_SIZE, flags, args);
+		} else if ((spank_func == SPANK_STEP_TASK_POST_FORK) &&
+			   spank_has_task_post_fork()) {
+			args = xmalloc(sizeof(*args));
+			args->step = step;
+			args->id = id;
+			stack = xmalloc(STACK_SIZE);
+			pid = clone(_spank_task_post_fork_child,
+				    stack + STACK_SIZE, flags, args);
+		} else if ((spank_func == SPANK_STEP_USER_INIT) &&
+			   spank_has_user_init()) {
+			/*
+			 * spank_user_init() runs as the user, but setns()
+			 * requires CAP_SYS_ADMIN. Reclaim privileges here so
+			 * setns() will function.
+			 */
+			if (reclaim_privileges(sprivs) < 0) {
+				error("Unable to reclaim privileges");
+				rc = 1;
+				goto fail;
+			}
+
+			stack = xmalloc(STACK_SIZE);
+			pid = clone(_spank_user_child,
+				    stack + STACK_SIZE, flags, step);
+		} else {
+			/* no action required */
+			return rc;
+		}
+
+		if (pid == -1) {
+			error("clone failed before spank call: %m");
+			rc = SLURM_ERROR;
+		} else {
+			waitpid(pid, &status, 0);
+			if (WEXITSTATUS(status))
+				rc = SLURM_ERROR;
+		}
+
+		if (spank_func == SPANK_STEP_USER_INIT) {
+			if (drop_privileges(step, true, sprivs, true) < 0) {
+				error("drop_privileges: %m");
+				rc = 2;
+			}
+		}
+
+fail:
+		xfree(args);
+		xfree(stack);
+		return rc;
+	}
+#endif
+	/*
+	 * Default case is to run these spank functions normally. To allow a
+	 * different exit path, set rc = SLURM_ERROR if the plugstack call
+	 * fails.
+	 */
+	if ((spank_func == SPANK_STEP_TASK_EXIT) &&
+	    (spank_task_exit(step, id) < 0)) {
+		rc = SLURM_ERROR;
+	} else if ((spank_func == SPANK_STEP_TASK_POST_FORK) &&
+		   (spank_task_post_fork(step, id) < 0)) {
+		rc = SLURM_ERROR;
+	} else if ((spank_func == SPANK_STEP_USER_INIT) &&
+		   (spank_user(step) < 0)) {
+		rc = SLURM_ERROR;
+	}
+
+	return rc;
+}
+
 static bool _need_join_container()
 {
 	/*
@@ -947,56 +1142,64 @@ static void _shutdown_x11_forward(stepd_step_rec_t *step)
 		error("%s: Unable to reclaim privileges", __func__);
 }
 
-static void *_x11_signal_handler(void *arg)
+static void _x11_signal_handler(conmgr_callback_args_t conmgr_args, void *arg)
 {
+	static bool run_once = false;
+	static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
+	bool bail = false;
 	stepd_step_rec_t *step = (stepd_step_rec_t *) arg;
-	int sig, status;
-	sigset_t set;
 	pid_t cpid, pid;
 
-	(void) pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
-	(void) pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
+	if (conmgr_args.status == CONMGR_WORK_STATUS_CANCELLED) {
+		debug4("%s: cancelled", __func__);
+		return;
+	}
 
-	while (1) {
-		xsignal_sigset_create(sig_array, &set);
-		if (sigwait(&set, &sig) == EINTR)
-			continue;
+	/*
+	 * Protect against race of SIGTERM and step shutdown causing this
+	 * function to run more than once.
+	 */
+	slurm_mutex_lock(&mutex);
+	if (run_once)
+		bail = true;
+	run_once = true;
+	slurm_mutex_unlock(&mutex);
 
-		switch (sig) {
-		case SIGTERM:	/* kill -15 */
-			debug("Terminate signal (SIGTERM) received");
-			if (!_need_join_container()) {
-				_shutdown_x11_forward(step);
-				return NULL;
-			}
-			if ((cpid = fork()) == 0) {
-				if (container_g_join(step->step_id.job_id,
-						     step->uid) !=
-				    SLURM_SUCCESS) {
-					error("%s: cannot join container",
-					      __func__);
-					_exit(1);
-				}
-				_shutdown_x11_forward(step);
-				_exit(0);
-			} else if (cpid < 0) {
-				error("%s: fork: %m", __func__);
-			} else {
-				pid = waitpid(cpid, &status, 0);
-				if (pid < 0)
-					error("%s: waitpid failed: %m",
-					      __func__);
-				else if (!WIFEXITED(status))
-					error("%s: child terminated abnormally",
-					      __func__);
-				else if (WEXITSTATUS(status))
-					error("%s: child returned non-zero",
-					      __func__);
-			}
-			return NULL;
-		default:
-			error("Invalid signal (%d) received", sig);
+	if (bail) {
+		debug4("%s: Already run. bailing.", __func__);
+		return;
+	}
+
+	debug("Terminate signal (SIGTERM) received");
+
+	if (!_need_join_container()) {
+		_shutdown_x11_forward(step);
+		return;
+	}
+	if ((cpid = fork()) == 0) {
+		if (container_g_join(&step->step_id, step->uid, false) !=
+		    SLURM_SUCCESS) {
+			error("%s: cannot join container",
+			      __func__);
+			_exit(1);
 		}
+		_shutdown_x11_forward(step);
+		_exit(0);
+	} else if (cpid < 0) {
+		error("%s: fork: %m", __func__);
+	} else {
+		int status;
+
+		pid = waitpid(cpid, &status, 0);
+		if (pid < 0)
+			error("%s: waitpid failed: %m",
+			      __func__);
+		else if (!WIFEXITED(status))
+			error("%s: child terminated abnormally",
+			      __func__);
+		else if (WEXITSTATUS(status))
+			error("%s: child returned non-zero",
+			      __func__);
 	}
 }
 
@@ -1010,12 +1213,25 @@ static int _set_xauthority(stepd_step_rec_t *step)
 		return SLURM_ERROR;
 	}
 
+	if (!xstrcasestr(slurm_conf.x11_params, "home_xauthority")) {
+		int fd;
+		/* protect against weak file permissions in old glibc */
+		umask(0077);
+		if ((fd = mkstemp(step->x11_xauthority)) == -1) {
+			error("%s: failed to create temporary XAUTHORITY file: %m",
+			      __func__);
+			rc = SLURM_ERROR;
+			goto endit;
+		}
+		close(fd);
+	}
+
 	if (x11_set_xauth(step->x11_xauthority, step->x11_magic_cookie,
 			  step->x11_display)) {
 		error("%s: failed to run xauth", __func__);
 		rc =  SLURM_ERROR;
 	}
-
+endit:
 	if (reclaim_privileges(&sprivs) < 0) {
 		error("%s: Unable to reclaim privileges after xauth", __func__);
 		return SLURM_ERROR;
@@ -1024,6 +1240,109 @@ static int _set_xauthority(stepd_step_rec_t *step)
 	return rc;
 }
 
+static int _run_prolog_epilog(stepd_step_rec_t *step, bool is_epilog)
+{
+	int rc = SLURM_SUCCESS;
+	job_env_t job_env;
+	list_t *tmp_list;
+
+	memset(&job_env, 0, sizeof(job_env));
+
+	tmp_list = gres_g_prep_build_env(step->job_gres_list, step->node_list);
+	gres_g_prep_set_env(&job_env.gres_job_env, tmp_list, step->nodeid);
+	FREE_NULL_LIST(tmp_list);
+
+	job_env.jobid = step->step_id.job_id;
+	job_env.step_id = SLURM_EXTERN_CONT;
+	job_env.node_list = step->node_list;
+	job_env.het_job_id = step->het_job_id;
+	job_env.partition = step->msg->cred->arg->job_partition;
+	job_env.spank_job_env = step->msg->spank_job_env;
+	job_env.spank_job_env_size = step->msg->spank_job_env_size;
+	job_env.work_dir = step->cwd;
+	job_env.uid = step->uid;
+	job_env.gid = step->gid;
+
+	if (!is_epilog)
+		rc = run_prolog(&job_env, step->msg->cred);
+	else
+		rc = run_epilog(&job_env, step->msg->cred);
+
+	if (job_env.gres_job_env) {
+		for (int i = 0; job_env.gres_job_env[i]; i++)
+			xfree(job_env.gres_job_env[i]);
+		xfree(job_env.gres_job_env);
+	}
+
+	if (rc) {
+		int term_sig = 0, exit_status = 0;
+		if (WIFSIGNALED(rc))
+			term_sig = WTERMSIG(rc);
+		else if (WIFEXITED(rc))
+			exit_status = WEXITSTATUS(rc);
+		error("[job %u] %s failed status=%d:%d", step->step_id.job_id,
+		      is_epilog ? "epilog" : "prolog", exit_status, term_sig);
+		rc = is_epilog ? ESLURMD_EPILOG_FAILED : ESLURMD_PROLOG_FAILED;
+	}
+
+	return rc;
+}
+
+static void _setup_x11_child(int to_parent[2], stepd_step_rec_t *step)
+{
+	uint32_t len = 0;
+
+	if (container_g_join(&step->step_id, step->uid, false) !=
+	    SLURM_SUCCESS) {
+		safe_write(to_parent[1], &len, sizeof(len));
+		_exit(1);
+	}
+
+	if (_set_xauthority(step) != SLURM_SUCCESS) {
+		safe_write(to_parent[1], &len, sizeof(len));
+		_exit(1);
+	}
+
+	len = strlen(step->x11_xauthority);
+	safe_write(to_parent[1], &len, sizeof(len));
+	safe_write(to_parent[1], step->x11_xauthority, len);
+
+	_exit(0);
+
+rwfail:
+	error("%s: failed to write to parent: %m", __func__);
+	_exit(1);
+}
+
+static int _setup_x11_parent(int to_parent[2], pid_t pid, char **tmp)
+{
+	uint32_t len = 0;
+	int status = 0;
+
+	safe_read(to_parent[0], &len, sizeof(len));
+
+	if (len) {
+		*tmp = xcalloc(len, sizeof(char));
+		safe_read(to_parent[0], *tmp, len);
+	}
+
+	if ((waitpid(pid, &status, 0) != pid) || WEXITSTATUS(status)) {
+		error("%s: Xauthority setup failed", __func__);
+		xfree(*tmp);
+		return SLURM_ERROR;
+	}
+
+	return SLURM_SUCCESS;
+
+rwfail:
+	error("%s: failed to read from child: %m", __func__);
+	xfree(*tmp);
+	waitpid(pid, &status, 0);
+	debug2("%s: status from child %d", __func__, status);
+
+	return SLURM_ERROR;
+}
+
 static int _spawn_job_container(stepd_step_rec_t *step)
 {
 	jobacctinfo_t *jobacct = NULL;
@@ -1032,16 +1351,8 @@ static int _spawn_job_container(stepd_step_rec_t *step)
 	int status = 0;
 	pid_t pid;
 	int rc = SLURM_SUCCESS;
-	uint32_t jobid;
+	uint32_t jobid = step->step_id.job_id;
 
-#ifdef HAVE_NATIVE_CRAY
-	if (step->het_job_id && (step->het_job_id != NO_VAL))
-		jobid = step->het_job_id;
-	else
-		jobid = step->step_id.job_id;
-#else
-	jobid = step->step_id.job_id;
-#endif
 	if (container_g_stepd_create(jobid, step)) {
 		error("%s: container_g_stepd_create(%u): %m", __func__, jobid);
 		return SLURM_ERROR;
@@ -1081,22 +1392,21 @@ static int _spawn_job_container(stepd_step_rec_t *step)
 			return SLURM_ERROR;
 		}
 
-		xsignal_block(sig_array);
-		slurm_thread_create(&x11_signal_handler_thread,
-				    _x11_signal_handler, step);
+		conmgr_add_work_signal(SIGTERM, _x11_signal_handler, step);
 
-		debug("x11 forwarding local display is %d", step->x11_display);
-		debug("x11 forwarding local xauthority is %s",
-		      step->x11_xauthority);
-	}
-
-	/*
-	 * When using job_container/tmpfs we need to get into
-	 * the correct namespace or .Xauthority won't be visible
-	 * in /tmp from inside the job.
-	 */
-	if (step->x11) {
+		/*
+		 * When using job_container/tmpfs we need to get into
+		 * the correct namespace or .Xauthority won't be visible
+		 * in /tmp from inside the job.
+		 */
 		if (_need_join_container()) {
+			int to_parent[2] = {-1, -1};
+
+			if (pipe(to_parent) < 0) {
+				error("%s: pipe failed: %m", __func__);
+				rc = SLURM_ERROR;
+				goto x11_fail;
+			}
 			/*
 			 * The fork is necessary because we cannot join a
 			 * namespace if we are multithreaded. Also we need to
@@ -1106,37 +1416,53 @@ static int _spawn_job_container(stepd_step_rec_t *step)
 			 */
 			pid = fork();
 			if (pid == 0) {
-				if (container_g_join(jobid, step->uid) !=
-				    SLURM_SUCCESS)
-					_exit(1);
-				_exit(_set_xauthority(step));
-			} else if (pid < 0) {
+				_setup_x11_child(to_parent, step);
+			} else if (pid > 0) {
+				char *tmp = NULL;
+				rc = _setup_x11_parent(to_parent, pid, &tmp);
+				xfree(step->x11_xauthority);
+				if (tmp)
+					step->x11_xauthority = tmp;
+			} else {
 				error("fork: %m");
 				rc = SLURM_ERROR;
 			}
-			if ((waitpid(pid, &status, 0) != pid) ||
-			    WEXITSTATUS(status)) {
-				error("%s: Xauthority setup failed", __func__);
-				rc = SLURM_ERROR;
-			}
+			close(to_parent[0]);
+			close(to_parent[1]);
 		} else {
 			rc = _set_xauthority(step);
 		}
-
+x11_fail:
 		if (rc != SLURM_SUCCESS) {
 			set_job_state(step, SLURMSTEPD_STEP_ENDING);
 			close_slurmd_conn(rc);
 			goto fail1;
 		}
+
+		debug("x11 forwarding local display is %d", step->x11_display);
+		debug("x11 forwarding local xauthority is %s",
+		      step->x11_xauthority);
 	}
 
 	pid = fork();
 	if (pid == 0) {
+		sigset_t sigset;
 		setpgid(0, 0);
 		setsid();
 		set_oom_adj(0);	/* the tasks may be killed by OOM */
 		acct_gather_profile_g_child_forked();
-		_unblock_signals();
+
+		/*
+		 * Mask all signals (except SIGTERM) to prevent the sleep
+		 * process from being inadvertently killed by sbatch or salloc
+		 * --signal options. If the sleep process dies then the extern
+		 * step terminates, which would break X11, pam_slurm_adopt or
+		 * stepmgr making the job to fail.
+		 */
+		sigfillset(&sigset);
+		sigdelset(&sigset, SIGTERM);
+		sigprocmask(SIG_BLOCK, &sigset, NULL);
+
 		/*
 		 * Need to exec() something for proctrack/linuxproc to
 		 * work, it will not keep a process named "slurmstepd"
@@ -1171,16 +1497,19 @@ static int _spawn_job_container(stepd_step_rec_t *step)
 	jobacct_id.step = step;
 	jobacct_gather_set_proctrack_container_id(step->cont_id);
 	jobacct_gather_add_task(pid, &jobacct_id, 1);
-	container_g_add_cont(jobid, step->cont_id);
 
 	set_job_state(step, SLURMSTEPD_STEP_RUNNING);
 	if (!slurm_conf.job_acct_gather_freq)
 		jobacct_gather_stat_task(0, true);
 
-	if (spank_task_post_fork(step, -1) < 0) {
+	if (_run_spank_func(SPANK_STEP_TASK_POST_FORK, step, -1, NULL) < 0) {
 		error("spank extern task post-fork failed");
 		rc = SLURM_ERROR;
+	} else if (slurm_conf.prolog_flags & PROLOG_FLAG_RUN_IN_JOB) {
+		rc = _run_prolog_epilog(step, false);
+	}
 
+	if (rc != SLURM_SUCCESS) {
 		/*
 		 * Failure before the tasks have even started, so we will need
 		 * to mark all of them as failed unless there is already an
@@ -1208,6 +1537,19 @@ static int _spawn_job_container(stepd_step_rec_t *step)
 		;	       /* Wait until above process exits from signal */
 	}
 
+	slurm_mutex_lock(&step->state_mutex);
+	while ((step->state < SLURMSTEPD_STEP_CANCELLED)) {
+		slurm_cond_wait(&step->state_cond, &step->state_mutex);
+	}
+	join_extern_threads();
+	slurm_mutex_unlock(&step->state_mutex);
+
+	/* Wait for all steps other than extern (this one) to complete */
+	if (!pause_for_job_completion(jobid, MAX(slurm_conf.kill_wait, 5),
+				      true)) {
+		warning("steps did not complete quickly");
+	}
+
 	/* remove all tracked tasks */
 	while ((jobacct = jobacct_gather_remove_task(0))) {
 		jobacctinfo_setinfo(jobacct,
@@ -1248,6 +1590,9 @@ static int _spawn_job_container(stepd_step_rec_t *step)
 		if (task_g_post_term(step, step->task[i]) == ENOMEM)
 			step->oom_error = true;
 
+	/* Lock to not collide with the _x11_signal_handler thread. */
+	auth_setuid_lock();
+
 	/*
 	 * This function below calls jobacct_gather_fini(). For the case of
 	 * jobacct_gather/cgroup, it ends up doing the cgroup hierarchy cleanup
@@ -1255,14 +1600,12 @@ static int _spawn_job_container(stepd_step_rec_t *step)
 	 * children processes from the step are gone.
 	 */
 	acct_gather_profile_fini();
-
 	task_g_post_step(step);
 
+	auth_setuid_unlock();
+
 fail1:
-	if (x11_signal_handler_thread) {
-		(void) pthread_kill(x11_signal_handler_thread, SIGTERM);
-		pthread_join(x11_signal_handler_thread, NULL);
-	}
+	conmgr_add_work_fifo(_x11_signal_handler, step);
 
 	debug2("%s: Before call to spank_fini()", __func__);
 	if (spank_fini(step) < 0)
@@ -1284,6 +1627,17 @@ fail1:
 
 	stepd_send_step_complete_msgs(step);
 
+	switch_g_extern_step_fini(jobid);
+
+	if (slurm_conf.prolog_flags & PROLOG_FLAG_RUN_IN_JOB) {
+		/* Force all other steps to end before epilog starts */
+		pause_for_job_completion(jobid, 0, true);
+
+		int epilog_rc = _run_prolog_epilog(step, true);
+		epilog_complete(step->step_id.job_id, step->node_list,
+				epilog_rc);
+	}
+
 	return rc;
 }
 
@@ -1380,7 +1734,7 @@ job_manager(stepd_step_rec_t *step)
 	    (mpi_g_slurmstepd_prefork(step, &step->env) != SLURM_SUCCESS)) {
 		error("Failed mpi_g_slurmstepd_prefork");
 		rc = SLURM_ERROR;
-		goto fail3;
+		goto fail2;
 	}
 
 	if (!step->batch && (step->step_id.step_id != SLURM_INTERACTIVE_STEP) &&
@@ -1411,7 +1765,7 @@ job_manager(stepd_step_rec_t *step)
 	if ((rc = _fork_all_tasks(step, &io_initialized)) < 0) {
 		debug("_fork_all_tasks failed");
 		rc = ESLURMD_EXECVE_FAILED;
-		goto fail3;
+		goto fail2;
 	}
 
 	/*
@@ -1421,7 +1775,7 @@ job_manager(stepd_step_rec_t *step)
 	 * launch to happen.
 	 */
 	if ((rc != SLURM_SUCCESS) || !io_initialized)
-		goto fail3;
+		goto fail2;
 
 	io_close_task_fds(step);
 
@@ -1453,12 +1807,6 @@ job_manager(stepd_step_rec_t *step)
 	acct_gather_profile_g_node_step_end();
 	set_job_state(step, SLURMSTEPD_STEP_ENDING);
 
-fail3:
-	if (!step->batch && (step->step_id.step_id != SLURM_INTERACTIVE_STEP) &&
-	    (switch_g_job_fini(step->switch_job) < 0)) {
-		error("switch_g_job_fini: %m");
-	}
-
 fail2:
 	/*
 	 * First call switch_g_job_postfini() - In at least one case,
@@ -1480,8 +1828,14 @@ fail2:
 	step_terminate_monitor_stop();
 	if (!step->batch && (step->step_id.step_id != SLURM_INTERACTIVE_STEP)) {
 		/* This sends a SIGKILL to the pgid */
-		if (switch_g_job_postfini(step) < 0)
-			error("switch_g_job_postfini: %m");
+		if (switch_g_job_postfini(step) < 0) {
+                        error("switch_g_job_postfini: %m");
+			/*
+			 * Drain the node since resources might still be kept.
+			 * (E.g, cxi_service for switch/hpe_slingshot.)
+			 */
+			stepd_drain_node("switch_g_job_postfini failed");
+		}
 	}
 
 	/*
@@ -1586,10 +1940,9 @@ static int _pre_task_child_privileged(
 
 	set_oom_adj(0); /* the tasks may be killed by OOM */
 
-#ifndef HAVE_NATIVE_CRAY
 	if (!(step->flags & LAUNCH_NO_ALLOC)) {
 		/* Add job's pid to job container, if a normal job */
-		if (container_g_join(step->step_id.job_id, step->uid)) {
+		if (container_g_join(&step->step_id, step->uid, false)) {
 			error("container_g_join failed: %u",
 			      step->step_id.job_id);
 			exit(1);
@@ -1602,7 +1955,6 @@ static int _pre_task_child_privileged(
 		 */
 		setwd = 1;
 	}
-#endif
 
 	if (spank_task_privileged(step, taskid) < 0)
 		return error("spank_task_init_privileged failed");
@@ -1770,12 +2122,12 @@ static int exec_wait_kill_child (struct exec_wait_info *e)
  *  Send all children in exec_wait_list SIGKILL.
  *  Returns 0 for success or  < 0 on failure.
  */
-static int exec_wait_kill_children (List exec_wait_list)
+static int exec_wait_kill_children(list_t *exec_wait_list)
 {
 	int rc = 0;
 	int count;
 	struct exec_wait_info *e;
-	ListIterator i;
+	list_itr_t *i;
 
 	if ((count = list_count (exec_wait_list)) == 0)
 		return (0);
@@ -1808,20 +2160,6 @@ static void prepare_stdio (stepd_step_rec_t *step, stepd_step_task_info_t *task)
 	return;
 }
 
-static void _unblock_signals (void)
-{
-	sigset_t set;
-	int i;
-
-	for (i = 0; slurmstepd_blocked_signals[i]; i++) {
-		/* eliminate pending signals, then set to default */
-		xsignal(slurmstepd_blocked_signals[i], SIG_IGN);
-		xsignal(slurmstepd_blocked_signals[i], SIG_DFL);
-	}
-	sigemptyset(&set);
-	xsignal_set_mask (&set);
-}
-
 /*
  * fork and exec N tasks
  */
@@ -1832,9 +2170,9 @@ _fork_all_tasks(stepd_step_rec_t *step, bool *io_initialized)
 	int i;
 	struct priv_state sprivs;
 	jobacct_id_t jobacct_id;
-	List exec_wait_list = NULL;
-	uint32_t jobid;
+	list_t *exec_wait_list = NULL;
 	uint32_t node_offset = 0, task_offset = 0;
+	char saved_cwd[PATH_MAX];
 
 	if (step->het_job_node_offset != NO_VAL)
 		node_offset = step->het_job_node_offset;
@@ -1846,6 +2184,11 @@ _fork_all_tasks(stepd_step_rec_t *step, bool *io_initialized)
 
 	xassert(step != NULL);
 
+	if (!getcwd(saved_cwd, sizeof(saved_cwd))) {
+		error ("Unable to get current working directory: %m");
+		strlcpy(saved_cwd, "/tmp", sizeof(saved_cwd));
+	}
+
 	if (task_g_pre_setuid(step)) {
 		error("Failed to invoke task plugins: one of task_p_pre_setuid functions returned error");
 		return SLURM_ERROR;
@@ -1947,18 +2290,26 @@ _fork_all_tasks(stepd_step_rec_t *step, bool *io_initialized)
 		}
 	}
 
-	if (spank_user (step) < 0) {
-		error("spank_user failed.");
-		rc = SLURM_ERROR;
+	if ((rc = _run_spank_func(SPANK_STEP_USER_INIT, step, -1, &sprivs))) {
+		if (rc < 0) {
+			error("spank_user failed.");
+			rc = SLURM_ERROR;
 
-		step->task[0]->estatus = W_EXITCODE(1, 0);
-		step->task[0]->exited = true;
-		slurm_mutex_lock(&step_complete.lock);
-		if (!step_complete.step_rc)
-			step_complete.step_rc = rc;
-		slurm_mutex_unlock(&step_complete.lock);
-
-		goto fail4;
+			step->task[0]->estatus = W_EXITCODE(1, 0);
+			step->task[0]->exited = true;
+			slurm_mutex_lock(&step_complete.lock);
+			if (!step_complete.step_rc)
+				step_complete.step_rc = rc;
+			slurm_mutex_unlock(&step_complete.lock);
+			goto fail4;
+		} else {
+			/*
+			 * A drop_privileges() or reclaim_privileges() failed,
+			 * In this case, bail out skipping the redundant
+			 * reclaim.
+			 */
+			goto fail2;
+		}
 	}
 
 	exec_wait_list = list_create ((ListDelF) _exec_wait_info_destroy);
@@ -2015,8 +2366,6 @@ _fork_all_tasks(stepd_step_rec_t *step, bool *io_initialized)
 
 			/* log_fini(); */ /* note: moved into exec_task() */
 
-			_unblock_signals();
-
 			/*
 			 *  Need to setup stdio before setpgid() is called
 			 *   in case we are setting up a tty. (login_tty()
@@ -2069,7 +2418,7 @@ _fork_all_tasks(stepd_step_rec_t *step, bool *io_initialized)
 		/* Don't bother erroring out here */
 	}
 
-	if (chdir(sprivs.saved_cwd) < 0) {
+	if (chdir(saved_cwd) < 0) {
 		error ("Unable to return to working directory");
 	}
 
@@ -2117,7 +2466,7 @@ _fork_all_tasks(stepd_step_rec_t *step, bool *io_initialized)
 			goto fail2;
 		}
 
-		if (spank_task_post_fork (step, i) < 0) {
+		if (_run_spank_func(SPANK_STEP_TASK_POST_FORK, step, i, NULL) < 0) {
 			error ("spank task %d post-fork failed", i);
 			rc = SLURM_ERROR;
 
@@ -2141,20 +2490,6 @@ _fork_all_tasks(stepd_step_rec_t *step, bool *io_initialized)
 		}
 	}
 //	jobacct_gather_set_proctrack_container_id(step->cont_id);
-#ifdef HAVE_NATIVE_CRAY
-	if (step->het_job_id && (step->het_job_id != NO_VAL))
-		jobid = step->het_job_id;
-	else
-		jobid = step->step_id.job_id;
-#else
-	jobid = step->step_id.job_id;
-#endif
-	if (container_g_add_cont(jobid, step->cont_id) != SLURM_SUCCESS)
-		error("container_g_add_cont(%u): %m", step->step_id.job_id);
-	if (!step->batch && (step->step_id.step_id != SLURM_INTERACTIVE_STEP) &&
-	    core_spec_g_set(step->cont_id, step->job_core_spec) &&
-	    (step->step_id.step_id == 0))
-		error("core_spec_g_set: %m");
 
 	/*
 	 * Now it's ok to unblock the tasks, so they may call exec.
@@ -2177,7 +2512,7 @@ _fork_all_tasks(stepd_step_rec_t *step, bool *io_initialized)
 	return rc;
 
 fail4:
-	if (chdir (sprivs.saved_cwd) < 0) {
+	if (chdir(saved_cwd) < 0) {
 		error ("Unable to return to working directory");
 	}
 fail3:
@@ -2286,7 +2621,7 @@ static int
 _wait_for_any_task(stepd_step_rec_t *step, bool waitflag)
 {
 	stepd_step_task_info_t *t = NULL;
-	int rc, status = 0;
+	int rc = 0, status = 0;
 	pid_t pid;
 	int completed = 0;
 	jobacctinfo_t *jobacct = NULL;
@@ -2351,6 +2686,7 @@ _wait_for_any_task(stepd_step_rec_t *step, bool waitflag)
 			step->envtp->uid = step->uid;
 			step->envtp->user_name = xstrdup(step->user_name);
 			step->envtp->nodeid = step->nodeid;
+			step->envtp->oom_kill_step = step->oom_kill_step ? 1 : 0;
 
 			/*
 			 * Modify copy of job's environment. Do not alter in
@@ -2385,11 +2721,10 @@ _wait_for_any_task(stepd_step_rec_t *step, bool waitflag)
 					error("--task-epilog failed status=%d",
 					      rc);
 			}
-
-			if (spank_task_exit(step, t->id) < 0) {
+			if (_run_spank_func(SPANK_STEP_TASK_EXIT, step, t->id,
+					    NULL) < 0)
 				error ("Unable to spank task %d at exit",
 				       t->id);
-			}
 			rc = task_g_post_term(step, t);
 			if (rc == ENOMEM)
 				step->oom_error = true;
@@ -2426,19 +2761,22 @@ _wait_for_all_tasks(stepd_step_rec_t *step)
 
 	for (i = 0; i < tasks_left; ) {
 		int rc;
-		rc = _wait_for_any_task(step, true);
-		if (rc != -1) {
-			i += rc;
-			if (i < tasks_left) {
-				/* To limit the amount of traffic back
-				 * we will sleep a bit to make sure we
-				 * have most if not all the tasks
-				 * completed before we return */
-				usleep(100000);	/* 100 msec */
-				rc = _wait_for_any_task(step, false);
-				if (rc != -1)
-					i += rc;
-			}
+		if ((rc = _wait_for_any_task(step, true)) == -1) {
+			error("%s: No child processes. node_tasks:%u, expected:%d, reaped:%d",
+			      __func__, step->node_tasks, tasks_left, i);
+			break;
+		}
+
+		i += rc;
+		if (i < tasks_left) {
+			/* To limit the amount of traffic back
+			 * we will sleep a bit to make sure we
+			 * have most if not all the tasks
+			 * completed before we return */
+			usleep(100000);	/* 100 msec */
+			rc = _wait_for_any_task(step, false);
+			if (rc != -1)
+				i += rc;
 		}
 
 		if (i < tasks_left) {
@@ -2700,7 +3038,7 @@ _send_launch_resp(stepd_step_rec_t *step, int rc)
 static int
 _send_complete_batch_script_msg(stepd_step_rec_t *step, int err, int status)
 {
-	int		rc, i, msg_rc;
+	int rc;
 	slurm_msg_t	req_msg;
 	complete_batch_script_msg_t req;
 
@@ -2722,18 +3060,16 @@ _send_complete_batch_script_msg(stepd_step_rec_t *step, int err, int status)
 		 slurm_strerror(err), status);
 
 	/* Note: these log messages don't go to slurmd.log from here */
-	for (i = 0; i <= MAX_RETRY; i++) {
-		msg_rc = slurm_send_recv_controller_rc_msg(&req_msg, &rc,
-							   working_cluster_rec);
-		if (msg_rc == SLURM_SUCCESS)
-			break;
-		info("Retrying job complete RPC for %ps", &step->step_id);
+
+	/*
+	 * Retry batch complete RPC, send to slurmctld indefinitely.
+	 */
+	while (slurm_send_recv_controller_rc_msg(&req_msg, &rc,
+						 working_cluster_rec)) {
+		info("Retrying job complete RPC for %ps [sleeping %us]",
+		     &step->step_id, RETRY_DELAY);
 		sleep(RETRY_DELAY);
 	}
-	if (i > MAX_RETRY) {
-		error("Unable to send job complete message: %m");
-		return SLURM_ERROR;
-	}
 
 	if ((rc == ESLURM_ALREADY_DONE) || (rc == ESLURM_INVALID_JOB_ID))
 		rc = SLURM_SUCCESS;
@@ -2905,25 +3241,17 @@ _run_script_as_user(const char *name, const char *path, stepd_step_rec_t *step,
 	if ((cpid = _exec_wait_get_pid (ei)) == 0) {
 		struct priv_state sprivs;
 		char *argv[2];
-		uint32_t jobid;
 
-#ifdef HAVE_NATIVE_CRAY
-		if (step->het_job_id && (step->het_job_id != NO_VAL))
-			jobid = step->het_job_id;
-		else
-			jobid = step->step_id.job_id;
-#else
-		jobid = step->step_id.job_id;
-#endif
 		/* container_g_join needs to be called in the
 		   forked process part of the fork to avoid a race
 		   condition where if this process makes a file or
 		   detacts itself from a child before we add the pid
 		   to the container in the parent of the fork.
 		*/
-		if ((jobid != 0) &&	/* Ignore system processes */
+		/* Ignore system processes */
+		if ((step->step_id.job_id != 0) &&
 		    !(step->flags & LAUNCH_NO_ALLOC) &&
-		    (container_g_join(jobid, step->uid) != SLURM_SUCCESS))
+		    (container_g_join(&step->step_id, step->uid, false)))
 			error("container_g_join(%u): %m", step->step_id.job_id);
 
 		argv[0] = (char *)xstrdup(path);
