@@ -48,38 +48,42 @@
 
 #include "src/common/assoc_mgr.h"
 #include "src/common/cpu_frequency.h"
-#include "src/interfaces/gres.h"
-#include "src/interfaces/hash.h"
+#include "src/common/forward.h"
+#include "src/common/macros.h"
+#include "src/common/node_features.h"
+#include "src/common/port_mgr.h"
 #include "src/common/run_command.h"
-#include "src/interfaces/route.h"
-#include "src/interfaces/select.h"
-#include "src/interfaces/topology.h"
 #include "src/common/setproctitle.h"
-#include "src/interfaces/auth.h"
-#include "src/interfaces/jobacct_gather.h"
-#include "src/interfaces/acct_gather_profile.h"
-#include "src/interfaces/mpi.h"
 #include "src/common/slurm_protocol_api.h"
 #include "src/common/slurm_protocol_pack.h"
 #include "src/common/slurm_rlimits_info.h"
 #include "src/common/spank.h"
 #include "src/common/stepd_api.h"
-#include "src/interfaces/switch.h"
 #include "src/common/xmalloc.h"
-#include "src/common/xsignal.h"
 #include "src/common/xstring.h"
+
+#include "src/conmgr/conmgr.h"
+
+#include "src/interfaces/acct_gather_energy.h"
+#include "src/interfaces/acct_gather_profile.h"
+#include "src/interfaces/auth.h"
 #include "src/interfaces/cgroup.h"
-
-#include "src/interfaces/core_spec.h"
 #include "src/interfaces/gpu.h"
+#include "src/interfaces/gres.h"
+#include "src/interfaces/hash.h"
 #include "src/interfaces/job_container.h"
+#include "src/interfaces/jobacct_gather.h"
+#include "src/interfaces/mpi.h"
+#include "src/interfaces/prep.h"
+#include "src/interfaces/proctrack.h"
+#include "src/interfaces/select.h"
+#include "src/interfaces/switch.h"
+#include "src/interfaces/task.h"
+#include "src/interfaces/topology.h"
 
+#include "src/slurmd/common/privileges.h"
 #include "src/slurmd/common/set_oomadj.h"
 #include "src/slurmd/common/slurmstepd_init.h"
-#include "src/interfaces/task.h"
-#include "src/interfaces/acct_gather_energy.h"
-#include "src/interfaces/proctrack.h"
-#include "src/slurmd/common/xcpuinfo.h"
 #include "src/slurmd/slurmd/slurmd.h"
 #include "src/slurmd/slurmstepd/container.h"
 #include "src/slurmd/slurmstepd/mgr.h"
@@ -87,90 +91,343 @@
 #include "src/slurmd/slurmstepd/slurmstepd.h"
 #include "src/slurmd/slurmstepd/slurmstepd_job.h"
 
+#include "src/stepmgr/stepmgr.h"
+
 static int _init_from_slurmd(int sock, char **argv, slurm_addr_t **_cli,
-			     slurm_addr_t **_self, slurm_msg_t **_msg);
+			    slurm_msg_t **_msg);
 
-static void _dump_user_env(void);
 static void _send_ok_to_slurmd(int sock);
-static void _send_fail_to_slurmd(int sock);
+static void _send_fail_to_slurmd(int sock, int rc);
 static void _got_ack_from_slurmd(int);
-static stepd_step_rec_t *_step_setup(slurm_addr_t *cli, slurm_addr_t *self,
-				     slurm_msg_t *msg);
+static stepd_step_rec_t *_step_setup(slurm_addr_t *cli, slurm_msg_t *msg);
 #ifdef MEMORY_LEAK_DEBUG
 static void _step_cleanup(stepd_step_rec_t *step, slurm_msg_t *msg, int rc);
 #endif
-static int _process_cmdline (int argc, char **argv);
+static void _process_cmdline(int argc, char **argv);
 
 static pthread_mutex_t cleanup_mutex = PTHREAD_MUTEX_INITIALIZER;
 static bool cleanup = false;
 
-/*
- *  List of signals to block in this process
- */
-int slurmstepd_blocked_signals[] = {
-	SIGINT,  SIGTERM, SIGTSTP,
-	SIGQUIT, SIGPIPE, SIGUSR1,
-	SIGUSR2, SIGALRM, SIGHUP, 0
-};
-
 /* global variable */
 slurmd_conf_t * conf;
 extern char  ** environ;
 
-int
-main (int argc, char **argv)
+list_t *job_list = NULL;
+job_record_t *job_step_ptr = NULL;
+list_t *job_node_array = NULL;
+time_t last_job_update = 0;
+bool time_limit_thread_shutdown = false;
+pthread_t time_limit_thread_id = 0;
+
+/* See _send_msg_maybe() in src/slurmctld/agent.c */
+static void _send_msg_maybe(slurm_msg_t *req)
+{
+	int fd = -1;
+
+	if ((fd = slurm_open_msg_conn(&req->address)) < 0) {
+		log_flag(NET, "%s: slurm_open_msg_conn(%pA): %m",
+			 __func__, &req->address);
+		return;
+	}
+
+	(void) slurm_send_node_msg(fd, req);
+
+	(void) close(fd);
+}
+
+static int _foreach_ret_data_info(void *x, void *arg)
+{
+	int rc;
+	ret_data_info_t *ret_data_info = x;
+
+	if ((rc = slurm_get_return_code(ret_data_info->type,
+					ret_data_info->data))) {
+		error("stepmgr failed to send message %s: rc=%d(%s)",
+		      rpc_num2string(ret_data_info->type), rc,
+		      slurm_strerror(rc));
+		return SLURM_ERROR;
+	}
+
+	return SLURM_SUCCESS;
+}
+
+static void *_rpc_thread(void *data)
+{
+	bool srun_agent = false;
+	agent_arg_t *agent_arg_ptr = data;
+	slurm_msg_t msg;
+	slurm_msg_t_init(&msg);
+
+	msg.data = agent_arg_ptr->msg_args;
+	msg.flags = agent_arg_ptr->msg_flags;
+	msg.msg_type = agent_arg_ptr->msg_type;
+	msg.protocol_version = agent_arg_ptr->protocol_version;
+
+	slurm_msg_set_r_uid(&msg, agent_arg_ptr->r_uid);
+
+	srun_agent = ((msg.msg_type == SRUN_PING) ||
+		      (msg.msg_type == SRUN_JOB_COMPLETE) ||
+		      (msg.msg_type == SRUN_STEP_MISSING) ||
+		      (msg.msg_type == SRUN_STEP_SIGNAL) ||
+		      (msg.msg_type == SRUN_TIMEOUT) ||
+		      (msg.msg_type == SRUN_USER_MSG) ||
+		      (msg.msg_type == RESPONSE_RESOURCE_ALLOCATION) ||
+		      (msg.msg_type == SRUN_NODE_FAIL));
+
+	if (agent_arg_ptr->addr) {
+		msg.address = *agent_arg_ptr->addr;
+
+		if (msg.msg_type == SRUN_JOB_COMPLETE) {
+			_send_msg_maybe(&msg);
+		} else if (slurm_send_only_node_msg(&msg) && !srun_agent) {
+			error("failed to send message type %d/%s",
+			      msg.msg_type, rpc_num2string(msg.msg_type));
+		}
+	} else {
+		list_t *ret_list = NULL;
+		if (!(ret_list = start_msg_tree(agent_arg_ptr->hostlist,
+						&msg, 0))) {
+			error("%s: no ret_list given", __func__);
+		} else {
+			list_for_each(ret_list, _foreach_ret_data_info, NULL);
+			FREE_NULL_LIST(ret_list);
+		}
+	}
+
+	purge_agent_args(agent_arg_ptr);
+
+	return NULL;
+}
+
+static void _agent_queue_request(agent_arg_t *agent_arg_ptr)
+{
+	slurm_thread_create_detached(_rpc_thread, agent_arg_ptr);
+}
+
+extern job_record_t *find_job_record(uint32_t job_id)
+{
+	xassert(job_step_ptr);
+
+	return job_step_ptr;
+}
+
+static void *_step_time_limit_thread(void *data)
+{
+	time_t now;
+
+	xassert(job_step_ptr);
+
+	while (!time_limit_thread_shutdown) {
+		now = time(NULL);
+		slurm_mutex_lock(&stepmgr_mutex);
+		list_for_each(job_step_ptr->step_list,
+			      check_job_step_time_limit, &now);
+		slurm_mutex_unlock(&stepmgr_mutex);
+		sleep(1);
+	}
+
+	return NULL;
+}
+
+stepmgr_ops_t stepd_stepmgr_ops = {
+	.find_job_record = find_job_record,
+	.last_job_update = &last_job_update,
+	.agent_queue_request = _agent_queue_request
+};
+
+static int _foreach_job_node_array(void *x, void *arg)
+{
+	node_record_t *job_node_ptr = x;
+	int *table_index = arg;
+
+	config_record_t *config_ptr =
+		config_record_from_node_record(job_node_ptr);
+
+	*table_index = bit_ffs_from_bit(job_step_ptr->node_bitmap, *table_index);
+
+	job_node_ptr->config_ptr = config_ptr;
+	insert_node_record_at(job_node_ptr, *table_index);
+
+	(*table_index)++;
+
+	job_node_ptr->tot_cores =
+		job_node_ptr->tot_sockets * job_node_ptr->cores;
+	/*
+	 * Sanity check to make sure we can take a version we
+	 * actually understand.
+	 */
+	if (job_node_ptr->protocol_version < SLURM_MIN_PROTOCOL_VERSION)
+		job_node_ptr->protocol_version = SLURM_MIN_PROTOCOL_VERSION;
+
+	return SLURM_SUCCESS;
+}
+
+static void _setup_stepmgr_nodes(void)
+{
+	int table_index = 0;
+	init_node_conf();
+
+	xassert(job_node_array);
+	/*
+	 * next_node_bitmap() asserts
+	 * bit_size(node_bitmap) == node_record_count
+	 */
+	node_record_count = bit_size(job_step_ptr->node_bitmap);
+	grow_node_record_table_ptr();
+	list_for_each(job_node_array, _foreach_job_node_array, &table_index);
+}
+
+static void _init_stepd_stepmgr(void)
+{
+	if (!job_step_ptr)
+		return;
+
+	stepd_stepmgr_ops.up_node_bitmap =
+		bit_alloc(bit_size(job_step_ptr->node_bitmap));
+	bit_set_all(stepd_stepmgr_ops.up_node_bitmap);
+	stepmgr_init(&stepd_stepmgr_ops);
+	reserve_port_stepmgr_init(job_step_ptr);
+
+	_setup_stepmgr_nodes();
+	node_features_build_active_list(job_step_ptr);
+
+	if (!xstrcasecmp(slurm_conf.accounting_storage_type,
+			 "accounting_storage/slurmdbd")) {
+		xfree(slurm_conf.accounting_storage_type);
+		slurm_conf.accounting_storage_type =
+			xstrdup("accounting_storage/ctld_relay");
+		acct_storage_g_init();
+	} else {
+			acct_storage_g_init();
+	}
+
+	slurm_thread_create(&time_limit_thread_id, _step_time_limit_thread,
+			    NULL);
+}
+
+static void _on_sigalrm(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	debug("Caught SIGALRM. Ignoring.");
+}
+
+static void _on_sigint(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	info("Caught SIGINT. Shutting down.");
+	conmgr_request_shutdown();
+}
+
+static void _on_sigterm(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	info("Caught SIGTERM. Shutting down.");
+	conmgr_request_shutdown();
+}
+
+static void _on_sigquit(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	info("Caught SIGQUIT. Shutting down.");
+	conmgr_request_shutdown();
+}
+
+static void _on_sigtstp(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	info("Caught SIGTSTP. Ignoring");
+}
+
+static void _on_sighup(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	info("Caught SIGHUP. Ignoring");
+}
+
+static void _on_sigusr1(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	info("Caught SIGUSR1. Ignoring.");
+}
+
+static void _on_sigusr2(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	info("Caught SIGUSR2. Ignoring.");
+}
+
+static void _on_sigpipe(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	info("Caught SIGPIPE. Ignoring.");
+}
+
+static void _on_sigttin(conmgr_callback_args_t conmgr_args, void *arg)
+{
+	debug("Caught SIGTTIN. Ignoring.");
+}
+
+extern int main(int argc, char **argv)
 {
 	log_options_t lopts = LOG_OPTS_INITIALIZER;
 	slurm_addr_t *cli;
-	slurm_addr_t *self;
 	slurm_msg_t *msg;
 	stepd_step_rec_t *step;
-	int rc = 0;
+	int rc = SLURM_SUCCESS;
+	bool only_mem = true;
 
-	if (_process_cmdline (argc, argv) < 0)
-		fatal ("Error in slurmstepd command line");
+	_process_cmdline(argc, argv);
 
-	run_command_init();
-
-	xsignal_block(slurmstepd_blocked_signals);
 	conf = xmalloc(sizeof(*conf));
-	conf->argv = &argv;
-	conf->argc = &argc;
+	conf->argv = argv;
+	conf->argc = argc;
 	init_setproctitle(argc, argv);
 
 	log_init(argv[0], lopts, LOG_DAEMON, NULL);
 
 	/* Receive job parameters from the slurmd */
-	_init_from_slurmd(STDIN_FILENO, argv, &cli, &self, &msg);
+	_init_from_slurmd(STDIN_FILENO, argv, &cli, &msg);
+
+	conmgr_init(0, 0, (conmgr_callbacks_t) {0});
+
+	conmgr_add_work_signal(SIGALRM, _on_sigalrm, NULL);
+	conmgr_add_work_signal(SIGINT, _on_sigint, NULL);
+	conmgr_add_work_signal(SIGTERM, _on_sigterm, NULL);
+	conmgr_add_work_signal(SIGQUIT, _on_sigquit, NULL);
+	conmgr_add_work_signal(SIGTSTP, _on_sigtstp, NULL);
+	conmgr_add_work_signal(SIGHUP, _on_sighup, NULL);
+	conmgr_add_work_signal(SIGUSR1, _on_sigusr1, NULL);
+	conmgr_add_work_signal(SIGUSR2, _on_sigusr2, NULL);
+	conmgr_add_work_signal(SIGPIPE, _on_sigpipe, NULL);
+	conmgr_add_work_signal(SIGTTIN, _on_sigttin, NULL);
+
+	conmgr_run(false);
+
+	if ((run_command_init(argc, argv, conf->stepd_loc) != SLURM_SUCCESS) &&
+	    conf->stepd_loc && conf->stepd_loc[0])
+		fatal("%s: Unable to reliably execute %s",
+		      __func__, conf->stepd_loc);
 
 	/* Create the stepd_step_rec_t, mostly from info in a
 	 * launch_tasks_request_msg_t or a batch_job_launch_msg_t */
-	if (!(step = _step_setup(cli, self, msg))) {
-		_send_fail_to_slurmd(STDOUT_FILENO);
+	if (!(step = _step_setup(cli, msg))) {
 		rc = SLURM_ERROR;
+		_send_fail_to_slurmd(STDOUT_FILENO, rc);
 		goto ending;
 	}
 
+	_init_stepd_stepmgr();
+
 	/* fork handlers cause mutexes on some global data structures
 	 * to be re-initialized after the fork. */
 	slurm_conf_install_fork_handlers();
 
 	/* sets step->msg_handle and step->msgid */
 	if (msg_thr_create(step) == SLURM_ERROR) {
-		_send_fail_to_slurmd(STDOUT_FILENO);
 		rc = SLURM_ERROR;
+		_send_fail_to_slurmd(STDOUT_FILENO, rc);
 		goto ending;
 	}
 
 	if (step->step_id.step_id != SLURM_EXTERN_CONT)
-		close_slurmd_conn();
+		close_slurmd_conn(rc);
 
 	/* slurmstepd is the only daemon that should survive upgrade. If it
 	 * had been swapped out before upgrade happened it could easily lead
 	 * to SIGBUS at any time after upgrade. Avoid that by locking it
 	 * in-memory. */
 	if (xstrstr(slurm_conf.launch_params, "slurmstepd_memlock")) {
-#ifdef _POSIX_MEMLOCK
 		int flags = MCL_CURRENT;
 		if (xstrstr(slurm_conf.launch_params, "slurmstepd_memlock_all"))
 			flags |= MCL_FUTURE;
@@ -178,9 +435,6 @@ main (int argc, char **argv)
 			info("failed to mlock() slurmstepd pages: %m");
 		else
 			debug("slurmstepd locked in memory");
-#else
-		info("mlockall() system call does not appear to be available");
-#endif
 	}
 
 	acct_gather_energy_g_set_data(ENERGY_DATA_STEP_PTR, step);
@@ -189,15 +443,19 @@ main (int argc, char **argv)
 	 * and blocks until the step is complete */
 	rc = job_manager(step);
 
-	return stepd_cleanup(msg, step, cli, self, rc, 0);
+	only_mem = false;
 ending:
-	return stepd_cleanup(msg, step, cli, self, rc, 1);
+	rc = stepd_cleanup(msg, step, cli, rc, only_mem);
+
+	conmgr_fini();
+	return rc;
 }
 
 extern int stepd_cleanup(slurm_msg_t *msg, stepd_step_rec_t *step,
-			 slurm_addr_t *cli, slurm_addr_t *self,
-			 int rc, bool only_mem)
+			 slurm_addr_t *cli, int rc, bool only_mem)
 {
+	time_limit_thread_shutdown = true;
+
 	slurm_mutex_lock(&cleanup_mutex);
 
 	if (cleanup)
@@ -208,14 +466,22 @@ extern int stepd_cleanup(slurm_msg_t *msg, stepd_step_rec_t *step,
 		goto done;
 	}
 
-	if (!only_mem) {
-		if (step->batch)
-			batch_finish(step, rc); /* sends batch complete message */
+	if (!only_mem && step->batch)
+		batch_finish(step, rc); /* sends batch complete message */
 
+	/*
+	 * Call auth_setuid_lock after sending the batch_finish message as in
+	 * there the auth_g_create function is called, and that function uses
+	 * the lock. The lock is needed to ensure that the privileges are not
+	 * dropped from a different thread, like X11 shutdown thread.
+	 */
+	auth_setuid_lock();
+
+	if (!only_mem) {
 		/* signal the message thread to shutdown, and wait for it */
 		if (step->msg_handle)
 			eio_signal_shutdown(step->msg_handle);
-		pthread_join(step->msgid, NULL);
+		slurm_thread_join(step->msgid);
 	}
 
 	mpi_fini();
@@ -232,25 +498,31 @@ extern int stepd_cleanup(slurm_msg_t *msg, stepd_step_rec_t *step,
 	if (step->container)
 		cleanup_container(step);
 
+	if (step->step_id.step_id == SLURM_EXTERN_CONT) {
+		if (container_g_stepd_delete(step->step_id.job_id))
+			error("container_g_stepd_delete(%u): %m",
+			      step->step_id.job_id);
+	}
+
+	auth_setuid_unlock();
 	run_command_shutdown();
 
-	if (step->step_id.step_id == SLURM_EXTERN_CONT) {
-		uint32_t jobid;
-#ifdef HAVE_NATIVE_CRAY
-		if (step->het_job_id && (step->het_job_id != NO_VAL))
-			jobid = step->het_job_id;
-		else
-			jobid = step->step_id.job_id;
-#else
-		jobid = step->step_id.job_id;
-#endif
-		if (container_g_stepd_delete(jobid))
-			error("container_g_stepd_delete(%u): %m", jobid);
-	}
+	/*
+	 * join() must be done before _step_cleanup() where job_step_ptr is
+	 * freed.
+	 */
+	slurm_thread_join(time_limit_thread_id);
 
 #ifdef MEMORY_LEAK_DEBUG
 	acct_gather_conf_destroy();
-	(void) core_spec_g_fini();
+	acct_storage_g_fini();
+
+	if (job_step_ptr) {
+		xfree(job_step_ptr->resv_ports);
+		reserve_port_stepmgr_init(job_step_ptr);
+		node_features_free_lists();
+	}
+
 	_step_cleanup(step, msg, rc);
 
 	fini_setproctitle();
@@ -258,9 +530,9 @@ extern int stepd_cleanup(slurm_msg_t *msg, stepd_step_rec_t *step,
 	cgroup_conf_destroy();
 
 	xfree(cli);
-	xfree(self);
 	xfree(conf->block_map);
 	xfree(conf->block_map_inv);
+	xfree(conf->conffile);
 	xfree(conf->hostname);
 	xfree(conf->hwloc_xml);
 	xfree(conf->logfile);
@@ -268,19 +540,37 @@ extern int stepd_cleanup(slurm_msg_t *msg, stepd_step_rec_t *step,
 	xfree(conf->node_topo_addr);
 	xfree(conf->node_topo_pattern);
 	xfree(conf->spooldir);
+	xfree(conf->stepd_loc);
 	xfree(conf->cpu_spec_list);
 	xfree(conf);
 #endif
 	cleanup = true;
 done:
 	slurm_mutex_unlock(&cleanup_mutex);
-	info("done with job");
+	/* skipping lock of step_complete.lock */
+	if (rc || step_complete.step_rc) {
+		info("%s: done with step (rc[0x%x]:%s, cleanup_rc[0x%x]:%s)",
+		     __func__, step_complete.step_rc,
+		     slurm_strerror(step_complete.step_rc), rc,
+		     slurm_strerror(rc));
+	} else {
+		info("done with step");
+	}
+
+	conmgr_request_shutdown();
+
 	return rc;
 }
 
-extern void close_slurmd_conn(void)
+extern void close_slurmd_conn(int rc)
 {
-	_send_ok_to_slurmd(STDOUT_FILENO);
+	debug("%s: sending %d: %s", __func__, rc, slurm_strerror(rc));
+
+	if (rc)
+		_send_fail_to_slurmd(STDOUT_FILENO, rc);
+	else
+		_send_ok_to_slurmd(STDOUT_FILENO);
+
 	_got_ack_from_slurmd(STDIN_FILENO);
 
 	/* Fancy way of closing stdin that keeps STDIN_FILENO from being
@@ -301,7 +591,7 @@ static slurmd_conf_t *_read_slurmd_conf_lite(int fd)
 	buf_t *buffer = NULL;
 	slurmd_conf_t *confl, *local_conf = NULL;
 	int tmp_int = 0;
-	List tmp_list = NULL;
+	list_t *tmp_list = NULL;
 	assoc_mgr_lock_t locks = { .tres = WRITE_LOCK };
 
 	/*  First check to see if we've already initialized the
@@ -415,7 +705,7 @@ static int _get_jobid_uid_gid_from_env(uint32_t *jobid, uid_t *uid, gid_t *gid)
 	return SLURM_SUCCESS;
 }
 
-static int _handle_spank_mode (int argc, char **argv)
+static int _handle_spank_mode(int argc, char **argv)
 {
 	char *prefix = NULL;
 	const char *mode = argv[2];
@@ -444,8 +734,8 @@ static int _handle_spank_mode (int argc, char **argv)
 	 *   This could happen if slurmstepd is run standalone for
 	 *   testing.
 	 */
-	conf = _read_slurmd_conf_lite (STDIN_FILENO);
-	close (STDIN_FILENO);
+	conf = _read_slurmd_conf_lite(STDIN_FILENO);
+	close(STDIN_FILENO);
 
 	if (_get_jobid_uid_gid_from_env(&jobid, &uid, &gid))
 		return error("spank environment invalid");
@@ -453,51 +743,55 @@ static int _handle_spank_mode (int argc, char **argv)
 	debug("Running spank/%s for jobid [%u] uid [%u] gid [%u]",
 	      mode, jobid, uid, gid);
 
-	if (xstrcmp (mode, "prolog") == 0) {
+	if (!xstrcmp(mode, "prolog")) {
 		if (spank_job_prolog(jobid, uid, gid) < 0)
-			return (-1);
-	}
-	else if (xstrcmp (mode, "epilog") == 0) {
+			return -1;
+	} else if (!xstrcmp(mode, "epilog")) {
 		if (spank_job_epilog(jobid, uid, gid) < 0)
-			return (-1);
+			return -1;
+	} else {
+		error("Invalid mode %s specified!", mode);
+		return -1;
 	}
-	else {
-		error ("Invalid mode %s specified!", mode);
-		return (-1);
-	}
-	return (0);
+
+	return 0;
 }
 
 /*
  *  Process special "modes" of slurmstepd passed as cmdline arguments.
  */
-static int _process_cmdline (int argc, char **argv)
+static void _process_cmdline(int argc, char **argv)
 {
-	if ((argc == 2) && (xstrcmp(argv[1], "getenv") == 0)) {
+	if ((argc == 2) && !xstrcmp(argv[1], "getenv")) {
 		print_rlimits();
-		_dump_user_env();
+		for (int i = 0; environ[i]; i++)
+			printf("%s\n", environ[i]);
 		exit(0);
 	}
-	if ((argc == 2) && (xstrcmp(argv[1], "infinity") == 0)) {
+	if ((argc == 2) && !xstrcmp(argv[1], "infinity")) {
 		set_oom_adj(-1000);
 		(void) poll(NULL, 0, -1);
 		exit(0);
 	}
-	if ((argc == 3) && (xstrcmp(argv[1], "spank") == 0)) {
+	if ((argc == 3) && !xstrcmp(argv[1], "spank")) {
 		if (_handle_spank_mode(argc, argv) < 0)
-			exit (1);
-		exit (0);
+			exit(1);
+		exit(0);
+	}
+	if (run_command_is_launcher(argc, argv)) {
+		run_command_launcher(argc, argv);
+		_exit(127); /* Should not get here */
 	}
-	return (0);
 }
 
-
 static void
 _send_ok_to_slurmd(int sock)
 {
-	/* If running under valgrind/memcheck, this pipe doesn't work correctly
-	 * so just skip it. */
-#if (SLURMSTEPD_MEMCHECK == 0)
+	/*
+	 * If running under memcheck, this pipe doesn't work correctly so just
+	 * skip it.
+	 */
+#if (SLURMSTEPD_MEMCHECK != 1)
 	int ok = SLURM_SUCCESS;
 	safe_write(sock, &ok, sizeof(int));
 	return;
@@ -506,18 +800,14 @@ rwfail:
 #endif
 }
 
-static void
-_send_fail_to_slurmd(int sock)
+static void _send_fail_to_slurmd(int sock, int rc)
 {
-	/* If running under valgrind/memcheck, this pipe doesn't work correctly
-	 * so just skip it. */
-#if (SLURMSTEPD_MEMCHECK == 0)
-	int fail = SLURM_ERROR;
-
-	if (errno)
-		fail = errno;
-
-	safe_write(sock, &fail, sizeof(int));
+	/*
+	 * If running under memcheck, this pipe doesn't work correctly so just
+	 * skip it.
+	 */
+#if (SLURMSTEPD_MEMCHECK != 1)
+	safe_write(sock, &rc, sizeof(int));
 	return;
 rwfail:
 	error("Unable to send \"fail\" to slurmd");
@@ -527,9 +817,11 @@ rwfail:
 static void
 _got_ack_from_slurmd(int sock)
 {
-	/* If running under valgrind/memcheck, this pipe doesn't work correctly
-	 * so just skip it. */
-#if (SLURMSTEPD_MEMCHECK == 0)
+	/*
+	 * If running under memcheck, this pipe doesn't work correctly so just
+	 * skip it.
+	 */
+#if (SLURMSTEPD_MEMCHECK != 1)
 	int ok;
 	safe_read(sock, &ok, sizeof(int));
 	return;
@@ -545,7 +837,9 @@ static void _set_job_log_prefix(slurm_step_id_t *step_id)
 
 	log_build_step_id_str(step_id, tmp_char, sizeof(tmp_char),
 			      STEP_ID_FLAG_NO_PREFIX);
-	buf = xstrdup_printf("[%s]", tmp_char);
+	buf = xstrdup_printf("[%s%s]",
+			     tmp_char,
+			     job_step_ptr ? " stepmgr" : "");
 
 	setproctitle("%s", buf);
 	/* note: will claim ownership of buf, do not free */
@@ -558,8 +852,8 @@ static void _set_job_log_prefix(slurm_step_id_t *step_id)
  *  sent by _send_slurmstepd_init() in src/slurmd/slurmd/req.c.
  */
 static int
-_init_from_slurmd(int sock, char **argv,
-		  slurm_addr_t **_cli, slurm_addr_t **_self, slurm_msg_t **_msg)
+_init_from_slurmd(int sock, char **argv, slurm_addr_t **_cli,
+		  slurm_msg_t **_msg)
 {
 	char *incoming_buffer = NULL;
 	buf_t *buffer;
@@ -567,7 +861,6 @@ _init_from_slurmd(int sock, char **argv,
 	int len;
 	uint16_t proto;
 	slurm_addr_t *cli = NULL;
-	slurm_addr_t *self = NULL;
 	slurm_msg_t *msg = NULL;
 	slurm_step_id_t step_id = {
 		.job_id = 0,
@@ -579,6 +872,11 @@ _init_from_slurmd(int sock, char **argv,
 	if (!(conf = _read_slurmd_conf_lite(sock)))
 		fatal("Failed to read conf from slurmd");
 
+	/*
+	 * Init select plugin after reading slurm.conf and before receiving step
+	 */
+	select_g_init(false);
+
 	slurm_conf.slurmd_port = conf->port;
 	slurm_conf.slurmd_syslog_debug = conf->syslog_debug;
 	/*
@@ -603,14 +901,19 @@ _init_from_slurmd(int sock, char **argv,
 	safe_read(sock, &step_complete.children, sizeof(int));
 	safe_read(sock, &step_complete.depth, sizeof(int));
 	safe_read(sock, &step_complete.max_depth, sizeof(int));
-	safe_read(sock, &step_complete.parent_addr, sizeof(slurm_addr_t));
+	safe_read(sock, &len, sizeof(int));
+	if (len) {
+		step_complete.parent_name = xmalloc(len + 1);
+		safe_read(sock, step_complete.parent_name, len);
+	}
+
 	if (step_complete.children)
 		step_complete.bits = bit_alloc(step_complete.children);
 	step_complete.jobacct = jobacctinfo_create(NULL);
 	slurm_mutex_unlock(&step_complete.lock);
 
-	debug3("slurmstepd rank %d, parent = %pA",
-	       step_complete.rank, &step_complete.parent_addr);
+	debug3("slurmstepd rank %d, parent = %s",
+	       step_complete.rank, step_complete.parent_name);
 
 	/* receive cli from slurmd */
 	safe_read(sock, &len, sizeof(int));
@@ -622,22 +925,6 @@ _init_from_slurmd(int sock, char **argv,
 		fatal("slurmstepd: problem with unpack of slurmd_conf");
 	FREE_NULL_BUFFER(buffer);
 
-	/* receive self from slurmd */
-	safe_read(sock, &len, sizeof(int));
-	if (len > 0) {
-		/* receive packed self from main slurmd */
-		incoming_buffer = xmalloc(len);
-		safe_read(sock, incoming_buffer, len);
-		buffer = create_buf(incoming_buffer,len);
-		self = xmalloc(sizeof(slurm_addr_t));
-		if (slurm_unpack_addr_no_alloc(self, buffer)
-		    == SLURM_ERROR) {
-			fatal("slurmstepd: problem with unpack of "
-			      "slurmd_conf");
-		}
-		FREE_NULL_BUFFER(buffer);
-	}
-
 	/* Grab the slurmd's spooldir. Has %n expanded. */
 	cpu_freq_init(conf);
 
@@ -670,12 +957,12 @@ _init_from_slurmd(int sock, char **argv,
 		break;
 	}
 
-	/* Init select and switch before unpack_msg to only init the default */
-	if (select_g_init(1) != SLURM_SUCCESS )
-		fatal( "failed to initialize node selection plugin" );
+	/* Init switch before unpack_msg to only init the default */
+	if (switch_g_init(true) != SLURM_SUCCESS)
+		fatal("failed to initialize switch plugin");
 
-	if (switch_init(1) != SLURM_SUCCESS)
-		fatal( "failed to initialize authentication plugin" );
+	if (cred_g_init() != SLURM_SUCCESS)
+		fatal("failed to initialize credential plugin");
 
 	if (gres_init() != SLURM_SUCCESS)
 		fatal("failed to initialize gres plugins");
@@ -691,10 +978,43 @@ _init_from_slurmd(int sock, char **argv,
 		step_id.step_het_comp = NO_VAL;
 		break;
 	case LAUNCH_TASKS:
-		memcpy(&step_id,
-		       &((launch_tasks_request_msg_t *)msg->data)->step_id,
-		       sizeof(step_id));
+	{
+		launch_tasks_request_msg_t *task_msg;
+		task_msg = (launch_tasks_request_msg_t *)msg->data;
+
+		memcpy(&step_id, &task_msg->step_id, sizeof(step_id));
+
+		if (task_msg->job_ptr &&
+		    !xstrcmp(conf->node_name, task_msg->job_ptr->batch_host)) {
+			slurm_addr_t *node_addrs;
+
+			/* only allow one stepd to be stepmgr. */
+			job_step_ptr = task_msg->job_ptr;
+			job_step_ptr->part_ptr = task_msg->part_ptr;
+			job_node_array = task_msg->job_node_array;
+
+			/*
+			 * job_record doesn't pack its node_addrs array, so get
+			 * it from the cred.
+			 */
+			if (task_msg->cred &&
+			    (node_addrs = slurm_cred_get(
+				     task_msg->cred,
+				     CRED_DATA_JOB_NODE_ADDRS))) {
+				add_remote_nodes_to_conf_tbls(
+					job_step_ptr->nodes, node_addrs);
+
+				job_step_ptr->node_addrs =
+					xcalloc(job_step_ptr->node_cnt,
+						sizeof(slurm_addr_t));
+				memcpy(job_step_ptr->node_addrs, node_addrs,
+				       job_step_ptr->node_cnt *
+				       sizeof(slurm_addr_t));
+			}
+		}
+
 		break;
+	}
 	default:
 		fatal("%s: Unrecognized launch RPC (%d)", __func__, step_type);
 		break;
@@ -702,22 +1022,23 @@ _init_from_slurmd(int sock, char **argv,
 
 	_set_job_log_prefix(&step_id);
 
+	if (cgroup_read_state(sock) != SLURM_SUCCESS)
+		fatal("Failed to read cgroup state from slurmd");
+
 	/*
 	 * Init all plugins after receiving the slurm.conf from the slurmd.
 	 */
-	if ((slurm_auth_init(NULL) != SLURM_SUCCESS) ||
+	if ((auth_g_init() != SLURM_SUCCESS) ||
 	    (cgroup_g_init() != SLURM_SUCCESS) ||
 	    (hash_g_init() != SLURM_SUCCESS) ||
 	    (acct_gather_conf_init() != SLURM_SUCCESS) ||
-	    (core_spec_g_init() != SLURM_SUCCESS) ||
-	    (slurm_proctrack_init() != SLURM_SUCCESS) ||
-	    (slurmd_task_init() != SLURM_SUCCESS) ||
+	    (prep_g_init(NULL) != SLURM_SUCCESS) ||
+	    (proctrack_g_init() != SLURM_SUCCESS) ||
+	    (task_g_init() != SLURM_SUCCESS) ||
 	    (jobacct_gather_init() != SLURM_SUCCESS) ||
 	    (acct_gather_profile_init() != SLURM_SUCCESS) ||
-	    (slurm_cred_init() != SLURM_SUCCESS) ||
 	    (job_container_init() != SLURM_SUCCESS) ||
-	    (route_init() != SLURM_SUCCESS) ||
-	    (slurm_topo_init() != SLURM_SUCCESS))
+	    (topology_g_init() != SLURM_SUCCESS))
 		fatal("Couldn't load all plugins");
 
 	/*
@@ -766,7 +1087,6 @@ _init_from_slurmd(int sock, char **argv,
 	msg->protocol_version = proto;
 
 	*_cli = cli;
-	*_self = self;
 	*_msg = msg;
 
 	return 1;
@@ -776,8 +1096,7 @@ rwfail:
 	exit(1);
 }
 
-static stepd_step_rec_t *
-_step_setup(slurm_addr_t *cli, slurm_addr_t *self, slurm_msg_t *msg)
+static stepd_step_rec_t *_step_setup(slurm_addr_t *cli, slurm_msg_t *msg)
 {
 	stepd_step_rec_t *step = NULL;
 
@@ -788,7 +1107,7 @@ _step_setup(slurm_addr_t *cli, slurm_addr_t *self, slurm_msg_t *msg)
 		break;
 	case REQUEST_LAUNCH_TASKS:
 		debug2("setup for a launch_task");
-		step = mgr_launch_tasks_setup(msg->data, cli, self,
+		step = mgr_launch_tasks_setup(msg->data, cli,
 					      msg->protocol_version);
 		break;
 	default:
@@ -840,7 +1159,9 @@ _step_setup(slurm_addr_t *cli, slurm_addr_t *self, slurm_msg_t *msg)
 				    step->step_id.job_id,
 				    step->step_id.step_id);
 	}
-	if (step->batch || (step->step_id.step_id == SLURM_INTERACTIVE_STEP)) {
+	if (step->batch ||
+	    (step->step_id.step_id == SLURM_INTERACTIVE_STEP) ||
+	    (step->flags & LAUNCH_EXT_LAUNCHER)) {
 		gres_g_job_set_env(step, 0);
 	} else if (msg->msg_type == REQUEST_LAUNCH_TASKS) {
 		gres_g_step_set_env(step);
@@ -853,11 +1174,11 @@ _step_setup(slurm_addr_t *cli, slurm_addr_t *self, slurm_msg_t *msg)
 			    conf->node_topo_addr);
 	env_array_overwrite(&step->env,"SLURM_TOPOLOGY_ADDR_PATTERN",
 			    conf->node_topo_pattern);
-	/*
-	 * Reset address for cloud nodes
-	 */
-	if (step->alias_list && set_nodes_alias(step->alias_list)) {
-		error("%s: set_nodes_alias failed: %s", __func__,
+
+	/* Reset addrs for dynamic/cloud nodes to hash tables */
+	if (step->node_addrs &&
+	    add_remote_nodes_to_conf_tbls(step->node_list, step->node_addrs)) {
+		error("%s: failed to add node addrs: %s", __func__,
 		      step->alias_list);
 		stepd_step_rec_destroy(step);
 		return NULL;
@@ -878,33 +1199,13 @@ _step_cleanup(stepd_step_rec_t *step, slurm_msg_t *msg, int rc)
 			stepd_step_rec_destroy(step);
 	}
 
-	if (msg) {
-		/*
-		 * The message cannot be freed until the jobstep is complete
-		 * because the job struct has pointers into the msg, such
-		 * as the switch jobinfo pointer.
-		 */
-		switch(msg->msg_type) {
-		case REQUEST_BATCH_JOB_LAUNCH:
-			slurm_free_job_launch_msg(msg->data);
-			break;
-		case REQUEST_LAUNCH_TASKS:
-			slurm_free_launch_tasks_request_msg(msg->data);
-			break;
-		default:
-			fatal("handle_launch_message: Unrecognized launch RPC");
-			break;
-		}
-		xfree(msg);
-	}
+	/*
+	 * The message cannot be freed until the jobstep is complete
+	 * because the job struct has pointers into the msg, such
+	 * as the switch jobinfo pointer.
+	 */
+	slurm_free_msg(msg);
+
 	jobacctinfo_destroy(step_complete.jobacct);
 }
 #endif
-
-static void _dump_user_env(void)
-{
-	int i;
-
-	for (i=0; environ[i]; i++)
-		printf("%s\n",environ[i]);
-}
