@@ -73,7 +73,6 @@
 #include "src/common/spank.h"
 #include "src/common/strlcpy.h"
 #include "src/common/xmalloc.h"
-#include "src/common/xsignal.h"
 #include "src/common/xstring.h"
 
 #include "src/interfaces/auth.h"
@@ -136,16 +135,6 @@ static void _proc_stdout(char *buf, stepd_step_rec_t *step)
 				equal_ptr--;
 			equal_ptr[0] = '\0';
 			end_line[0] = '\0';
-			if (!xstrcmp(name_ptr, "SLURM_PROLOG_CPU_MASK")) {
-				step->cpu_bind_type = CPU_BIND_MASK;
-				xfree(step->cpu_bind);
-				step->cpu_bind = xstrdup(val_ptr);
-				if (task_g_pre_launch(step)) {
-					error("Failed SLURM_PROLOG_CPU_MASK "
-					      "setup");
-					exit(1);
-				}
-			}
 			debug("export name:%s:val:%s:", name_ptr, val_ptr);
 			if (setenvf(env, name_ptr, "%s", val_ptr)) {
 				error("Unable to set %s environment variable",
@@ -203,6 +192,7 @@ _run_script_and_set_env(const char *name, const char *path,
 		.max_wait = -1,
 		.script_path = path,
 		.script_type = name,
+		.direct_exec = true,
 		.status = &status
 	};
 
@@ -304,7 +294,6 @@ _setup_mpi(stepd_step_rec_t *step, int ltaskid)
 		info->gtaskid = step->het_job_task_offset +
 				step->task[ltaskid]->gtid;
 		info->ltaskid = step->task[ltaskid]->id;
-		info->self    = step->envtp->self;
 		info->client  = step->envtp->cli;
 	} else {
 		info->step_id.step_id  = step->step_id.step_id;
@@ -315,7 +304,6 @@ _setup_mpi(stepd_step_rec_t *step, int ltaskid)
 		info->ltasks  = step->node_tasks;
 		info->gtaskid = step->task[ltaskid]->gtid;
 		info->ltaskid = step->task[ltaskid]->id;
-		info->self    = step->envtp->self;
 		info->client  = step->envtp->cli;
 	}
 
@@ -370,6 +358,7 @@ extern void exec_task(stepd_step_rec_t *step, int local_proc_id)
 	step->envtp->job_licenses = xstrdup(step->job_licenses);
 	step->envtp->job_start_time = step->job_start_time;
 	step->envtp->user_name = xstrdup(step->user_name);
+	step->envtp->oom_kill_step = step->oom_kill_step ? 1 : 0;
 
 	/*
 	 * Modify copy of step's environment. Do not alter in place or
@@ -397,7 +386,7 @@ extern void exec_task(stepd_step_rec_t *step, int local_proc_id)
 
 	if (!step->batch && (step->step_id.step_id != SLURM_EXTERN_CONT) &&
 	    (step->step_id.step_id != SLURM_INTERACTIVE_STEP)) {
-		if (switch_g_job_attach(step->switch_job, &step->env,
+		if (switch_g_job_attach(step->switch_step, &step->env,
 					step->nodeid, (uint32_t) local_proc_id,
 					step->nnodes, step->ntasks,
 					task->gtid + task_offset) < 0) {
@@ -448,7 +437,7 @@ extern void exec_task(stepd_step_rec_t *step, int local_proc_id)
 
 #ifdef WITH_SELINUX
 	if (setexeccon(step->selinux_context)) {
-		error("Failed to set SELinux context to %s: %m",
+		error("Failed to set SELinux context to '%s': %m",
 		      step->selinux_context);
 		_exit(1);
 	}
@@ -543,12 +532,12 @@ extern void exec_task(stepd_step_rec_t *step, int local_proc_id)
 			eol = strchr(buf, '\n');
 			if (eol)
 				eol[0] = '\0';
-			slurm_seterrno(saved_errno);
+			errno = saved_errno;
 			error("execve(): bad interpreter(%s): %m", buf+2);
 			_exit(errno);
 		}
 	}
-	slurm_seterrno(saved_errno);
+	errno = saved_errno;
 	error("execve(): %s: %m", task->argv[0]);
 	_exit(errno);
 }
